/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001,2002 Robert Chatley, Alan Chan                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

/*
 * DataFlowAnalyser.java
 *
 * Data flow analyser for kenya programs. Attempts to identify
 * usage of variables before definition.
 */

package com.chatley.kenya;

import com.chatley.kenya.minijava.analysis.*;
import com.chatley.kenya.minijava.node.*;
import com.chatley.kenya.utils.*;

import java.util.*;
import java.io.*;


public class DataFlowAnalyser extends DepthFirstAdapter
{
    /** All error output should be written to this stream.*/
    ErrorMsgWriter mErrorDest;

    /** The name of the function we are currently checking
     *  (if we're checking one).*/
    private String mCurrentFuncName;

    /** The state of variables in the enclosing scope.*/
    private Hashtable mParentVars = null;

    /** The state of variable in this scope.*/
    Hashtable mVars;

    private Vector mUninitialisedVars;

    /** whether any errors have been detected by the analyser.*/
    private boolean mErrorsDetected = false;

    /*
     * Set this variable to true to indicate that we are in a block
     * where the break statement is valid, i.e. inside a loop or switch
     * block.
     */
    private boolean mBreakable = false;

    /** Set this if a break statement has been encountered in a block*/
    private boolean mBreakFound = false;

    /** Set this if a return statement has been encountered in a block*/
    private boolean mReturnFound = false;

    /** Set this if we have found an infinite while loop.*/
    private boolean mInfiniteLoop = false;

    public DataFlowAnalyser(ErrorMsgWriter errorDest)
    {
        mCurrentFuncName = "";
        mErrorDest = errorDest;
        mVars = new Hashtable();
    }

    private DataFlowAnalyser(ErrorMsgWriter errorDest, String aFuncName,
      Hashtable aParentVars)
    {
        this(errorDest);
        mCurrentFuncName = aFuncName;
        mParentVars = aParentVars;
    }

    public boolean errorsHaveOccurred()
    {
        return mErrorsDetected;
    }

    ///////////////////////////////////////////////////////////////

    // Empty body. We don't need class declarations to be analysed.
    // The only thing which could be analysed is for duplicate declarations
    // of variables, which should have been done by the type checker,
    // so there's no point us doing it again (is there?).
    public void caseAClassDecDeclaration(AClassDecDeclaration node){}


    public void caseAConstDecDeclaration(AConstDecDeclaration node)
    {
        int ln = node.getConst().getLine();

        // if a break or return has been found previously,
        // this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        String varName = node.getIdentifier().toString().trim();

        IMTrace.out("constname = " + varName);

        if(mVars.containsKey(varName))
        {
            giveError("constant " + varName + "already defined.",
                      node.getSemicolon().getLine());
        }
        else
        {
            if(node.getInitialiser()!=null)
            {
                if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
                else mUninitialisedVars.clear();

                node.getInitialiser().apply(this);

                if(!printExpressionErrors(ln))
                    mVars.put(varName,new Boolean(true));
                else
                    mVars.put(varName,new Boolean(false));
            }
            else
                mVars.put(varName,new Boolean(false));
        }
    }

    public void caseAVarDecDeclaration(AVarDecDeclaration node)
    {
        int ln = node.getSemicolon().getLine();

        // if a break or return has been found previously,
        // this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        String varName = node.getIdentifier().toString().trim();

        IMTrace.out("varname = " + varName);

        if(mVars.containsKey(varName))
        {
            giveError("variable " + varName + " already defined.",
                      node.getSemicolon().getLine());
        }
        else
        {
            if(!isPrimitiveType(node.getType().toString().trim()))
            {
                mVars.put(varName, new Boolean(true));
            }
            else if(node.getInitialiser()!=null)
            {
                if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
                else mUninitialisedVars.clear();

                node.getInitialiser().apply(this);

                if(!printExpressionErrors(ln))
                    mVars.put(varName,new Boolean(true));
                else
                    mVars.put(varName,new Boolean(false));
            }
            else
                mVars.put(varName,new Boolean(false));
        }
    }

    public void caseAFuncDecDeclaration(AFuncDecDeclaration node)
    {
        String fName = node.getIdentifier().toString().trim();

        DataFlowAnalyser forFunc =
          new DataFlowAnalyser(mErrorDest,fName,mVars);

        if(node.getFormalParamList() != null)
        {
            StringTokenizer st = new StringTokenizer(node.getFormalParamList().toString().trim(),"[] ,");
            while (st.hasMoreTokens()) // get param names and types
            {
                 String type, name;
                 type = st.nextToken();
                 name = st.nextToken();
                 forFunc.mVars.put(name,new Boolean(true));
            }
        }

        node.getBlock().apply(forFunc);
        if(forFunc.errorsHaveOccurred()) mErrorsDetected = true;

        // check that a function that is declared as returning a value
        // actually does
        if(!node.getType().toString().trim().equals("void") &&
           !forFunc.returnFound())
        {
            giveError("Function might not return a value, " +
              "return statement missing.",node.getLParenthese().getLine());
        }
    }

    public void caseAArrayDecDeclaration(AArrayDecDeclaration node)
    {
        String varName = node.getIdentifier().toString().trim();
        int ln = node.getSemicolon().getLine();

        // if a break or return has been found previously,
        // this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        IMTrace.out("Array dec " + varName);
        if(mVars.containsKey(varName))
        {
            giveError("variable " + varName + "already defined.",
                      node.getSemicolon().getLine());
        }
        else
        {
            if(node.getArrayInitialiser()!=null)
            {
                if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
                else mUninitialisedVars.clear();

                node.getArrayInitialiser().apply(this);

                if(!printExpressionErrors(ln))
                    mVars.put(varName,new Boolean(true));
                else
                    mVars.put(varName,new Boolean(false));
            }
            else
                mVars.put(varName,new Boolean(false));
        }
    }

    ///////////////////////////////////////////////////////////////
    public void caseAAssignmentStatement(AAssignmentStatement node)
    {
        String lhs = node.getFieldAccess().toString().trim();
        int ln = node.getSemicolon().getLine();

        // if a break has been found previously, this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        lhs = trimDotsAndBrackets(lhs);
        IMTrace.out("Intialised lhs = " + lhs);

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        node.getExpression().apply(this);

        if(!printExpressionErrors(ln))
            mVars.put(lhs,new Boolean(true));
        else
            mVars.put(lhs,new Boolean(false));
    }

    public void caseABreakStatement(ABreakStatement node)
    {
        int ln = node.getBreak().getLine();

        // if a break or return has been found previously,
        // this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        mBreakFound = true; // flag that a break statement has been found.
        if(!mBreakable)
            giveError("Break found outside of loop or switch.",ln);
    }

    public void caseAIfStatement(AIfStatement node)
    {
        int ln = node.getIf().getLine();

        // if a break has been found previously, this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        // check the IF condition
        node.getBoolExpression().apply(this);
        printExpressionErrors(ln);

        DataFlowAnalyser _if, _else;
        // analyse the IF block
        _if = new DataFlowAnalyser(mErrorDest,"",mVars);
        _if.setBreakable(mBreakable);
        node.getBlock1().apply(_if);
        if(_if.errorsHaveOccurred()) mErrorsDetected = true;
        String if_cond = node.getBoolExpression().toString().trim();

        // analyse ELSE block if one exists, update var initialised tables.
        if(node.getBlock2()!=null)
        {
            _else = new DataFlowAnalyser(mErrorDest,"",mVars);
            _else.setBreakable(mBreakable);
            node.getBlock2().apply(_else);
            if(_else.errorsHaveOccurred()) mErrorsDetected = true;

            if_updateVarTable(mVars,_if,_else);
            if(mParentVars!=null) if_updateVarTable(mParentVars,_if,_else);

            // if a return statement was found in both the if and else blocks,
            // then any statement which follows the if statement is unreachable
            if(_if.returnFound() && _else.returnFound()) mReturnFound = true;
        }

        // if IF condition is the constant literal 'true' or '!false',
        // update var tables.
        if(if_cond.equals("true") || if_cond.equals("! false"))
        {
            updateVarTable(mVars,_if);
            updateVarTable(mParentVars,_if);
        }
    }

    public void caseAWhileStatement(AWhileStatement node)
    {
        int ln = node.getWhile().getLine();

        // if a break or return has been found previously,
        // this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        node.getBoolExpression().apply(this);
        printExpressionErrors(ln);

        IMTrace.out("while condition = " + node.getBoolExpression().toString().trim());

        DataFlowAnalyser _while = new DataFlowAnalyser(mErrorDest,"",mVars);
        _while.setBreakable();
        node.getBlock().apply(_while);
        if(_while.errorsHaveOccurred()) mErrorsDetected =true;

        String while_cond = node.getBoolExpression().toString().trim();
        if(while_cond.equals("true") || while_cond.equals("! false")){
            updateVarTable(mVars,_while);
            if(mParentVars!=null) updateVarTable(mParentVars,_while);

            // detects infinite while loops
            if(!_while.breakFound()) this.mInfiniteLoop = true;
        }
    }

    public void caseAPrintlnStrStatement(APrintlnStrStatement node)
    {
        int ln = node.getSemicolon().getLine();

        // if a break or return has been found previously,
        // this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        node.getExpression().apply(this);

        printExpressionErrors(ln);
    }

    public void caseAPrintStrStatement(APrintStrStatement node)
    {
        int ln = node.getSemicolon().getLine();

        // if a break or return has been found previously,
        // this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        node.getExpression().apply(this);

        printExpressionErrors(ln);
    }

    public void caseAForStatement(AForStatement node)
    {
        int ln = node.getFor().getLine();

        // if a break or return has been found previously,
        // this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        node.getForControl().apply(this);

        DataFlowAnalyser _for = new DataFlowAnalyser(mErrorDest,"",mVars);
        _for.setBreakable();
        node.getBlock().apply(_for);
        if(_for.errorsHaveOccurred())mErrorsDetected =true;
    }

    public void caseAJavaForControl(AJavaForControl node)
    {
        int ln = node.getLParenthese().getLine();

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        String var = node.getName().toString().trim();
        IMTrace.out("for loop control var = "+ var);

        node.getExpression().apply(this);
        if(!printExpressionErrors(ln)) mVars.put(var,new Boolean(true));

        node.getBoolExpression().apply(this);
        node.getUnaryExp().apply(this);
        printExpressionErrors(ln);
    }

    public void caseASimpleForControl(ASimpleForControl node)
    {
        int ln = node.getAssign().getLine();

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        String var = node.getName().toString().trim();
        IMTrace.out("simple for loop control var = "+ var);

        node.getExp1().apply(this);
        if(!printExpressionErrors(ln)) mVars.put(var,new Boolean(true));

        node.getExp2().apply(this);
        if(node.getStepBit()!=null) node.getStepBit().apply(this);
        printExpressionErrors(ln);
    }

    public void caseAReturnStatement(AReturnStatement node)
    {
        int ln = node.getReturn().getLine();

        // if a break or return has been found previously,
        // this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        if(node.getExpression()!=null){
            node.getExpression().apply(this);
            printExpressionErrors(ln);
        }
        mReturnFound = true;
    }

    public void caseASwitchStatement(ASwitchStatement node)
    {
        int ln = node.getSwitch().getLine();

        // if a break or return has been found previously,
        // this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        node.getExpression().apply(this);
        printExpressionErrors(ln);
        node.getSwitchBlock().apply(this);
    }

    public void caseASwitchBlock(ASwitchBlock node)
    {
        Hashtable assignedVars = new Hashtable();
        DataFlowAnalyser _caseBlk;
        boolean defaultCaseExists = false;
        boolean returnStatInAllCases = true;

        Object temp[] = node.getPossibleCase().toArray();
        for(int i = 0; i < temp.length; i++)
        {
            _caseBlk = new DataFlowAnalyser(mErrorDest,"",mVars);
            _caseBlk.setBreakable();
            ((PPossibleCase) temp[i]).apply(_caseBlk);
            if(_caseBlk.errorsHaveOccurred())mErrorsDetected =true;

            if(((APossibleCase) temp[i]).getSwitchLabel()
              instanceof ADefaultSwitchLabel) defaultCaseExists = true;

            returnStatInAllCases &= _caseBlk.returnFound();
            switch_var_count_update(assignedVars,mVars,_caseBlk);
            //if(mParentVars!=null)
            //    switch_var_count_update(assignedVars,mParentVars,_caseBlk);

        }

        // nothing is definitely assigned in a switch statement if there
        // isn't a default case.
        if(!defaultCaseExists) return;

        // if a a default case existed and a return statement was found
        // in each of the cases of the switch statement, then any statements
        // which follow are unreachable.
        if(returnStatInAllCases) mReturnFound = true;

        switch_updateVarTable(mVars,assignedVars,temp.length);
        //if(mParentVars!=null)
        //    switch_updateVarTable(mParentVars,assignedVars,temp.length);
    }

    public void caseAPossibleCase(APossibleCase node)
    {
        PSwitchLabel S_node = node.getSwitchLabel();
        if(S_node instanceof ACaseSwitchLabel)
        {
            ACaseSwitchLabel _case = (ACaseSwitchLabel) S_node;
            int ln = _case.getCase().getLine();

            if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
            else mUninitialisedVars.clear();

            _case.getExpression().apply(this);
            if(!printExpressionErrors(ln)) node.getBlock().apply(this);
        }
        else node.getBlock().apply(this); // default case
    }

    /*public void caseACaseSwitchLabel(ACaseSwitchLabel node)
    {
        int ln = node.getCase().getLine();

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        node.getExpression().apply(this);
        if(printExpressionErrors(ln)) mVars.clear();
    }*/

    private void switch_var_count_update(Hashtable counts, Hashtable vars,
     DataFlowAnalyser _caseBlk)
    {
        for (Enumeration e = vars.keys();e.hasMoreElements();)
        {
            String var = (String)e.nextElement();

            // if variable already initialised, no need to check again
            if(((Boolean)vars.get(var)).booleanValue())continue;

            if(_caseBlk.mVars.containsKey(var))
            {
                if(((Boolean)_caseBlk.mVars.get(var)).booleanValue())
                {
                    if(counts.containsKey(var))
                    {
                        int n = ((Integer)counts.get(var)).intValue();
                        counts.put(var,new Integer((++n)));
                    }else{
                        counts.put(var,new Integer(1));
                    }
                }
            }
        }
    }

    public void caseAAssertStatement(AAssertStatement node)
    {
        int ln = node.getAssert().getLine();

        // if a break or return has been found previously,
        // this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        node.getBoolExpression().apply(this);
        if(node.getColonString()!=null) node.getColonString().apply(this);
        printExpressionErrors(ln);
    }

    public void caseAArraycopyStatement(AArraycopyStatement node)
    {
        int ln = node.getArraycopy().getLine();

        // if a break or return has been found previously,
        // this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        String src = node.getSrc().toString(),
              dest = node.getDest().toString();

        src = trimDotsAndBrackets(src);
        dest = trimDotsAndBrackets(dest);

        if(!mVars.containsKey(src) ||
           !((Boolean)mVars.get(src)).booleanValue())
        {
            giveError("variable: " + src +
                " might not have been initialised.",ln);
        }

        if(!mVars.containsKey(dest) ||
           !((Boolean)mVars.get(dest)).booleanValue())
        {
            giveError("variable: " + dest +
                " might not have been initialised.",ln);
        }
    }

    // ------------------- graphics primitives --------------------

    public void caseADrawlineStatement(ADrawlineStatement node)
    {
        int ln = node.getDrawline().getLine();

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();
        node.getFourNumbers().apply(this);
        printExpressionErrors(ln);
    }

    public void caseADrawboxStatement(ADrawboxStatement node)
    {
        int ln = node.getDrawbox().getLine();

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();
        node.getFourNumbers().apply(this);
        printExpressionErrors(ln);
    }

    public void caseADrawtriangleStatement(ADrawtriangleStatement node)
    {
        int ln = node.getDrawtriangle().getLine();

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();
        node.getSixNumbers().apply(this);
        printExpressionErrors(ln);
    }

    public void caseADrawellipseStatement(ADrawellipseStatement node)
    {
        int ln = node.getDrawellipse().getLine();

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();
        node.getFourNumbers().apply(this);
        printExpressionErrors(ln);
    }

    public void caseADrawstringStatement(ADrawstringStatement node)
    {
        int ln = node.getDrawstring().getLine();

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        node.getString().apply(this);
        node.getX().apply(this);
        node.getY().apply(this);
        printExpressionErrors(ln);
    }

    public void caseASetcolourStatement(ASetcolourStatement node)
    {
        int ln = node.getSetcolour().getLine();

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();
        node.getRgb().apply(this);
        printExpressionErrors(ln);
    }

    public void caseASetbgcolourStatement(ASetbgcolourStatement node)
    {
        int ln = node.getSetbgcolour().getLine();

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();
        node.getRgb().apply(this);
        printExpressionErrors(ln);
    }

    public void caseAFillStatement(AFillStatement node)
    {
        int ln = node.getFill().getLine();

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        node.getBoolExpression().apply(this);
        printExpressionErrors(ln);
    }

    // ------------------ function call ---------------------------

    public void caseAFunctionApplication(AFunctionApplication node)
    {
        int ln = node.getLParenthese().getLine();

        // if a break has been found previously, this statement is unreachable
        if(mBreakFound || mReturnFound || mInfiniteLoop){
            giveError("Statement unreachable.",ln);
            return;
        }

        if(mUninitialisedVars==null) mUninitialisedVars = new Vector();
        else mUninitialisedVars.clear();

        node.getActualParamList().apply(this);
        printExpressionErrors(ln);
    }

    ///////////////////////////////////////////////////////////////

    public void caseAFieldaccessFactor(AFieldaccessFactor node)
    {
        String var = node.getFieldAccess().toString();

        var = trimDotsAndBrackets(var);

        IMTrace.out("caseAFieldaccessFactor: var = " + var + ".");

        if(mVars.containsKey(var))
        {
            Boolean b = (Boolean)mVars.get(var);

            if(!b.booleanValue()) {
                mUninitialisedVars.add(var);
                IMTrace.out("var " + var + " not initialised");
            }
        }
        else if(mParentVars.containsKey(var))
        {
            if(!((Boolean)mParentVars.get(var)).booleanValue()){
                mUninitialisedVars.add(var);
                IMTrace.out("var " + var + " not initialised");
            }
        }
    }


    public void setBreakable()
    {
        setBreakable(true);
    }

    public void setBreakable(boolean b)
    {
        mBreakable = b;
    }

    /**
     * returns true if a break statement has been found in the current
     * scope of the program so far.
     */
    public boolean breakFound()
    {
        return mBreakFound;
    }

    /**
     * returns true if a return statement has been found in the current
     * scope of the program so far.
     */
    public boolean returnFound()
    {
        return mReturnFound;
    }

    ///////////////////////////////////////////////////////////////


    private String trimDotsAndBrackets(String varName)
    {
        String var = new String(varName);
        var = (var.indexOf('[')!=-1)?var.substring(0,var.indexOf('[')):var;
        var = (var.indexOf('.')!=-1)?var.substring(0,var.indexOf('.')):var;
        var = var.trim();
        return var;
    }

    private void giveError( String s , int line )
    {
        mErrorsDetected = true;
        mErrorDest.giveError("Error (line " + line + "): " + s + "\n",line);
    }

    private void giveError( String s )
    {
        mErrorsDetected = true;
        mErrorDest.giveError("Error : " + s + "\n");
    }

    private boolean printExpressionErrors(int line)
    {
        boolean errorsOccurred = false;
        if(!mUninitialisedVars.isEmpty()){
            errorsOccurred = true;
            for(int i=0;i<mUninitialisedVars.size();i++)
                giveError("variable: " + (String)mUninitialisedVars.get(i) +
                " might not have been initialised.",line);
        }
        return errorsOccurred;
    }

    /**
     * updates the table of declared variables, if a variable is initialised
     * in the scope of both the IF and ELSE parts, then we can conclude
     * that it is initialised in the current scope too.
     */
    private void if_updateVarTable(Hashtable vars, DataFlowAnalyser _if,
      DataFlowAnalyser _else)
    {
        for (Enumeration e = vars.keys();e.hasMoreElements();)
        {
            String var = (String)e.nextElement();

            if(((Boolean)vars.get(var)).booleanValue())continue;

            if(_if.mVars.containsKey(var) &&
               _else.mVars.containsKey(var))
            {
                boolean b1, b2;
                b1 = ((Boolean)_if.mVars.get(var)).booleanValue();
                b2 = ((Boolean)_else.mVars.get(var)).booleanValue();
                if(b1&&b2)
                    mVars.put(var,new Boolean(true));
            }
         }
    }

    /**
     * updates the table of declared variables, if a variable is initialised
     * in the scope of every case of a SWITCH statement, then we can conclude
     * that it is initialised in the current scope.
     */
    private void switch_updateVarTable(Hashtable vars, Hashtable counts,
     int numOfCases)
    {
        for (Enumeration e = vars.keys();e.hasMoreElements();)
        {
            String var = (String)e.nextElement();

            if(((Boolean)vars.get(var)).booleanValue())continue;

            if(counts.containsKey(var))
            {
                if(((Integer)counts.get(var)).intValue()==numOfCases)
                    mVars.put(var,new Boolean(true));
            }
        }
    }

    private void updateVarTable(Hashtable vars, DataFlowAnalyser stmt)
    {
        for (Enumeration e = vars.keys();e.hasMoreElements();)
        {
            String var = (String)e.nextElement();

            if(((Boolean)vars.get(var)).booleanValue())continue;

            if(stmt.mVars.containsKey(var))
            {
                if(((Boolean)stmt.mVars.get(var)).booleanValue())
                    mVars.put(var,new Boolean(true));
            }
        }
    }


    private boolean isPrimitiveType(String type)
    {
        return ( type.equals("int") || type.equals("double") ||
                 type.equals("String") || type.equals("boolean") ||
                 type.equals("char") );
    }
}