/*********************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001 Robert Chatley                                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

// >>> askc98 2002/2
package com.chatley.kenya;
// <<< askc98

import javax.swing.*;
import java.awt.event.*;
import java.awt.*;
import java.io.*;
import java.util.Properties;
import com.chatley.kenya.gui.*;
import com.chatley.kenya.utils.IMTrace;

public class Kenya {

	public static void main( String[] args )
	{
		int i = 0;
		boolean nowindowmode = false;
		boolean compile = false;
		boolean savejava = false;
		String arg;
		String config = "config";
		String filename = "";
		Properties prop = null;
		String slash = System.getProperty("file.separator");

		while ( i < args.length && args[i].startsWith("-")) {

			arg = args[i++];

			if ( arg.equals("-help") || arg.equals("--help") ) {
				System.out.println("Kenya is a teaching language, designed by Robert Chatley 2001");
				System.out.println("Usage: java Kenya [-nowindow] [-savejava] [-compile] [-debug] filename\n");
				System.out.println("To launch graphical interface use no arguments.");
				System.out.println("To run on the command line use -nowindow and supply a Kenya file.");
				System.out.println("This will output Java code to the console.");
				System.out.println("To create a Java source file, use -savejava in conjunction with -nowindow");
				System.out.println("This will create a package directory containing a Java file.");
				System.out.println("To create a .class file, use -compile in conjunction with -nowindow.");
				System.out.println("This will create a package directory containing a class file and a Java file.");
				System.out.println("To print debug messages, use the -debug switch");
				return;
			}

			else if ( arg.equals("-config") ) {
				if(i+1 == args.length) config = args[i++];
				else {
				    System.err.println("Error: No config file specified.");
				    System.exit(-1);
				}
			}

			else if ( arg.equals("-nowindow") ) {
				nowindowmode = true;
			}

			else if ( arg.equals("-savejava") ) {
				savejava = true;
			}

			else if ( arg.equals("-compile") ) {
				compile = true;
			}

			else if ( arg.equals("-debug")){
			    IMTrace.setDebugOn();
			}
		}

		if ( i < args.length ) {
			filename = args[i];
		}


		//get properties from config file

		try {
			FileInputStream in = new FileInputStream( config );
			prop = new Properties();
			prop.load(in);
		}
		catch ( IOException e ) {System.err.println( "Problem encountered loading default config file. Quitting." ); System.exit(-1); }

		//if the user has a .kenya file in their home directory, load properties from that

		String lookingfor =  System.getProperty( "user.home" ) + System.getProperty( "file.separator" ) + ".kenya";
		File customconfig = new File( lookingfor );

		if ( customconfig.exists() ) {
			System.err.println( "Using personal configuration file: " + lookingfor );
			try {
			FileInputStream in = new FileInputStream( customconfig );
			prop.load(in);
			}
			catch ( IOException e ) { System.err.println( "Problem encountered loading custom config file."); }
		}


		//batch mode processing

		if ( nowindowmode ) {

			if ( filename.equals("") ) {
				System.err.println("You must supply an input file name when running in no window mode.");
				return;
			}

			CommandLineProcessor c = new CommandLineProcessor( filename , compile , savejava , prop );
			c.process();

		} else {

			//create GUI

			/*try{
				UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
			}catch(Exception e){
				System.err.println("Unable to set look and feel.");
			}*/
            Util.showSplash();

			final KenyaUI ui = new KenyaUI( prop );

			ui.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);

			ui.addWindowListener(new WindowAdapter() {
				public void windowClosing(WindowEvent e) {

				if (SaveListener.isDirty()) {
						int save = JOptionPane.showConfirmDialog( ui , "Save current Kenya before quitting?" , "Quit" , JOptionPane.YES_NO_CANCEL_OPTION );

						if ( save == JOptionPane.CANCEL_OPTION ) {
							return;
						} else if ( save == JOptionPane.YES_OPTION ) {
							SaveListener.save(ui,false);
						}
				}
				ui.dispose();
				System.exit(0);
				}
			});

            Util.advanceSplashProgress();

			//load a file if one was specified in the command line args

			if ( ! filename.equals("") ) {

			try {

				BufferedReader infile = new BufferedReader( new FileReader( filename ) );

				while ( infile.ready() )
				{
					ui.kenyaArea().append( infile.readLine() );
					ui.kenyaArea().append("\n");
				}
				infile.close();
				ui.setMsg("Opened file " + filename);


				if ( filename.indexOf(slash) >= 0 ) {
					filename = filename.substring( filename.lastIndexOf(slash) + 1 , filename.length() );
				}

				ui.setTitle( "Kenya - " + filename );

				if ( filename.indexOf(".") >= 0 ) {
					ui.setPackageName(filename.substring( 0 , filename.indexOf(".") ));
				} else {
					ui.setPackageName(filename);
				}
			}

			catch (FileNotFoundException e ) {
				if ( nowindowmode ) {
					System.err.println("File not found - " + filename );
					System.exit(-1);
				} else {
					JOptionPane.showMessageDialog(ui, "File not found - " + filename , "Error" , JOptionPane.ERROR_MESSAGE );
				}
			}
			catch (IOException e ) {
				if ( nowindowmode ) {
					System.err.println( "An error occurred while reading file " + filename );
					System.exit(-1);
				} else {
					JOptionPane.showMessageDialog(ui, "An error occurred while reading file " + filename , "Error" , JOptionPane.ERROR_MESSAGE );
				}
			}
			}

			Util.hideSplash();
			ui.pack();
			//Center the window
			Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
			Dimension frameSize = ui.getSize();
			if (frameSize.height > screenSize.height) {
				frameSize.height = screenSize.height;
			}
			if (frameSize.width > screenSize.width) {
				frameSize.width = screenSize.width;
			}
			ui.setLocation((screenSize.width - frameSize.width) / 2, (screenSize.height - frameSize.height) / 2);

			// start and run preloader thread to preload translator classes
			// (speeds things up later  :-)
			(new PreLoaderThread()).start();

			ui.setVisible(true);
		}
	}
}
