/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001 Robert Chatley                                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

// >>> askc98 2002/2
package com.chatley.kenya;
// <<< askc98

/**
 * The typechecker performs two passes to complete type checking, instead of
 * using this class directly, the <code>KTypeChecker</code> class which acts as
 * an interface for invoking the different parts of typechecker and initialising
 * the separate passes should be used.
 *
 * @see KTypeChecker
 */

import com.chatley.kenya.minijava.analysis.*;
import com.chatley.kenya.minijava.node.*;
import com.chatley.kenya.utils.*;

import java.util.*;
import java.io.*;
import java.awt.*;


public class TypeChecker extends DepthFirstAdapter
{
    private boolean nowindowmode;

    ErrorMsgWriter mErrorWriter;

    SymbolTable symbolTable;
    SymbolTable functions;
    HashMap knownTypes;
    boolean checked;
    boolean genCode;
    boolean noMainFlagged;
    String expected;
    String found;
    String currentfunc;
    String currentclass;
    boolean expectingArrayIndex;
    int arraydim;
    int expdim, founddim;
    int curdepth , maxdepth;
    int ln;
    boolean mFirstPass;

    // added for function param check
    Vector mParamsFound = null;

    public TypeChecker(ErrorMsgWriter errorDest, HashMap kt, SymbolTable s)
    {
        this();
        mErrorWriter = errorDest;
        symbolTable = s;
        knownTypes = kt;
        functions = new SymbolTable();

        addStandardTypes();
    }

    public TypeChecker(ErrorMsgWriter errorDest, HashMap kt)
    {
        this();
        mErrorWriter = errorDest;
        symbolTable = new SymbolTable();
        knownTypes = kt;
        functions = new SymbolTable();

        addStandardTypes();
    }

    public TypeChecker(ErrorMsgWriter errorDest)
    {
        this();
        mErrorWriter = errorDest;
        symbolTable = new SymbolTable();
        functions = new SymbolTable();
        knownTypes = new HashMap();

        addStandardTypes();
    }

    public TypeChecker()
    {
        currentfunc = "";
        currentclass = "";
        expected = "";
        found = "";
        checked = true;
        genCode = true;
        nowindowmode = false;
        founddim = expdim = arraydim = 0;
        curdepth = maxdepth = 0;
        expectingArrayIndex = true;
        noMainFlagged = false;
        mFirstPass = true;
    }

    public void setNoWindows() {
        nowindowmode = true;
    }

    private void addStandardTypes() {

        knownTypes.put("void", new SymbolTable( "void" ));
        knownTypes.put("char", new SymbolTable( "char" ));
        knownTypes.put("int",  new SymbolTable( "int" ));
        knownTypes.put("double", new SymbolTable( "double" ));
        knownTypes.put("String", new SymbolTable( "String" ));
        knownTypes.put("boolean", new SymbolTable( "boolean" ));
        knownTypes.put("Vector",  new SymbolTable( "Vector" ));
        knownTypes.put("Stack", new SymbolTable( "Stack" ));
        knownTypes.put("Dictionary", new SymbolTable( "Dictionary" ));
        knownTypes.put("Hashtable", new SymbolTable( "Hashtable" ));
        knownTypes.put("Enumeration", new SymbolTable( "Enumeration" ));
    }

    public SymbolTable getSymbolTable()
    {
        return symbolTable;
    }

    // cla97 >>
    public void caseAListStatements(AListStatements node)
    {
        expected = "";
        founddim = 0;
        // check for existance of main() method in Kenya program
        if ("".equals(currentfunc) && !noMainFlagged && !symbolTable.containsKey("main")) {
            giveError("No main() function found.");
            noMainFlagged = true;
        }
	
        super.caseAListStatements(node);
    }
    // cla97 <<

    public void caseAClassDecDeclaration( AClassDecDeclaration node )
    {

        // >>> askc98 2002/5
        // we only parse and check class declarations on the first pass.
        if(!mFirstPass) return;
        // <<< askc98

        if ( node.getIdentifier() != null )
        {
            String classname = node.getIdentifier().toString().trim();

            if (!knownTypes.containsKey( classname ))
            {
                ln = node.getKlass().getLine();
                giveError("A type with the name " + classname + " has already been defined." , ln );
                return;
            }
            else//( knownTypes.containsKey( classname )/* && knownTypes.get( classname ) != null*/ )
            {
                SymbolTable s = (SymbolTable)knownTypes.get( classname );
                if(!s.incompleteClass){
                    ln = node.getKlass().getLine();
                    giveError("A type with the name " + classname + " has already been defined." , ln );
                    return;
                }
            //} else {

                //knownTypes.put( classname , null );

                TypeChecker ltc = new TypeChecker(mErrorWriter, knownTypes);
                ltc.nowindowmode = nowindowmode;
                ltc.currentclass = classname;

                LinkedList decs = node.getDeclaration();
                ListIterator decsIterator = decs.listIterator(0);

                while( decsIterator.hasNext() )
                {
                    Node t = (Node)decsIterator.next();
                    t.apply(ltc);
                }

                s = ltc.getSymbolTable();

                if ( s == null ) { System.err.println("Symboltable returned from recursed tc is null"); }
                s.isClass = true;
                s.type = classname;
                knownTypes.put( classname , s );
                genCode &= ltc.genCode;
                currentclass = "";

                // we have just finished checking a class declaration, update
                // any 'old'incomplete information held about this class in the
                // symbol table and knowntypes table
                updateClassDefinitions(knownTypes,s);
                updateClassDefinitions(symbolTable,s);
            }
        }
    }

    private void updateClassDefinitions(HashMap original, SymbolTable toUpdate)
    {
        String thisType = "";
        if(original instanceof SymbolTable)
            thisType = ((SymbolTable)original).type;


        // for each field declaration in a class declaration...
        // replace type information held if the field's type == toUpdate.type
        // watch out for recursive class definitions!
        for(Iterator decls = original.keySet().iterator();decls.hasNext();)
        {
            Object o = decls.next();
            SymbolTable s = (SymbolTable)original.get(o);
            if(s.isClass && !s.type.equals(toUpdate.type) &&
               !s.type.equals(thisType))
            {
                IMTrace.out("thisType      ------> " + thisType);
                IMTrace.out("toUpdate.type ------> " + toUpdate.type);
                IMTrace.out("recursive call -----> " + o);
                updateClassDefinitions(s,toUpdate);
            }
            else if(s.isClass && s.type.equals(toUpdate.type)
                    && !s.equals(toUpdate))
            {
                SymbolTable copy = (SymbolTable)toUpdate.clone();
                copy.arraydim = s.arraydim;
                copy.constant = s.constant;
                IMTrace.out("Updating --------> " + o);
                original.put(o,copy);
            }
        }
    }

    public void caseAFuncDecDeclaration( AFuncDecDeclaration node )
    {
        // >>> askc98 2002/5
        // we only parse and check function declarations on the second pass.
        if(mFirstPass) return;
        // <<< askc98

        String name = node.getIdentifier().toString().trim();
        String type =  node.getType().toString().trim();

        ln = node.getIdentifier().getLine();

        if ( symbolTable.containsKey( name ) && ! ((SymbolTable)symbolTable.get( name )).prototypeOnly )
        {
            giveError("A function with the name " + name + " has already been defined." , ln );
            return;
        }
        else
        {
            if (! knownTypes.containsKey( type ) ) {
                giveError("The return type " + type + " of function " + name + " is unknown." , ln );
            } else {

                FuncInfo params = ((SymbolTable)symbolTable.get(name)).funcInfo;

                TypeChecker forfunc = new TypeChecker(mErrorWriter, knownTypes , (SymbolTable)/*functions.clone()*/symbolTable.clone() );
                forfunc.nowindowmode = nowindowmode;

                forfunc.currentfunc = name;

                if ( node.getFormalParamList() != null ) {
                    node.getFormalParamList().apply( forfunc );
                }

                if ( node.getBlock() != null ) {
                    node.getBlock().apply( forfunc );
                }

                SymbolTable lst = forfunc.getSymbolTable();
                lst.type = type;
                lst.isFunction = true;
                lst.funcInfo = params;

                if ( node.getBracketPair().size() > 0 ) {
                    lst.arraydim = node.getBracketPair().size();
                }

                symbolTable.put( name , lst );
                genCode &= forfunc.genCode;

                currentfunc = "";
            }
        }
    }

    public void caseAReturnStatement( AReturnStatement node )
    {
	// rbc 11/9/02 >>>
	// fixes problem with return statements in switch blocks
	// not sure this is the optimum fix but it seems to work

	checked = true;

	// <<<

        ln = node.getReturn().getLine();

        if ( currentfunc.equals("") ) {
            giveError("return found outside a function" , ln );
            return;
        }

        SymbolTable sk = (SymbolTable)symbolTable.get( currentfunc );
        String type = sk.type;

        if ( type.equals("void") ) {
            if ( node.getExpression() != null )
                giveError("Value returned from void function " + currentfunc , ln );

            return;
        }

        // askc98 >>>
        if(!type.equals("void") && node.getExpression()==null)
        {
            giveError("Return value of type " + type + " expected.",ln);
            return;
        }

        if ( node.getExpression() == null && type.equals("void") ) {
            checked &= true;
            return;
        }

	/*
	  // don't think this is useful
        if ( node.getExpression() == null ) {
            checked = false;
            return;
	

	    } else { */

	// node.getExpression != null
	// type != void

            expected = type;
            founddim = 0;

	    node.getExpression().apply(this);

            if(founddim!=sk.arraydim){
                giveError("Return value of type " +
                  typeDimErrorString(type,sk.arraydim) + " expected, " +
                  typeDimErrorString(found,arraydim) + " found.",ln);
            }

            if ( expected.equals( "double" ) ) {
                checked &= ( found.equals("int") || found.equals("double") );
            } else if ( expected.equals( "int" ) ) {
                checked &= ( found.equals( "int" ) || found.equals("char") );
            } else {
                checked &= ( expected.equals(found) || expected.equals("") );
            } 

	    //}

        if ( /*! expected.equals( found ) */!checked) {
            giveError("Value returned from function " + currentfunc + " should have type " + type , ln );
        }
    }

    public void caseATypeName( ATypeName node )
    {
        String name = node.getIdentifier().toString().trim();
        String type = node.getType().toString().trim();
        ln = node.getIdentifier().getLine();

        if ( node.getBracketPair() != null ) {
            arraydim = node.getBracketPair().size();
        }

        if (! knownTypes.containsKey( type ) ) {
            giveError("The type " + type + " for function parameter " + name + " is unknown." , ln );

        } else {

            SymbolTable st;

            if ( arraydim > 0 ) {
                st  = (SymbolTable)((SymbolTable)knownTypes.get(type)).clone();
                st.arraydim = arraydim;
                st.isClass = true;
                st.put( "length" , (SymbolTable)knownTypes.get( "int" ) );
            } else {
                st = (SymbolTable)knownTypes.get( type );
            }

            symbolTable.put( name , st );
        }
    }

    public void caseAConstDecDeclaration( AConstDecDeclaration node )
    {
        // >>> askc98 2002/5
        // we only parse and check constants on the second pass if they
        // occur within functions.
        if(!mFirstPass && currentfunc.equals("")) return;
        // <<< askc98

        String id = node.getIdentifier().toString().trim();
        String type =  node.getType().toString().trim();
        ln = node.getConst().getLine();

        if ( symbolTable.containsKey( id ) ) {
            giveError("A variable with the name " + id + " has already been declared." , ln );
            return;
        } else {

            if (! knownTypes.containsKey( type ) ) {
            giveError("The type " + type + " has not been defined." , ln );

            } else {

            if ( node.getInitialiser() != null ) {
                expected = type;
                node.getInitialiser().apply( this );
                if( !checked ) {
                giveError( "Constant " + id + " must be initialised with type " + type , ln );
                return;
                }

            }

            SymbolTable st = (SymbolTable)((SymbolTable)knownTypes.get( type )).clone();
            st.constant = true;
                symbolTable.put( id , st );
            }
        }
    }

    public void caseAVarDecDeclaration( AVarDecDeclaration node )
    {
        String type =  node.getType().toString();
        ln = node.getIdentifier().getLine();

        if("".equals(currentfunc) && "".equals(currentclass)){
            giveError("Global statements are not allowed.",ln);
        }

        type = type.trim();
        arraydim = 0;


        if ( !knownTypes.containsKey( type ) ) {
            giveError( "The type " + type + " has not been defined." , ln );
            return;
        }

        String id = node.getIdentifier().toString().trim();

        IMTrace.out("varDec :: type = " + type + " id = " + id + " currentclass = " + currentclass);

        if ( symbolTable.containsKey( id ) ) {
            giveError("A variable with the name " + id + " has already been declared." , ln );
            return;
        }

        if ( node.getTypeParam() != null ) {
            node.getTypeParam().apply(this);
        }

        Node init = node.getInitialiser();

        if ( init != null ) {
            expected = type;
            init.apply( this );
            if( !checked ) {
                giveError( "Variable " + id + " must be initialised with type " + type , ln );
                return;
            }

        }

        if ( type.equals( currentclass )) {

            symbolTable.put( id , symbolTable );

        } else {
            symbolTable.put( id , (SymbolTable)knownTypes.get( type ) );
        }

    }

    public void caseAArrayDecDeclaration(AArrayDecDeclaration node)
    {
        ln = node.getIdentifier().getLine();

        if("".equals(currentfunc) && "".equals(currentclass)){
            giveError("Global statements are not allowed.",ln);
        }

        arraydim = node.getBracketPair().size();

        String type = node.getType().toString().trim();

        if ( !knownTypes.containsKey( type ) ) {
            giveError( "The type " + type + " has not been defined." , ln );
            return;
        }

        String id = node.getIdentifier().toString().trim();

        if ( symbolTable.containsKey( id ) ) {
            giveError("A variable with the name " + id + " has already been declared." , ln );
            return;
        }

        IMTrace.out("Array declaration:- " + id);
        IMTrace.out("Dimensions == " + arraydim);

        Node init = node.getArrayInitialiser();

        if ( init != null )
        {
            expected = type;
            init.apply( this );
            // found will be contain the type of the initialiser
            if(!checked) {
                giveError("Variable " + id + " must be initialised with type " +
                  typeDimErrorString(type,arraydim), ln );
            }

            if ( maxdepth != arraydim) {
                giveError( "Array declared with " + arraydim + " dimensions. Initialiser has " + maxdepth + " dimensions." , ln );
            }

            maxdepth = curdepth = 0;
        }

        if ( type.equals( currentclass ))
        {
            SymbolTable st = (SymbolTable)symbolTable.clone();
            st.arraydim = arraydim;
            st.type = currentclass;
            st.isClass = true;
            st.put( "length" , (SymbolTable)knownTypes.get( "int" ) );
            symbolTable.put( id , st );

        } else {
            SymbolTable st = (SymbolTable)((SymbolTable)knownTypes.get( type )).clone();
            st.arraydim = arraydim;
            st.type = type;
            st.isClass = true;
            st.put( "length" , (SymbolTable)knownTypes.get( "int" ) );
            symbolTable.put( id , st );
        }

    }

    public void caseATypeParamList( ATypeParamList node ) {

    if ( ! knownTypes.containsKey( node.getType().toString().trim() ) ) {
       giveError("Unknown type given as type parameter" , ln );
    }

    if ( node.getCommaType() != null ) {
        LinkedList ctl = node.getCommaType();
        ListIterator ctlIterator = ctl.listIterator(0);
        while( ctlIterator.hasNext() )
        {
        ACommaType ct = (ACommaType)ctlIterator.next();

        if ( ! knownTypes.containsKey( ct.getType().toString().trim() ) ) {
            giveError("Unknown type given as type parameter" , ln );
        }
        }
    }
    }

    public void caseAArrayInit( AArrayInit node ) {

    if ( ++curdepth > maxdepth )
        maxdepth = curdepth;

    if ( node.getInitList() != null ) {
        node.getInitList().apply( this );
    }

    --curdepth;
    }

    public void caseAInitialiser( AInitialiser node )
    {
        node.getExpression().apply ( this );
    }

    public StringPair separateOnDot( String from )
    {
        int p = from.indexOf('.');
        StringPair pair = new StringPair();

        if ( p < 0 ) {
            pair.a = pair.b = from;
            pair.separated = false;
        } else {
            pair.a = from.substring( 0 , p ).trim();
            pair.b = from.substring( p+1 , from.length() ).trim();
            pair.separated = true;
        }
    return pair;
    }

    private String nameFromFA( PFieldAccess node )
    {
        if ( node instanceof AScalarFieldAccess ) {
            return ((AScalarFieldAccess)node).getName().toString().trim();
        } else {
            return ((AArrayFieldAccess)node).getName().toString().trim();
        }
    }

    private String uptolbracket( String a ) {

	if ( a.indexOf('[') < 0 ) { return a.trim(); }
	else { return a.substring( 0 , a.indexOf('[') ).trim(); }
    }

    private String uptorbracket( String a ) {

	if ( a.indexOf(']') < 0 ) { return a.trim(); }
	else { return a.substring( 0 , a.indexOf(']') ).trim(); }
    }

    //***********For loops*****************

    public void caseAJavaForControl( AJavaForControl node ) {

        ln = node.getS1().getLine();

        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }

        if ( node.getExpression() != null ) {
            expected = "double";
            node.getExpression().apply(this);
            if ( checked == false ) { giveError("For loop starting index must be numeric" , ln ); return; }
        }

        if ( node.getBoolExpression() != null ) {
            expected = "boolean";
            node.getBoolExpression().apply(this);
            if ( checked == false ) { giveError("For loop termination condition must be boolean" , ln ); return; }
        }
    }

    public void caseASimpleForControl( ASimpleForControl node ) {

        ln = node.getTo().getLine();

        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }

        if ( node.getExp1() != null ) {
            expected = "double";
            node.getExp1().apply(this);
            if ( checked == false ) { giveError("For loop starting index must be numeric" , ln ); return; }
        }
        if ( node.getExp2() != null ) {
            expected = "double";
            node.getExp2().apply(this);
            if ( checked == false ) { giveError("For loop ending index must be numeric" , ln ); return; }
        }

        if ( node.getStepBit() != null ) {
            AStepBit sb = (AStepBit)node.getStepBit();
            expected = "double";
            sb.getExpression().apply(this);
            if ( checked == false ) { giveError("For loop step must be numeric" , ln ); return; }
        }
    }

    public void caseABreakStatement(ABreakStatement node)
    {
        ln = node.getBreak().getLine();
        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }
    }

    public void caseAWhileStatement( AWhileStatement node )
    {
        ln = node.getWhile().getLine();

        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }

        if ( node.getBoolExpression() != null ) {
            expected = "boolean";
            node.getBoolExpression().apply( this );

            if ( checked == false ) {
                giveError("While condition must be boolean" , ln );
                return;
            }
        }

        if ( node.getBlock() != null ) {

            expected = "";
            node.getBlock().apply( this );
        }
    }

    public void caseASwitchStatement( ASwitchStatement node ) {

        ln = node.getSwitch().getLine();
        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }

        if ( node.getExpression() != null ) {

            expected = "int";
            node.getExpression().apply(this);

            if ( checked == false ) {

                checked = true;
                expected = "char";
                node.getExpression().apply( this );

                if ( checked == false ) {
                    ln = node.getSwitch().getLine();
                    giveError("You can only switch on an integer or a char." , ln );
                    return;
                }
            }

            expected = "";

            if ( node.getSwitchBlock() != null ) {
                node.getSwitchBlock().apply( this );
            }
        }
    }

    public void caseAFunctioncallStatement(AFunctioncallStatement node)
    {
        ln = node.getSemicolon().getLine();
        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }
        node.getFunctionApplication().apply(this);
    }

    public void caseAAssignmentStatement( AAssignmentStatement node )
    {
	curdepth = maxdepth = 0;

        ln = node.getAssign().getLine();
        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }

        PFieldAccess fieldaccess = node.getFieldAccess();
        IMTrace.out("caseAAssignmentStatement() fieldaccess = " + fieldaccess);

        fieldaccess.apply(this);

        // get the dimesions of LHS of assignment, which would have been
        // evaluated and stored in founddim by fieldaccess.apply(this);
        int LHSDim = founddim;

        String name = fieldaccess.toString();
        SymbolTable s = symbolTable;
        StringPair ab = separateOnDot( name );

	//*****************
	expected = found;

	node.getExpression().apply( this );

	if ( LHSDim != founddim ) {
	    checked = false;
	}

        if (!checked ){
            giveError("Error in expression assigned to " + name + " - " +
              formattedDimErrorString(expected,found,LHSDim,founddim),ln);
        }

        founddim = 0;
    }

    public void caseAPrintStrStatement( APrintStrStatement node ) {

        ln = node.getPrint().getLine();

        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }
        if ( node.getExpression() != null ) {

            expected = "";
            node.getExpression().apply( this );
        }
    }

    public void caseAPrintlnStrStatement( APrintlnStrStatement node ) {

        ln = node.getPrintln().getLine();

        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }
        if ( node.getExpression() != null ) {
            expected = "";
            node.getExpression().apply( this );
        }
    }


    // ************* Graphics Primitives ***************** //

    public void caseAClearStatement(AClearStatement node)
    {
        ln = node.getClear().getLine();
        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }
    }

    public void caseADrawlineStatement(ADrawlineStatement node)
    {
        ln = node.getDrawline().getLine();
        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }
    }

    public void caseADrawtriangleStatement(ADrawtriangleStatement node)
    {
        ln = node.getDrawtriangle().getLine();
        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }
    }

    public void caseADrawellipseStatement(ADrawellipseStatement node)
    {
        ln = node.getDrawellipse().getLine();
        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }
    }

    public void caseADrawstringStatement( ADrawstringStatement node ) {


    ln = node.getDrawstring().getLine();
        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }

    expected = "String";
    node.getString().apply( this );
    if ( !checked ) {
        giveError("First argument to drawString must be a String" , ln );
        return;
    }

    expected = "int";
    node.getX().apply( this );
    if ( !checked ) {
        giveError("Second argument to drawString (x co-ord) must be an int" , ln );
        return;
    }

    expected = "int";
    node.getY().apply( this );
    if ( !checked ) {
        giveError("Third argument to drawString (y co-ord) must be an int" , ln );
        return;
    }
    }

    public void caseASetcolourStatement(ASetcolourStatement node)
    {
        ln = node.getSetcolour().getLine();
        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }
    }

    public void caseASetbgcolourStatement(ASetbgcolourStatement node)
    {
        ln = node.getSetbgcolour().getLine();
        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }
    }

    public void caseAFillStatement( AFillStatement node )
    {
        ln = node.getFill().getLine();
        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }
        expected = "boolean";
        node.getBoolExpression().apply( this );
        if ( ! checked ) {
            giveError("Fill takes a boolean parameter." , ln );
        }

    }

    public void caseAFourNumbers( AFourNumbers node ) {

        ln = node.getC1().getLine();

    if ( node.getX1() != null ) {

        expected = "int";
        node.getX1().apply( this );

        if ( !checked ) {
        giveError("First parameter should have type int" , ln );
        }
    }

    if ( node.getY1() != null ) {

        expected = "int";
        node.getY1().apply( this );

        if ( !checked ) {
        giveError("Second parameter should have type int" , ln );
        }
    }

    if ( node.getX2() != null ) {

        expected = "int";
        node.getX2().apply( this );

        if ( !checked ) {
        giveError("Third parameter should have type int" , ln );
        }
    }

    if ( node.getY2() != null ) {

        expected = "int";
        node.getY2().apply( this );

        if ( !checked ) {
        giveError("Fourth parameter should have type int" , ln );
        }
    }
    }

    public void caseASixNumbers ( ASixNumbers node ) {

    if ( node.getFourNumbers() != null ) {
        node.getFourNumbers().apply( this );
    }

    if ( node.getX() != null ) {

        expected = "int";
        node.getX().apply( this );

        if ( !checked ) {
        giveError("Fifth parameter should have type int" , ln );
        }
    }

    if ( node.getY() != null ) {

        expected = "int";
        node.getY().apply( this );

        if ( !checked ) {
        giveError("Sixth parameter should have type int" , ln );
        }
    }
    }

    public void caseARgb( ARgb node ) {

    ln = node.getC1().getLine();

    if ( node.getR() != null ) {

        expected = "int";
        node.getR().apply( this );

        if ( !checked ) {
        giveError("First (red) parameter should have type int" , ln );
        }
    }

    if ( node.getG() != null ) {

        expected = "int";
        node.getG().apply( this );

        if ( !checked ) {
        giveError("Second (green) parameter should have type int" , ln );
        }
    }

    if ( node.getB() != null ) {

        expected = "int";
        node.getB().apply( this );

        if ( !checked ) {
        giveError("Third (blue) parameter should have type int" , ln );
        }
    }
    }


    //********** Expression terms and factors ******************//

    public void caseAArrayAllocate(AArrayAllocate node)
    {
        maxdepth = node.getArrayAccess().size() +
          node.getBracketPair().size();
        founddim = maxdepth;
        found = node.getType().toString().trim();

        IMTrace.out("caseAArrayAllocate()::founddim = " + founddim +
          " type = " + found);

        checked &= expected.equals(found);
    }


    public void caseAPlusMathExpression(APlusMathExpression node)
    {
        checkMathExpression(node.getMathExpression(),node.getTerm(),
          "+",node.getPlus().getLine(),node);
    }


    public void caseAMinusMathExpression(AMinusMathExpression node)
    {
        checkMathExpression(node.getMathExpression(),node.getTerm(),
          "-",node.getMinus().getLine(),node);
    }

    // Dimension check on + and -.
    // i.e. check that we are not attempting to use one of those
    // operators on an array.
    private void checkMathExpression(PMathExpression LHS, PTerm RHS,
      String op, int line, PMathExpression node)
    {
        int LHSDim, RHSDim;
        String LHSType, RHSType;
        RHS.apply(this);
        RHSDim = founddim;
        RHSType = found;
        LHS.apply(this);
        LHSDim = founddim;
        LHSType = found;

        if(LHSDim!=RHSDim || LHSDim!=0){
            founddim = 0;
            giveError("operator " + op + " cannot be applied to " +
              typeDimErrorString(LHSType,LHSDim) + "," +
              typeDimErrorString(RHSType,RHSDim) + ": " +
              node.toString(),
              line);
        }
    }

    public void caseAMultTerm( AMultTerm node )
    {
       checkArithTerm(node.getTerm(),node.getUnaryExp(),"*",
         node.getTimes().getLine(),node);
    }

    public void caseADivTerm( ADivTerm node )
    {
       checkArithTerm(node.getTerm(),node.getUnaryExp(),"/",
         node.getDivide().getLine(),node);
    }

    public void caseAPowerTerm( APowerTerm node )
    {
       int LHSDim = 0, RHSDim = 0;
       String LHSType = "", RHSType = "";

       checked &= ( expected.equals("double") || expected.equals("") );
       if ( ! checked ) { return; }
       expected = "double";

       checkArithTerm(node.getTerm(),node.getUnaryExp(),"^",
         node.getPower().getLine(),node);


       found = "double";
    }

    // Dimension check on *, /, ^ and %.
    // i.e. check that we are not attempting to use one of those
    // operators on an array.
    private void checkArithTerm(PTerm LHS, PUnaryExp RHS,
      String op, int line, PTerm node)
    {
       int LHSDim, RHSDim;
       String LHSType, RHSType;

       LHS.apply(this);
       LHSDim = founddim;
       LHSType = found;

       RHS.apply(this);
       RHSDim = founddim;
       RHSType = found;

       if(LHSDim!=RHSDim || LHSDim!=0)
       {
            founddim = 0;
            giveError("operator " + op + " cannot be applied to " +
              typeDimErrorString(LHSType,LHSDim) + "," +
              typeDimErrorString(RHSType,RHSDim) + ": " +
              node.toString(),
              line);
       }
    }

    public void caseAMinusUnaryExp(AMinusUnaryExp node)
    {
        node.getUnaryExp().apply(this);
        if(founddim>0){
            giveError("operator - cannot be applied to " +
              typeDimErrorString(found,founddim) + ": " +
              node.toString(),
              node.getMinus().getLine());
       }else if(founddim <0){
            giveError("operator - usage error, array " +
              "access has too many dimensions: " +
              node.toString(),
              node.getMinus().getLine());
       }
    }

    public void caseAPostdecr(APostdecr node)
    {
        checkPlusPlusOrMinusMinus(node.getPostfixExp(),"--",
          node.getMinusminus().getLine(),node);
    }

    public void caseAPredecr(APredecr node)
    {
        checkPlusPlusOrMinusMinus(node.getPostfixExp(),"--",
          node.getMinusminus().getLine(),node);
    }

    public void caseAPostincr(APostincr node)
    {
        checkPlusPlusOrMinusMinus(node.getPostfixExp(),"++",
          node.getPlusplus().getLine(),node);
    }

    public void caseAPreincr(APreincr node)
    {
        checkPlusPlusOrMinusMinus(node.getPostfixExp(),"++",
          node.getPlusplus().getLine(),node);
    }

    // Dimension check on ++ and --.
    // i.e. check that we are not attempting to use one of those
    // operators on an array.
    private void checkPlusPlusOrMinusMinus(PPostfixExp exp, String op,
      int line, Node node)
    {
        exp.apply(this);

        if(founddim>0){
            giveError("operator " + op + " cannot be applied to " +
              typeDimErrorString(found,founddim) + ": " +
              node.toString(),
              line);
        }else if(founddim <0){
            giveError("operator " + op + " usage error, array " +
              "access has too many dimensions: " +
              node.toString(),
              line);
        }
    }

    public void caseAFunctionApplication( AFunctionApplication node )
    {
        String name = node.getName().toString().trim();
        ln = node.getLParenthese().getLine();

        if ( !symbolTable.containsKey( name ) ) {
            giveError( "Function " + name + " has not been defined." , ln );
            return;
        }

        IMTrace.out("Calling function ------> " + name);
        SymbolTable s = (SymbolTable)symbolTable.get( name );

        Vector outerParams = null;
        if(mParamsFound!=null) outerParams = mParamsFound;

        mParamsFound = new Vector();

        // the recursive calls to check parameters will alter this
        // value, must save it.
        boolean savedCheck = checked;

        node.getActualParamList().apply(this);

        FuncInfo f = s.funcInfo;

        IMTrace.out("f.numParams() = " + f.numParams());
        IMTrace.out("mParamsFound.size() = " + mParamsFound.size());

        checked = savedCheck;

        if(f.numParams()!=mParamsFound.size()){ // check the number of params
            giveError("Wrong number of parameters in function call.",ln);
        }

        // param type check
        int i = 0;
        for (Enumeration e = mParamsFound.elements(); e.hasMoreElements();i++)
        {
            String expectedParam = f.getParamType(i);
            String foundParam = (String)e.nextElement();

            IMTrace.out("Expected param " + i + " = " + expectedParam);
            IMTrace.out("Found param " + i + " = " + foundParam);

            if(expectedParam.equals("double")){
                checked &=
                  (foundParam.equals("int")||foundParam.equals("double"));
            } else if(expectedParam.equals("int")){
                checked &=
                  (foundParam.equals( "int" )||foundParam.equals("char"));
            } else{
                checked &=
                  (expectedParam.equals( foundParam )||expectedParam.equals(""));
            }

            if( /*!foundParam.equals(expectedParam)*/!checked){
                giveError("function call: " + name + ": param " + (i+1) +
                  " is wrong type." + " Expected: " + expectedParam + ", found: " +
                  foundParam,
                  node.getLParenthese().getLine());
            }

        }

        if(outerParams!=null) mParamsFound = outerParams;
        else mParamsFound = null;

        String type = s.type;
        IMTrace.out("\n\nFunction : "+name);
        IMTrace.out("Expected  : "+expected + "|");
        IMTrace.out("Return type ---> " + type + "|");

        if ( expected.equals( "double" ) ) {
            checked &= ( type.equals("int") || type.equals("double") );
        } else if ( expected.equals( "int" ) ) {
            checked &= ( type.equals( "int" ) || type.equals("char") );
        } else {
            checked &= ( expected.equals( type ) || expected.equals("") );
        }

        IMTrace.out("Checked : "+checked);
        IMTrace.out("Type : "+type);
        IMTrace.out("arraydim : " + s.arraydim);

        found = type;
        founddim = s.arraydim;
    }

    public void caseAListActualParamList(AListActualParamList node)
    {
        node.getActualParamList().apply(this);
        node.getExpression().apply(this);

        mParamsFound.add(typeDimErrorString(found,founddim));
    }

    public void caseAExpActualParamList(AExpActualParamList node)
    {
        node.getExpression().apply(this);
        mParamsFound.add(typeDimErrorString(found,founddim));
    }



    public void caseAFNumber( AFNumber node ) {

        checked &= ( expected.equals( "double" ) || expected.equals("") );
        found = "double";
        // >> askc98 2002/5/18
        founddim = 0;
        // <<< askc98
    }

    public void caseAINumber( AINumber node ) {

        checked &= ( expected.equals( "int" ) || expected.equals( "double") || expected.equals("") );
        found = "int";
        // >> askc98 2002/5/18
        founddim = 0;
        // <<< askc98
    }

    public void caseAStringliteralFactor( AStringliteralFactor node ) {

    checked &= ( expected.equals( "String" )  || expected.equals("") );
    found = "String";
    }

    public void caseACharliteralFactor( ACharliteralFactor node ) {

    checked &= ( expected.equals( "char" ) || expected.equals("int") || expected.equals("") );
    found = "char";
    }

    public void caseABoolliteralFactor( ABoolliteralFactor node ) {

    checked &= ( expected.equals( "boolean" ) || expected.equals("") );
    found = "boolean";
    }

    public void caseAArrayAccess( AArrayAccess node )
    {
       ln = node.getLBracket().getLine();

       if ( expectingArrayIndex && node.getMathExpression() == null ) {
           giveError("Missing array index" , ln );
           return;
       }

       if ( ! expectingArrayIndex && node.getMathExpression() != null ) {
           giveError("Not expecting array index." , ln );
           return;
       }

       if ( ! expectingArrayIndex ) return;

       String stackedexpected = expected;
       //   String stackedfound = found;

       if ( node.getMathExpression() != null ) {
           expected = "int";
           node.getMathExpression().apply(this);
           if ( !checked ) {
               giveError("Array index must be integer" , ln = node.getLBracket().getLine() );
           } 

       } else {
           giveError("Missing array index" , ln = node.getLBracket().getLine() );
       }
       
       expected = stackedexpected;

    }

    public void caseAArrayFieldAccess(AArrayFieldAccess node)
    {
        checkArrayFieldAccess(node);
    }

    public void caseAScalarFieldAccess(AScalarFieldAccess node)
    {
        checkFieldAccess(node);
    }

    private void checkFieldAccess(PFieldAccess node)
    {
        String name = node.toString();
        SymbolTable s = symbolTable;
        StringPair ab = separateOnDot( name );
        String arrname = uptolbracket( ab.a );

        while ( ab.separated )
        {
	    if ( ! s.containsKey( arrname ) ) {
                giveError("Variable " + arrname + " has not been declared" , ln );
                return;

            } else {
                s = ((SymbolTable)s.get( arrname ));
                if ( ! s.isClass ) {
                    giveError( ab.a + " is not a class." , ln );
                    return;
                }
            }

            ab = separateOnDot( ab.b );
            arrname = uptolbracket( ab.a );
        }
        // on exit from loop, ab.a = ab.b = arrname
       IMTrace.out("checkFieldAccess() arrname = " + arrname);

        founddim = 0;
        int nextBracket = 0;

        while((nextBracket = ab.b.indexOf('[',nextBracket+1))!=-1)
            founddim++;

        if ( ! s.containsKey(arrname ) ) {

            giveError("Variable " + ab.b + " has not been declared." , ln );
      
	} else {

            s = (SymbolTable)s.get(arrname);
            String type = s.type;

            IMTrace.out("checkFieldAccess() founddim before mod  = " + founddim);
            IMTrace.out("checkFieldAccess() ss.arraydim = " + s.arraydim);
            // >>> askc98 2002/5/16

            founddim = s.arraydim - founddim;

            // <<< askc98
            IMTrace.out("checkFieldAccess() founddim after mod  = " + founddim);

            // tried to access a non array variable like an array
            if(founddim < 0){
                giveError("array expected, but " + type + " found. -> "
                  + name, ln);
            }

            IMTrace.out("checkFieldAccess()::expected ---> " + expected);

            if ( type != null ) {
                if ( expected.equals( "double" ) ) {

                    checked &= ( type.equals("int") || type.equals("double") );

                } else if ( expected.equals("int") ) {

                    checked &= ( type.equals("int") || type.equals("char") );

                } else {
                    checked &= ( expected.equals( type ) || expected.equals("") );
                }

                found = type;

            } else { System.err.println("type in symbol table for " + name + " is null"); }
        }
    }

    private void checkArrayFieldAccess( AArrayFieldAccess node ) {
	
	String name = node.getName().toString().trim();
	StringPair ab = separateOnDot( name );

	SymbolTable s = symbolTable;

	while ( ab.separated ) {

	    String key = uptolbracket( ab.a );

	    if ( ! s.containsKey( key ) ) {
		giveError("Variable " + key + " has not been declared" , ln );
                return;
	    } else {
		s = (SymbolTable)s.get( key );
		if ( ! s.isClass ) {
                    giveError( key + " is not a class." , ln );
                    return;
                }
	    }

	    ab = separateOnDot( ab.b );
	}

	String key = uptolbracket( ab.a );

	if ( ! s.containsKey( key ) ) {
	    giveError("Variable " + key + " has not been declared" , ln );
	    return;
	}

	s = (SymbolTable)s.get( key );

	if ( node.getArrayAccess().size() != s.arraydim ) {
	    giveError("Array " + key + " was declared with " + s.arraydim + " dimensions, but accessed with " + node.getArrayAccess().size() , ln );
	    return;
	}

	String stackedexpected = expected;

	for ( Iterator i = node.getArrayAccess().iterator() ; i.hasNext() ; ) {
	   ((Node)i.next()).apply(this);
	}
	
	expected = stackedexpected;
	found = s.type;
    }

    public void caseAReadFactor( AReadFactor node ) {

    checked &= ( expected.equals( "char" ) || expected.equals( "int" ) || expected.equals("") );
    found = "char";
    }

    public void caseAReadintFactor( AReadintFactor node ) {

    checked &= ( expected.equals( "int" )  || expected.equals("") );
    found = "int";
    }

    public void caseAReadrealFactor( AReadrealFactor node ) {

    checked &= ( expected.equals( "double" ) || expected.equals("") );
    found = "double";
    }

    public void caseAReadstringFactor( AReadstringFactor node ) {

    checked &= ( expected.equals( "String" ) || expected.equals("") );
    found = "String";
    }

    public void caseAReadcharFactor( AReadcharFactor node ) {

    checked &= ( expected.equals( "char" ) || expected.equals( "int" ) || expected.equals("") );
    found = "char";
    }

    public void caseASinFactor( ASinFactor node ) {

    ln = node.getSin().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) { return; }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) { giveError( "Argument to sin must be numeric" , ln ); }
    found = "double";
    }

    public void caseAAsinFactor( AAsinFactor node ) {

    ln = node.getAsin().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) { return; }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) { giveError( "Argument to asin must be numeric" , ln ); }
    found = "double";
    }

    public void caseACosFactor( ACosFactor node ) {
    ln = node.getCos().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) { return; }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) { giveError( "Argument to cos must be numeric" , ln ); }
    found = "double";
    }

    public void caseAAcosFactor( AAcosFactor node ) {

    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) { return; }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) { giveError( "Argument to acos must be numeric" , node.getAcos().getLine() ); }
    found = "double";
    }

    public void caseATanFactor( ATanFactor node ) {

    ln = node.getTan().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) { return; }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) { giveError( "Argument to tan must be numeric" , ln ); }
    found = "double";
    }

    public void caseAAtanFactor( AAtanFactor node ) {

    ln = node.getAtan().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) { return; }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) { giveError( "Argument to atan must be numeric" , ln ); }
    found = "double";
    }

    public void caseASqrtFactor( ASqrtFactor node ) {

    ln =  node.getSqrt().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) { return; }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) { giveError( "Argument to sqrt must be numeric" , ln ); }
    found = "double";
    }

    public void caseALogFactor( ALogFactor node ) {

    ln =  node.getLog().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) { return; }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) { giveError( "Argument to log must be numeric" , ln ); }
    found = "double";
    }

    public void caseAExpFactor( AExpFactor node ) {

    ln =  node.getExp().getLine();
    checked &= ( expected.equals( "double") || expected.equals("") );
    if ( !checked ) { return; }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) { giveError( "Argument to exp must be numeric" , ln ); }
    found = "double";
    }

    public void caseARoundFactor( ARoundFactor node ) {

    ln = node.getRound().getLine();
    checked &= ( expected.equals( "int" ) || expected.equals("") );
    if ( !checked ) { return; }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) { giveError( "Argument to round must be numeric" , ln ); }
    found = "int";
    }

    public void caseAFloorFactor( AFloorFactor node ) {

    ln = node.getFloor().getLine();
    checked &= ( expected.equals( "int" ) || expected.equals( "double" ) || expected.equals("") );
    if ( !checked ) { return; }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) { giveError( "Argument to floor must be numeric" , ln ); }
    found = "int";
    }

    public void caseARandomFactor( ARandomFactor node ) {

    checked &= ( expected.equals( "double" ) || expected.equals("") );
    found = "double";
    }

    public void caseACeilFactor( ACeilFactor node ) {

    ln =  node.getCeil().getLine();
    checked &= ( expected.equals( "int" ) || expected.equals( "double" ) || expected.equals("") );
    if ( !checked ) { return; }
    expected = "double";
    node.getMathExpression().apply( this );
    if ( !checked ) { giveError( "Argument to ceil must be numeric" , ln ); }
    found = "double";
    }

    public void caseAFieldaccessFactor( AFieldaccessFactor node )
    {
        node.getFieldAccess().apply(this);
    }

    public void caseANullFactor( ANullFactor node ) {

    //matches any type

    found = expected;
    }

    /***** relational expressions *****/

    public void caseALtRelational( ALtRelational node ) {

       checked &= expected.equals("boolean");
       if ( !checked ) {
           found = "boolean";
           return;
       }

       expected = "";

       ln = node.getLess().getLine();

       if ( node.getRelational() != null ) {
           node.getRelational().apply( this );
       }

       String lhs_type = expected = found;

       if ( node.getMathExpression() != null ) {
           node.getMathExpression().apply( this );
       }

       if ( !checked ) {
           giveError( "Cannot compare " + lhs_type + " to " + found + " with < " , ln );
       }
    }

    public void caseAGtRelational( AGtRelational node ) {

    checked &= expected.equals("boolean");
    if ( !checked ) {
        found = "boolean";
        return;
    }

    expected = "";

    ln = node.getGreater().getLine();

    if ( node.getRelational() != null ) {
        node.getRelational().apply( this );
    }

    String lhs_type = expected = found;

    if ( node.getMathExpression() != null ) {
        node.getMathExpression().apply( this );
    }

    if ( !checked ) {
        giveError( "Cannot compare " + lhs_type + " to " + found + " with > " , ln );
    }
    }

    public void caseALteqRelational( ALteqRelational node ) {

    checked &= expected.equals("boolean");
    if ( !checked ) {
        found = "boolean";
        return;
    }

    expected = "";

    ln = node.getLessequal().getLine();

    if ( node.getRelational() != null ) {
        node.getRelational().apply( this );
    }

    String lhs_type = expected = found;

    if ( node.getMathExpression() != null ) {
        node.getMathExpression().apply( this );
    }

    if ( !checked ) {
        giveError( "Cannot compare " + lhs_type + " to " + found + " with <= " , ln );
    }
    }

    public void caseAGteqRelational( AGteqRelational node ) {

    checked &= expected.equals("boolean");
    if ( !checked ) {
        found = "boolean";
        return;
    }

    expected = "";

    ln = node.getGreaterequal().getLine();

    if ( node.getRelational() != null ) {
        node.getRelational().apply( this );
    }

    String lhs_type = expected = found;

    if ( node.getMathExpression() != null ) {
        node.getMathExpression().apply( this );
    }

    if ( !checked ) {
        giveError( "Cannot compare " + lhs_type + " to " + found + " with >= " , ln );
    }
    }

    public void caseAEqEquality( AEqEquality node ) {

    checked &= expected.equals("boolean");
    if ( !checked ) {
        found = "boolean";
        return;
    }

    expected = "";

    ln = node.getEqual().getLine();

    if ( node.getEquality() != null ) {
        node.getEquality().apply( this );
    }

    String lhs_type = expected = found;

    if ( node.getRelational() != null ) {
        node.getRelational().apply( this );
    }

    if ( !checked ) {
        giveError( "Cannot compare " + lhs_type + " to " + found + " with == " , ln );
    }
    }

    public void caseANeqEquality( ANeqEquality node ) {

    checked &= expected.equals("boolean");
    if ( !checked ) {
        found = "boolean";
        return;
    }

    expected = "";

    ln = node.getNotequal().getLine();

    if ( node.getEquality() != null ) {
        node.getEquality().apply( this );
    }

    String lhs_type = expected = found;

    if ( node.getRelational() != null ) {
        node.getRelational().apply( this );
    }

    if ( !checked ) {
        giveError( "Cannot compare " + lhs_type + " to " + found + " with != " , ln );
    }
    }

    public void caseAAndBoolTerm( AAndBoolTerm node )
    {
    expected = "boolean";

    if ( node.getBoolTerm() != null )
        node.getBoolTerm().apply( this );

    if ( !checked )
        giveError( found + " expression found where boolean expected, before " + node.getAnd().toString() , ln );

    expected = "boolean";

    if ( node.getEquality() != null )
        node.getEquality().apply( this );

    if ( !checked )
        giveError( found + " expression found where boolean expected, after " + node.getAnd().toString() , ln );
    }

    public void caseAOrBoolExpression( AOrBoolExpression node )
    {
    expected = "boolean";

    if ( node.getBoolExpression() != null )
        node.getBoolExpression().apply( this );

    if ( !checked )
        giveError( found + " expression found where boolean expected, after " + node.getOr().toString() , ln );

    expected = "boolean";

    if ( node.getBoolTerm() != null )
        node.getBoolTerm().apply( this );

    if ( !checked )
        giveError( found + " expression found where boolean expected, after " + node.getOr().toString() , ln );
    }

    public void caseAXorBoolExpression( AXorBoolExpression node )
    {
    expected = "boolean";

    if ( node.getBoolExpression() != null )
        node.getBoolExpression().apply( this );

    if ( !checked )
        giveError( found + " expression found where boolean expected, after " + node.getXor().toString() , ln );

    expected = "boolean";

    if ( node.getBoolTerm() != null )
        node.getBoolTerm().apply( this );

    if ( !checked )
        giveError( found + " expression found where boolean expected, after " + node.getXor().toString() , ln );
    }

    /******************* Other statments ******8************/

    public void caseAAssertStatement( AAssertStatement node )
    {
        ln = node.getAssert().getLine();

        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }

        if ( node.getBoolExpression() != null ) {
            expected = "boolean";
            node.getBoolExpression().apply( this );
            if ( ! checked ) {
                giveError( "Can only assert boolean conditions" , ln );
                return;
            }
        }

        if ( node.getColonString() != null ) {
            node.getColonString().apply( this );
        }
    }

    public void caseAColonString( AColonString node )
    {
    expected = "String";

    if ( node.getExpression() != null )
        node.getExpression().apply( this );

    if ( !checked )
        giveError( "Assertion message must be a String." , ln );
    }

    public void caseAArraycopyStatement( AArraycopyStatement node )
    {
        ln = node.getArraycopy().getLine();

        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }

        String src  = node.getSrc().toString().trim();
        String dest = node.getDest().toString().trim();

        SymbolTable srcST = checkQualifiedName(src);
        SymbolTable destST = checkQualifiedName(dest);

        if(srcST==null || destST==null) {
             //System.err.println("**** one of the operands not an array");
             //giveError("one or both operands is not an array.",ln);
             return;
        }
        int indexOfLastDot = src.lastIndexOf('.');
        if(indexOfLastDot!=-1) src = src.substring(indexOfLastDot);
        indexOfLastDot = dest.lastIndexOf('.');
        if(indexOfLastDot!=-1) dest = dest.substring(indexOfLastDot);

        IMTrace.out("caseAArrayCopyStatement dest type = " + destST.type);
        IMTrace.out("caseAArrayCopyStatement src type = " + srcST.type);

        if(!srcST.containsKey(src)){
            giveError("Variable " + src + " has not been declared" , ln );
        }else{
            srcST = (SymbolTable)srcST.get(src);
        }

        if(!destST.containsKey(src)){
            giveError("Variable " + dest + " has not been declared" , ln );
            return;
        }else{
            destST = (SymbolTable)destST.get(dest);
        }

        if(!srcST.type.equals(destST.type))
        {
            giveError( "arraycopy can only be applied to two arrays of the same type",ln );
        }

        if(srcST.arraydim<=0 || destST.arraydim <=0)
        {
            giveError("both of the operands to arraycopy must be arrays.",ln);
        }
        else if(srcST.arraydim!=destST.arraydim)
        {
            giveError( "arraycopy can only be applied to two arrays of the same dimensions",ln );
        }
        /*expected = "";
        node.getSrc().apply( this );

        expected = found;

        node.getDest().apply( this );

        if( !checked ) {
            giveError( "arraycopy can only be applied to two arrays of the same type" );
        }*/
    }

    public void caseAIfStatement( AIfStatement node )
    {
        ln = node.getIf().getLine();

        if("".equals(currentfunc)){
            giveError("Global statements are not allowed.",ln);
        }

        if ( node.getBoolExpression() != null ) {
            expected = "boolean";
            node.getBoolExpression().apply( this );
            if ( !checked ) {
                giveError("Condition must be boolean" , ln );
                return;
            }
        }

        if ( node.getBlock1() != null )
            node.getBlock1().apply( this );

        if ( node.getBlock2() != null )
            node.getBlock2().apply( this );
    }

    public void giveError( String s , int line )
    {
        mErrorWriter.giveError( "Error (line " + line + "): " + s + "\n" ,line);
        genCode = false;
        checked = true;
    }

    public void giveError( String s )
    {
        mErrorWriter.giveError("Error : " + s + "\n");
        genCode = false;
        checked = true;
    }

    private String typeDimErrorString(String type, int dim)
    {
        StringBuffer errorMsg = new StringBuffer(type);
        for(int i=0;i<dim;i++) errorMsg.append("[ ]");
        return errorMsg.toString();
    }

    private String formattedDimErrorString(String reqType,String foundType,
      int required,int found)
    {
        StringBuffer errorMsg = new StringBuffer("required: ");
        errorMsg.append(typeDimErrorString(reqType,required));
        errorMsg.append("    found: ").append(typeDimErrorString(foundType,found));

        return errorMsg.toString();
    }

    /**
     * utility method for returning the line number of a field access. USed
     * for producing better error messages usually.
     */
    private int getLineNum(PFieldAccess node)
    {
        if(node instanceof AArrayFieldAccess)
        {
            return getLineNumFromName(((AArrayFieldAccess)node).getName());
        }else{
            return getLineNumFromName(((AScalarFieldAccess)node).getName());
        }
    }

    /**
     * utility method for returning the line number of a field access. USed
     * for producing better error messages usually.
     */
    private int getLineNumFromName(PName name)
    {
        if(name instanceof ASimpleNameName)
        {
            PSimpleName p = ((ASimpleNameName)name).getSimpleName();
            return ((ASimpleName) p).getIdentifier().getLine();
        }else{
            PQualifiedName ka = ((AQualifiedNameName)name).getQualifiedName();
            return ((AQualifiedName)ka).getIdentifier().getLine();
        }
    }

    private SymbolTable checkQualifiedName(String name)
    {
        SymbolTable s = symbolTable;
        StringPair ab = separateOnDot( name );

        while ( ab.separated )
        {
            String arrname = uptolbracket( ab.a );

            if ( ! s.containsKey( arrname ) ) {
                giveError("Variable " + arrname + " has not been declared" , ln );
                return null;

            } else {

                s = ((SymbolTable)symbolTable.get( arrname ));
                if ( ! s.isClass ) {
                    giveError( ab.a + " is not a class." , ln );
                    return null;
                }
            }

            ab = separateOnDot( ab.b );
        }
        return s;
    }

    public SymbolTable getFunctions()
    {
        return functions;
    }

    public void setFunctions(SymbolTable s)
    {
        functions = s;
    }

    public boolean genCode()
    {
        return genCode;
    }

    /**
     * The typechecker performs two passes to complete type checking, call
     * this method after invoking the typechecker once, and then invoke it
     * again.&nbsp;See the <code>KTypeChecker</code> class for an example of
     * invoking the typechecker.
     */
    public void setSecondPass()
    {
        mFirstPass = false;
    }
}
