/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001,2002 Robert Chatley, Alan Chan                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

/**
 * KOptionsDialog.java
 *
 * Dialog box for setting tab sizes and font size.
 */

package com.chatley.kenya.gui;

import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import java.awt.event.*;
import com.chatley.kenya.utils.IMTrace;
import com.chatley.kenya.gui.syntaxHighlight.*;
import com.chatley.kenya.gui.syntaxHighlight.jeditSyntax.*;

public class KOptionsDialog extends JDialog
{
    private KenyaUI mUi;
    private JPanel mFontSizePanel;
    private JPanel mTabSizePanel;
    private JPanel mKenyaColoursPanel;
    private JPanel mJavaColoursPanel;
    private ButtonGroup mFontBtnGrp;
    private ButtonGroup mTabBtnGrp;

    private JComboBox mKenyaColours;
    private JComboBox mJavaColours;
    private JButton mKenyaColoursBtn;
    private JButton mJavaColoursBtn;
    private JButton mKenyaReset;
    private JButton mJavaReset;

    private SyntaxStyle[] mKenyaStyles;
    private SyntaxStyle[] mJavaStyles;

    private static final Object[] mItems =
      {"Keywords","Strings","Comments"};

    private static final int[] mTokenMap =
      {Token.KEYWORD1, Token.LITERAL1, Token.COMMENT1};

    private static final SyntaxStyle[] mDefaultColours =
      KenyaDefaults.getKenyaSyntaxStyles();


    public KOptionsDialog(KenyaUI ui)
    {
        super((Frame)ui,"Kenya Options",true);
        mUi = ui;
        initComponents();
    }

    private void initComponents()
    {
        mFontSizePanel     = new JPanel();
        mTabSizePanel      = new JPanel();
        mKenyaColoursPanel = new JPanel();
        mJavaColoursPanel  = new JPanel();

        mFontSizePanel.setBorder(
            makeBorder("Font size",mFontSizePanel));

        mTabSizePanel.setBorder(
            makeBorder("Tab size",mTabSizePanel));

        mKenyaColoursPanel.setBorder(
            makeBorder("Kenya text colours",mKenyaColoursPanel));

        mJavaColoursPanel.setBorder(
            makeBorder("Java text colours",mJavaColoursPanel));

        mFontBtnGrp = new ButtonGroup();
        JRadioButton smallFontsBtn = new JRadioButton("Small font");
        JRadioButton largeFontsBtn = new JRadioButton("Large font");
        smallFontsBtn.setActionCommand(String.valueOf(KenyaTextArea.SMALL_FONT_SIZE));
        largeFontsBtn.setActionCommand(String.valueOf(KenyaTextArea.LARGE_FONT_SIZE));

        mFontBtnGrp.add(smallFontsBtn);
        mFontBtnGrp.add(largeFontsBtn);
        mFontSizePanel.add(smallFontsBtn);
        mFontSizePanel.add(largeFontsBtn);


        mTabBtnGrp = new ButtonGroup();
        JRadioButton twoSpaces = new JRadioButton("2 spaces");
        JRadioButton fourSpaces = new JRadioButton("4 spaces");
        JRadioButton eightSpaces = new JRadioButton("8 spaces");
        twoSpaces.setActionCommand(String.valueOf(2));
        fourSpaces.setActionCommand(String.valueOf(4));
        eightSpaces.setActionCommand(String.valueOf(8));

        mTabBtnGrp.add(twoSpaces);
        mTabBtnGrp.add(fourSpaces);
        mTabBtnGrp.add(eightSpaces);
        mTabSizePanel.add(twoSpaces);
        mTabSizePanel.add(fourSpaces);
        mTabSizePanel.add(eightSpaces);

        JPanel btnPanel = new JPanel();
        btnPanel.setLayout(new FlowLayout());
        JButton ok = new JButton("ok");
        JButton cancel = new JButton("Cancel");
        btnPanel.add(ok);
        btnPanel.add(cancel);


        mKenyaColours = new JComboBox(mItems);
        mKenyaColoursPanel.add(mKenyaColours);

        mKenyaColoursBtn = new JButton();
        mKenyaColoursBtn.setPreferredSize(new Dimension(60,25));
        mKenyaColoursPanel.add(mKenyaColoursBtn);

        mKenyaReset = new JButton("Reset");
        mKenyaColoursPanel.add(mKenyaReset);

        mJavaColours = new JComboBox(mItems);
        mJavaColoursPanel.add(mJavaColours);

        mJavaColoursBtn = new JButton();
        mJavaColoursBtn.setPreferredSize(new Dimension(60,25));
        mJavaColoursPanel.add(mJavaColoursBtn);

        mJavaReset = new JButton("Reset");
        mJavaColoursPanel.add(mJavaReset);

        Container contentPane = getContentPane();
        contentPane.setLayout(new BoxLayout(contentPane,BoxLayout.Y_AXIS));
        contentPane.add(mFontSizePanel);
        contentPane.add(mTabSizePanel);
        contentPane.add(mKenyaColoursPanel);
        contentPane.add(mJavaColoursPanel);
        contentPane.add(btnPanel);

        // read in current settings
        int tabSize = mUi.kenyaArea().getTabSize();
        int fontSize = mUi.kenyaArea().getFontSize();

        if(fontSize >= KenyaTextArea.LARGE_FONT_SIZE){
            largeFontsBtn.setSelected(true);
        }else{
            smallFontsBtn.setSelected(true);
        }

        switch(tabSize){
            case 2: twoSpaces.setSelected(true);break;
            case 4: fourSpaces.setSelected(true);break;
            case 8: eightSpaces.setSelected(true);break;
            default: twoSpaces.setSelected(true);break;
        }

        mKenyaStyles =
          new SyntaxStyle[mUi.kenyaArea().getPainter().getStyles().length];
        mJavaStyles  =
          new SyntaxStyle[mUi.javaArea().getPainter().getStyles().length];

        System.arraycopy(mUi.kenyaArea().getPainter().getStyles(),
                         0,mKenyaStyles,
                         0,mKenyaStyles.length);

        System.arraycopy(mUi.javaArea().getPainter().getStyles(),
                         0,mJavaStyles,
                         0,mJavaStyles.length);

        // this complicated expression looks up the colour that is
        // currently selected as the colour to highlight a particular
        // style
        mKenyaColoursBtn.setBackground(mKenyaStyles[
            mTokenMap[mKenyaColours.getSelectedIndex()]].getColor());

        mJavaColoursBtn.setBackground(mJavaStyles[
            mTokenMap[mJavaColours.getSelectedIndex()]].getColor());

        mKenyaColoursBtn.addActionListener(
            new ColourListener(mKenyaColoursBtn,mKenyaColours,mKenyaStyles));

        mJavaColoursBtn.addActionListener(
            new ColourListener(mJavaColoursBtn,mJavaColours,mJavaStyles));

        mKenyaColours.addActionListener(
            new StyleListListener(mKenyaColoursBtn,mKenyaStyles));

        mJavaColours.addActionListener(
            new StyleListListener(mJavaColoursBtn,mJavaStyles));

        mKenyaReset.addActionListener(
            new ResetListener(mKenyaStyles,mKenyaColoursBtn,mKenyaColours));

        mJavaReset.addActionListener(
            new ResetListener(mJavaStyles,mJavaColoursBtn,mJavaColours));

        IMTrace.out("----> Tab size = " + tabSize);
        IMTrace.out("----> Font size = " + fontSize);


        ok.addActionListener( new ActionListener(){
            public void actionPerformed(ActionEvent e)
            {
                String fontsize = mFontBtnGrp.getSelection().getActionCommand();
                String tabsize = mTabBtnGrp.getSelection().getActionCommand();

                IMTrace.out("----> selected tab = " + tabsize);
                IMTrace.out("----> selected font = " + fontsize);

                mUi.kenyaArea().setTabSize(Integer.parseInt(tabsize));
                mUi.kenyaArea().setFontSize(Integer.parseInt(fontsize));
                mUi.javaArea().setTabSize(Integer.parseInt(tabsize));
                mUi.javaArea().setFontSize(Integer.parseInt(fontsize));
                mUi.kenyaArea().getPainter().setStyles(mKenyaStyles);
                mUi.javaArea().getPainter().setStyles(mJavaStyles);

                hide();
                dispose();
            }
        });


        cancel.addActionListener( new ActionListener(){
            public void actionPerformed(ActionEvent e)
            {
                hide();
                dispose();
            }
        });
    }

    private Border makeBorder(String title, JPanel panel)
    {
        Border b =
        BorderFactory.createCompoundBorder(
            BorderFactory.createCompoundBorder(
              BorderFactory.createTitledBorder(title),
              BorderFactory.createEmptyBorder(5,5,5,5)),
              panel.getBorder());
        return b;
    }

    // pops up a JColorChooser Dialog so that the user can
    // choose a colour.
    private class ColourListener implements ActionListener
    {
        JButton parent;
        JComboBox theList;
        SyntaxStyle[] styles;

        public ColourListener(JButton parent, JComboBox theList,
          SyntaxStyle[] styles)
        {
            this.parent = parent;
            this.styles = styles;
            this.theList = theList;
        }

        public void actionPerformed(ActionEvent e)
        {
            Color newColor = JColorChooser.showDialog(
                                        KOptionsDialog.this,
                                        "Choose Text Color",
                                        parent.getBackground());
            if (newColor != null) {
                parent.setBackground(newColor);
                styles[mTokenMap[theList.getSelectedIndex()]] =
                  new SyntaxStyle(newColor,false,false);
            }
        }
    }

    // looks up the colour that is currently selected as the
    // highlight colour for the selected text type in the combo box,
    // and then updates the colour of the button to reflect this
    private class StyleListListener implements ActionListener
    {
        JButton theBtn;
        SyntaxStyle[] styles;

        public StyleListListener(JButton theBtn, SyntaxStyle[] styles)
        {
            this.theBtn = theBtn;
            this.styles = styles;
        }

        public void actionPerformed(ActionEvent e)
        {
            JComboBox cb = (JComboBox)e.getSource();
            int style = cb.getSelectedIndex();
            theBtn.setBackground(styles[
              mTokenMap[style]].getColor());
        }
    }


    // attached to the reset buttons, when clicked, resets the
    // colours for the syntax highlighted to the system defaults
    // and then repaints the colour button.
    private class ResetListener implements ActionListener
    {
        SyntaxStyle[] styles;
        JButton theBtn;
        JComboBox theList;

        public ResetListener(SyntaxStyle[] styles, JButton theBtn,
          JComboBox theList)
        {
            this.styles  = styles;
            this.theBtn  = theBtn;
            this.theList = theList;
        }

        public void actionPerformed(ActionEvent e)
        {
            styles[Token.COMMENT1] =
              mDefaultColours[Token.COMMENT1];
            styles[Token.LITERAL1] = mDefaultColours[Token.LITERAL1];
            styles[Token.KEYWORD1] = mDefaultColours[Token.KEYWORD1];
            theBtn.setBackground(styles[mTokenMap[theList.getSelectedIndex()]].getColor());
        }
    }

    /*public static void main(String args[])
    {
        KOptionsDialog ka = new KOptionsDialog(null);
        ka.pack();
        ka.show();
    }*/
}