/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001,2002 Robert Chatley, Alan Chan                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

/*
 * KOutputWindow.java
 *
 * Displays output and accepts input from Interpreter, also contains
 * the variable watch panel window.
 */

package com.chatley.kenya.gui;


import javax.swing.*;
import javax.swing.text.*;

import java.awt.*;
import java.awt.event.*;
import java.io.*;

import com.chatley.kenya.minijava.node.*;
import com.chatley.kenya.minijava.parser.*;
import com.chatley.kenya.minijava.lexer.*;
import com.chatley.kenya.interpreter.*;
import com.chatley.kenya.gui.varview.*;
import com.chatley.kenya.utils.IMTrace;
import com.chatley.kenya.*;

class KOutputWindow extends JFrame implements KenyaOutputWindow{

    private JTextArea outputArea;
    private JScrollPane outputScrollPane;
    private InputBox inputbox;
    private JToolBar toolbar;

    private KVarWatchPanel mVarWatchPanel;
    private JSplitPane     mSplitPane;
    private JPanel         mOutputPanel;

    private KenyaUI ui;

    /** Output from interpreter is written to this stream, which echoes
        it to the output window.*/
    private JTextAreaWriter mStdout;

    /** Reads output from interpreter.*/
    private /*Piped*/Reader mOutputFromInt;

    /** Pipe which input to interpreter from input box is written to.*/
    private PipedWriter mInputToInt;

    /** Pipe which interpreter reads input from.*/
    private PipedReader mStdin;

    /** Writer used by interpreters to write to pipe.*/
    private PrintWriter mStdoutWriter;

    /** print writer used by input box to write to mInputToInt.*/
    private PrintWriter mOutputWriter;

    private Start mAst; // the parsed abstract syntax tree
    private SwingWorker mWorker = null;

    private KInterpreter mInterpreter = null;

    private ExecutionManager mExecutionManager = null;

    private String mLastParsedCode = "";

    private boolean mVarViewWindowVisible = false;

    public KOutputWindow( KenyaUI u , Start ast)
    {
        super("Kenya Interpreter");
        setIconImage(Util.getKenyaIcon().getImage());
        ui = u;
        mAst = ast;

        getContentPane().setLayout( new BorderLayout() );

        outputArea = new JTextArea();
        inputbox = new InputBox(this);

        Font courier = new Font( "Courier", Font.PLAIN, Integer.parseInt ( ui.prop.getProperty( "font.size" , "14") ) );

        inputbox.setFont( courier );

        outputArea.setMargin( new Insets( 5, 5, 5, 5 ) );
        outputArea.setFont( courier );
        outputArea.setLineWrap(true);
        outputArea.setWrapStyleWord(true);
        outputArea.setEditable(false);

        outputScrollPane = new JScrollPane(outputArea);
        int sizex = Integer.parseInt(ui.prop.getProperty("outputwindow.width","400"));
        int sizey = Integer.parseInt(ui.prop.getProperty("outputwindow.height","300"));

        outputScrollPane.setPreferredSize( new Dimension ( sizex, sizey ) );
        outputScrollPane.setVerticalScrollBarPolicy(
                            JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        outputScrollPane.setBorder(
          BorderFactory.createCompoundBorder(
            BorderFactory.createCompoundBorder(
              BorderFactory.createTitledBorder("Program Output"),
              BorderFactory.createEmptyBorder(5,5,5,5)),
              outputScrollPane.getBorder()));

        toolbar = new KInterToolBar(ui);

        mOutputPanel = new JPanel(new BorderLayout());
        mOutputPanel.add(toolbar, BorderLayout.NORTH);
        mOutputPanel.add( outputScrollPane , BorderLayout.CENTER );

        JPanel inputArea = new JPanel(new BorderLayout());
        inputArea.add(inputbox,BorderLayout.CENTER);
        inputArea.add(new EOFButton(this),BorderLayout.EAST);
        mOutputPanel.add( inputArea , BorderLayout.SOUTH );

        mVarWatchPanel = new KVarWatchPanel();
        mSplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
        mSplitPane.setLeftComponent(mOutputPanel);

        getContentPane().add(mSplitPane,BorderLayout.CENTER);

        addWindowListener(new WindowAdapter(){
            public void windowActivated(WindowEvent e){
                // query the font size and set it
                outputArea.setFont(ui.kenyaArea().getPainter().getFont());
            }
        });
    }

    /**
     * creates IO connections between interpreter and the input source
     * and output destination IO pipes.
     */
    private void createIOPipes() throws IOException
    {
        // output pipes
        mStdout = new JTextAreaWriter(outputArea);
        mStdoutWriter = new PrintWriter(mStdout,true);

        // input pipes
        mInputToInt  = new PipedWriter();
        mOutputWriter = new PrintWriter(mInputToInt);
        mStdin = new PipedReader(mInputToInt);

    }

    private Start parse(String code)
    {
        Start ast = null;
        try{
            // thread safe >>>
            ui.msgsArea.append("Parsing...\n");
            //ui.javaArea.setText("");
            // <<<

            Lexer lexer =
              new Lexer(new PushbackReader(new StringReader(code)));
            Parser parser = new KParser(lexer);
            ast = parser.parse();
            ui.msgsArea.append("done.\n");
            ui.msgsArea.append("Type checking...\n");
            ui.kenyaArea.removeAllLineHighlights();

            // Typecheck

            KUIErrorMsgWriter msgWriter = new KUIErrorMsgWriter(ui);

            KTypeChecker tc = new KTypeChecker(msgWriter,ast);
            ui.msgsArea.append("done.\n");

            if(tc.execute())
            {
                DataFlowAnalyser dfa = new DataFlowAnalyser(msgWriter);
                ast.apply(dfa);
                if(dfa.errorsHaveOccurred()) ast = null;
            }
            else
            {
                ast = null;
            }
        }
        catch (ParserException pe )
        {
            ui.msgsArea.append( pe.getMessage() );
            ui.kenyaArea.addLineHighlight(pe.getToken().getLine(),Color.orange);
            ui.kenyaArea.scrollTo(pe.getToken().getLine(),0);
            ast = null;
        }
        catch (LexerException e )
        {
            ui.msgsArea.append( e.getMessage() );;
            ui.kenyaArea.addLineHighlight(e.line,Color.orange);
            ui.kenyaArea.scrollTo(e.line,0);
            ast = null;
        }
        catch ( java.io.IOException e ) {
            ui.msgsArea.append("Error. Probable syntax error.");
            ast = null;
        }
        catch ( Exception e ) {
            e.printStackTrace();
            ast = null;
        }
        return ast;
    }

    /**
     * Output poller thread, reads output from interpreter and prints it.
     */
    /*public void run()
    {
        try
        {
            int c;

            while ( ( c = mOutputFromInt.read() ) != -1 ){
                outputArea.append( String.valueOf( (char)c ) );
                outputArea.setCaretPosition(outputArea.getText().length());
            }
            IMTrace.out("Output reading done.");
            //mExecutionManager.outputDone();
        }
        catch(IOException e) {
            IMTrace.out("Error on inStream.readLine()");
            e.printStackTrace();
        }
        finally{
            try{
                if(mOutputFromInt!=null) mOutputFromInt.close();
                if(mInputToInt!=null) mInputToInt.close();
            }catch(IOException ie){}
        }
    }*/

    /**
     * create and start a new instance of the interpreter in a separate
     * thread.
     * @param step whether to start the thread in 'step into code' mode.
     */
    private void createInterpreterThread(boolean step)
    {
        if (mExecutionManager==null){
            mExecutionManager = new ExecutionManager();
        }

        mExecutionManager.addListener(ui);
        mExecutionManager.addListener(mVarWatchPanel);


        if(step) mExecutionManager.stepModeOn();
        else mExecutionManager.stepModeOff();

        mWorker = new SwingWorker()
        {
            public Object construct()
            {
                try
                {
                    createIOPipes();

                    ui.msgsArea.append(">>> Starting interpreter\n");
                    mInterpreter = new KInterpreter(
                                   mStdin,mStdoutWriter,
                                   mExecutionManager,mAst);
                    mInterpreter.execute();
                    ui.msgsArea.append("<<< Exiting interpreter\n");
                }
                catch(IOException ie){
                    ui.msgsArea.append(
                        "\nError initialising interpreter I/O pipes!");
                    mInputToInt = null;
                    mStdin = null;
                    mStdout = null;
                    mOutputFromInt = null;
                    mStdoutWriter = null;
                }
                catch(KInterException ke)
                {
                    ui.msgsArea.setText("");
                    ui.msgsArea.append(ke.getMessage());
                    int line = ke.getLine();
                    if(line!=-1){
                        ui.removeAllLineHighlights();
                        ui.addLineHighlight(line,Color.orange);
                    }
                }
                catch(KInterHaltException ihe)
                {
                    ui.msgsArea.append("<<< Interpreter halted");
                }
                catch(Exception e){e.printStackTrace();}

                return new Integer(0);
            }

            public void finished()
            {
                mWorker = null;
                mInterpreter = null;
                IMTrace.out("Interpreter SwingWorker thread dying.");
            }
        };
        mWorker.start();
    }

    private class RunListener implements ActionListener
    {
        public void actionPerformed(ActionEvent event)
        {
            if(mWorker==null)
            {
                String code =
                  Util.trimCarriageReturns(ui.kenyaArea.getText());

                if(!mLastParsedCode.equals(code))
                {
                    ui.msgsArea.setText("");
                    ui.msgsArea.setText("Code changed since last parse.\n");
                    ui.removeAllLineHighlights();
                    mLastParsedCode = code;

                    // re-parse... parse() will return ast if parse
                    // was successful, null otherwise.
                    mAst = parse(code);
                    if(mAst==null)
                        return;
                }
                createInterpreterThread(false);
            }
            else
            {
                mExecutionManager.run();
            }
        }
    }// end class RunListener

    private class StepListener implements ActionListener
    {
        public void actionPerformed(ActionEvent event)
        {
            if(mWorker==null)
            {
                String code =
                  Util.trimCarriageReturns(ui.kenyaArea.getText());

                if(!mLastParsedCode.equals(code))
                {
                    ui.msgsArea.setText("Code changed since last parse.\n");
                    ui.removeAllLineHighlights();
                    mLastParsedCode = code;

                    // re-parse... parse() will return ast if parse
                    // was successful, null otherwise.
                    mAst = parse(code);
                    if(mAst==null)
                        return;
                }
                createInterpreterThread(true);
            }
            else
            {
                mExecutionManager.stepModeOn();
                mExecutionManager.step();
            }
        }
    }// end class StepListener

    private class AddBreakPointListener implements ActionListener
    {
        public void actionPerformed(ActionEvent event)
        {
            IMTrace.out("Breakpoint requested on line: "+ (ui.kenyaArea.getCaretLine() + 1));
            if(mExecutionManager==null){
                mExecutionManager = new ExecutionManager();
                mExecutionManager.addListener(ui);
            }

            mExecutionManager.setBreakpoint(ui.kenyaArea.getCaretLine() + 1);
        }
    }

    private class StopListener implements ActionListener
    {
        public void actionPerformed(ActionEvent event)
        {
            if(mInterpreter!=null){
                ui.msgsArea.append("Requesting halt...\n");
                mInterpreter.requestHalt();
                mExecutionManager.step();
            }
        }
    }

    private class ClearListener implements ActionListener
    {
        public void actionPerformed(ActionEvent event)
        {
            outputArea.setText("");
        }
    }

    private class RemoveBreakPointListener implements ActionListener
    {
        public void actionPerformed(ActionEvent event)
        {
            IMTrace.out("Breakpoint removal requested on line: "+ (ui.kenyaArea.getCaretLine() + 1));
            if(mExecutionManager!=null)
                mExecutionManager.removeBreakpoint(ui.kenyaArea.getCaretLine() + 1);
        }
    }

    private class ClearBreakPointsListener implements ActionListener
    {
        public void actionPerformed(ActionEvent event)
        {
            IMTrace.out("Remove all breakpoints requested");
            if(mExecutionManager!=null)
                mExecutionManager.clearBreakpoints();
        }
    }

    private class VarViewListener implements ActionListener
    {
        public void actionPerformed(ActionEvent event)
        {
            if(mVarViewWindowVisible){
                mSplitPane.remove(mVarWatchPanel);
                mVarViewWindowVisible = false;
            }
            else {
                mSplitPane.setRightComponent(mVarWatchPanel);
                mVarViewWindowVisible = true;
            }
        }
    }


    // the toolbar which goes at top of the output window.
    private class KInterToolBar extends JToolBar
    {
        KInterToolBar(KenyaUI ui)
        {
            String slash = System.getProperty( "file.separator" );

            JButton button;

            button = new JButton(new ImageIcon(ClassLoader.getSystemResource("images/run.gif")));
            button.setToolTipText("Run");
            button.addActionListener( new RunListener() );
            add(button);

            button = new JButton(new ImageIcon(ClassLoader.getSystemResource("images/stepInto.gif")));
            button.setToolTipText("Step into");
            button.addActionListener( new StepListener() );
            add(button);

            button = new JButton(new ImageIcon(ClassLoader.getSystemResource("images/stop.gif")));
            button.setToolTipText("Stop");
            button.addActionListener(new StopListener());
            add(button);
            addSeparator();

            button = new JButton(new ImageIcon(ClassLoader.getSystemResource("images/addBreak.gif")));
            button.setToolTipText("Set breakpoint");
            button.addActionListener( new AddBreakPointListener() );
            add(button);

            button = new JButton(new ImageIcon(ClassLoader.getSystemResource("images/removeBreak.gif")));
            button.setToolTipText("Remove breakpoint");
            button.addActionListener( new RemoveBreakPointListener() );
            add(button);

            button = new JButton(new ImageIcon(ClassLoader.getSystemResource("images/clearBreaks.gif")));
            button.setToolTipText("Clear All breakpoints");
            button.addActionListener( new ClearBreakPointsListener() );
            add(button);
            addSeparator();

            button = new JButton(new ImageIcon(ClassLoader.getSystemResource("images/clear.gif")));
            button.setToolTipText("Clear output window");
            button.addActionListener( new ClearListener() );
            add(button);

            button = new JButton(new ImageIcon(ClassLoader.getSystemResource("images/varWatch.gif")));
            button.setToolTipText("Variable watch");
            button.addActionListener( new VarViewListener() );
            add(button);

        }
    } // end class KInterToolBar

    // >>> HACK! this is a foul hack, but it speeds up the output
    // of the interpreter, and saves us having to spend too much
    // time keeping the separate threads synchronised. Basically, instead
    // of buffering output from the interpreter, and then using a separate
    // thread to print it to the output window, the printing is now
    // done by the interpreter thread, this makes sure that output is kept
    // in sync with interpreter execution, as well releasing the need for
    // an extra thread. 'Apparent Responsiveness' seems to be better too.
    //private class KenyaOutputStream extends ByteArrayOutputStream
    //{
    //    public synchronized void write(byte b[], int off, int len) {
   //         outputArea.append(new String(b,off,len));
   //     }
    //}

    // -------- implementation of the KenyaOutputWindow interface -------

    /**
     * Returns a printWriter to which user input to a kenya program should
     * be written to (in this case, we send it to the interpreter).
     */
    public PrintWriter getOutputWriter()
    {
        return mOutputWriter;
    }

    /**
     * Append the given string to the output window.
     * @param s the string to append.
     */
    public void appendText(String s)
    {
        outputArea.append(s);
    }
}
