/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001,2002 Robert Chatley, Alan Chan                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

/*  KenyaTextArea.java
 *
 *  Kenya's text area component. A customised version
 *  of <code>JEditTextArea</code>.
 *  Alan Chan 2002/1
 */

package com.chatley.kenya.gui.syntaxHighlight;

import com.chatley.kenya.gui.syntaxHighlight.jeditSyntax.*;
import java.awt.*;
import javax.swing.text.*;

public class KenyaTextArea extends JEditTextArea
{
    public static final int KENYA_MODE = 1;
    public static final int JAVA_MODE  = 2;

    public static final int SMALL_FONT_SIZE = 12;
    public static final int LARGE_FONT_SIZE = 18;

    /**
     * Creates a new KenyaTextArea in Kenya syntax highlighting mode
     */
    public KenyaTextArea()
    {
        this(KENYA_MODE);
    }

    /**
     * Creates a new KenyaTextArea with the specified syntax highlighting mode
     * @param mode the specified mode, should be <code>KenyaTextArea.KENYA_MODE</code>
     *             or <code>KenyaTextArea.JAVA_MODE</code>.&nsbp;If anything else
     *             is given, Kenya mode is used.
     */
    public KenyaTextArea(int mode)
    {
        super(new KenyaDefaults());
        if(mode != JAVA_MODE)
            this.setTokenMarker(new KenyaTokenMarker());
        else
            this.setTokenMarker(new JavaTokenMarker());
        // set default font to be courier size 12
        //getPainter().setFont(new Font("Courier",Font.PLAIN,12));
        setFontSize(SMALL_FONT_SIZE);
    }

    /**
     * Sets the number of characters to expand tabs to.
     * This will be multiplied by the maximum advance for
     * variable width fonts.  A PropertyChange event ("tabSize") is fired
     * when the tab size changes.
     *
     * @param size number of characters to expand to
     * @see #getTabSize
     * @beaninfo
     *   preferred: true
     *       bound: true
     * description: the number of characters to expand tabs to
     */
    public void setTabSize(int size) {
        Document doc = getDocument();
        if (doc != null) {
            int old = getTabSize();
            doc.putProperty(PlainDocument.tabSizeAttribute, new Integer(size));
            firePropertyChange("tabSize", old, size);
        }
    }

    /**
     * Gets the number of characters used to expand tabs.  If the document is
     * null or doesn't have a tab setting, return a default of 8.
     *
     * @return the number of characters
     */
    public int getTabSize() {
        int size = 8;
        Document doc = getDocument();
        if (doc != null) {
            Integer i = (Integer) doc.getProperty(PlainDocument.tabSizeAttribute);
            if (i != null) {
                size = i.intValue();
            }
        }
        return size;
    }

    /**
     * Appends the given text to the end of the document.  Does nothing if
     * the model is null or the string is null or empty.
     *
     * @param str the text to insert
     */
    public void append(String str) {
        Document doc = getDocument();
        if (doc != null) {
            try {
                doc.insertString(doc.getLength(), str, null);
            } catch (BadLocationException e) {
            }
        }
    }

    /**
     * Sets the point size of the text used.
     * @param size the point size to set.
     */
    public void setFontSize(int size)
    {
        getPainter().setFont(new Font("Courier",Font.PLAIN,size));
        getGutter().repaint();
    }

    /**
     * Returns the point size of the font being used rounded to an integer.
     */
    public int getFontSize()
    {
        return getPainter().getFont().getSize();
    }

    /**
     * Causes the specified line to be highlighted in the specified
     * colour.
     * @param line the line number of line to highlight.
     * @param c the colour to highlight in.
     */
    public void addLineHighlight(int line, Color c)
    {
        getPainter().addLineHighlight(line,c);
    }

    /**
     * Returns true iff the line specified has been set to highlight, false
     * otherwise.
     * @param line the line to test
     */
    public boolean isLineHighlighted(int line)
    {
        return getPainter().isLineHighlighted(line);
    }

    /**
     * Returns the colour which is used to highlight the specified line
     * if this line has been set to be highlighted. null is returned otherwise.
     * @param line the specified line
     */
    public Color getLineHighlightColour(int line)
    {
        return getPainter().getLineHighlightColour(line);
    }

    /**
     * Causes the specified line to stop being highlighted if it
     * is currently set to be highlighted.
     * @param line the line to remove highlighting from.
     */
    public void removeLineHighlight(int line)
    {
        getPainter().removeLineHighlight(line);
    }

    /**
     * Removes all user defined highlighting.
     */
    public void removeAllLineHighlights()
    {
        getPainter().removeAllLineHighlights();
    }
}