/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001,2002 Robert Chatley, Alan Chan                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

/**
 * KVarWatchPanel.java
 *
 * Variable watch panel of the interpreter output window (KOutputWindow.java)
 * Alan Chan 2002/3
 */
package com.chatley.kenya.gui.varview;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;
import com.chatley.kenya.interpreter.*;
import com.chatley.kenya.interpreter.value.*;
import com.chatley.kenya.utils.IMTrace;

public class KVarWatchPanel extends JPanel
  implements ExecutionListener, ValueSwitch
{
    /** GUI widgets.*/
    private JLabel mScopeLabel;
    private JComboBox  mScopeList;
    private JScrollPane mTreeScrollPane;
    private JTree  mVarTree;

    /** hashtable of scopes and treemodels, one treemodel per scope.
        A "scope" in this case represents the stack of a particular
        function call in a kenya program. Each call has an individual
        stack representing a snapshot of the state of variables of
        the call. Associated with each scope is a treemodel, which is
        used to hold the data to be displayed in the variable watch
        tree.
    */
    private Hashtable mScopeModels = new Hashtable();

    //String mPreviousScope = "<Globals>";

    private Stack mStack = new Stack();

    private Vector mScopes = new Vector();

    /** The scope that is currently being displayed. The scope
        "<Default>" represents a default placeholder without any
        meaning within a kenya program.*/
    private String mCurrentScope = "<Default>";

    private DefaultTreeModel mEmptyModel;

    public KVarWatchPanel()
    {
        initComponents();
    }

    private void initComponents()
    {
        GridBagLayout gbl = new GridBagLayout();
        this.setLayout(gbl);
        GridBagConstraints gbc = new GridBagConstraints();
        // initialise...
        gbc.gridx = 0;
        gbc.gridy = 0;
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gbc.gridheight = 1;
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbc.insets = new Insets(4,4,4,4);
        gbc.weightx = 1;
        gbc.weighty = 0;

        mScopeLabel = new JLabel("Select scope...");

        DefaultMutableTreeNode top =
        new DefaultMutableTreeNode("Welcome to the Kenya variable watcher!");
        //createNodes(top);
        mEmptyModel = new DefaultTreeModel(top);
        mVarTree = new JTree(mEmptyModel);
        mVarTree.setCellRenderer(new ValueTreeCellRenderer());
        mScopes.add("<Default>");
        mStack.push("<Default>");
        mScopeModels.put("<Default>",mEmptyModel);

        mVarTree.putClientProperty("JTree.lineStyle", "Angled");
        mScopeList = new JComboBox(mScopes);
        mTreeScrollPane = new JScrollPane(mVarTree);
        gbl.setConstraints(mScopeLabel,gbc);
        this.add(mScopeLabel);

        gbc.gridy++;
        gbl.setConstraints(mScopeList,gbc);
        this.add(mScopeList);

        gbc.gridy++;
        gbc.weighty = 1;
        gbc.fill = GridBagConstraints.BOTH;
        gbc.gridheight = GridBagConstraints.REMAINDER;
        gbl.setConstraints(mTreeScrollPane,gbc);
        this.add(mTreeScrollPane);

        mScopeList.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e) {
                JComboBox cb = (JComboBox)e.getSource();
                String scope = (String)cb.getSelectedItem();
                DefaultTreeModel m =
                  (DefaultTreeModel) mScopeModels.get(scope);

                if(m!=null) mVarTree.setModel(m);
            }
        });
    }

    // ----- implementation of ExecutionListener interface --------

    public void startingStep(int line){}

    public void endingStep(int line){}

    /** The kenya interpreter will let us know when a variable has
       changed via the ExecutionManager. The ExecutionManager will
       let us know by calling this method.
       @param varName the fully qualifed name of the variable.
       @param scope the name of the scope that is currently executing.
       @param v the new value of the variable.
     */
    public void variableChanged(String varName, String scope, Value v)
    {
        int lastdot = varName.lastIndexOf('.');
        String fldName = (lastdot==-1)? varName : varName.substring(lastdot+1);
        v.setVarName(fldName);

        IMTrace.out("variableChanged:: " + varName);

        // We haven't heard of the scope of the changed variable yet,
        // create a new treemodel for this scope, and add the variable
        // and it's value to the tree by creating a tree node for it and
        // adding it to the tree.
        if(!mScopeModels.containsKey(scope))
        {
            DefaultMutableTreeNode root = new DefaultMutableTreeNode();
            DefaultMutableTreeNode var = (DefaultMutableTreeNode)v.apply(this);
            root.add(var);
            DefaultTreeModel model = new DefaultTreeModel(root);
            mScopeModels.put(scope,model);
            mScopes.add(scope);
            mVarTree.setModel(model);
            mScopeList.setSelectedItem(scope);
            IMTrace.out("*** making new scope model");
        }
        else // otherwise, modify the tree node representing the variable
        {
            // get the treemodel representing the scope
            DefaultTreeModel model = (DefaultTreeModel)mScopeModels.get(scope);
            DefaultMutableTreeNode root =
              (DefaultMutableTreeNode) model.getRoot();

            // try and find the treenode representing the variable
            IMTrace.out("*** existing scope model found");
            ValueTreeNode varNode = (ValueTreeNode)findNode(varName,root);

            Value clone = null;

            if(v.isClass()){
                clone = v;
            }else{
                try{
                    clone = (Value)v.clone();
                }catch(Exception e){ e.printStackTrace(); }
            }

            // if we find the treenode representing variable, update it.
            if(varNode!=null){
                IMTrace.out("variableChanged:: A node with name " + varNode.getVarName() + " was found.");
                IMTrace.out("variableChanged:: v.getVarName = " + v.getVarName());

                IMTrace.out("variableChanged:: cloned object varNAme = " + clone.getVarName());
                if(clone.isReferenceValue()/*clone.isArray() || clone.isClass()*/){
                    IMTrace.out("*** clone is an array/class");
                    DefaultMutableTreeNode parent =
                      (DefaultMutableTreeNode)varNode.getParent();

                    int index = model.getIndexOfChild(parent,varNode);
                    model.removeNodeFromParent(varNode);
                    model.insertNodeInto((ValueTreeNode)clone.apply(this),
                      parent,index);
                }else{
                    varNode.setUserObject(clone);
                    model.nodeChanged(varNode);
                }

                IMTrace.out("*** variableChanged:: Modifying node");
            }
            // otherwise, we must make a new treenode for the variable and
            // add it to the tree. We fall into this case whenever we come
            // across a variable declaration in a scope we have already seen.
            else{
                 IMTrace.out("variableChanged:: Making a new node.");
                 DefaultMutableTreeNode newNode =
                   (DefaultMutableTreeNode)clone.apply(this);
                 if(root.getChildCount()<1){
                    root.add(newNode);
                    model.reload(root);
                 }else{
                    model.insertNodeInto(newNode,root,model.getChildCount(root));
                    IMTrace.out("*** variableChanged:: Adding node");
                }
            }
        }

    }

    // again, called by ExecutionManager
    public void functionEntered(String scope)
    {
        if(!mScopeModels.containsKey(scope))
        {
            DefaultMutableTreeNode root = new DefaultMutableTreeNode();
            DefaultTreeModel model = new DefaultTreeModel(root);
            mScopeModels.put(scope,model);
            mScopes.add(scope);
            //mPreviousScope = (String)mScopeList.getSelectedItem();
            mVarTree.setModel(model);
            IMTrace.out("*** making new scope model " + scope);
        }
        else
        {
            //mPreviousScope = (String)mScopeList.getSelectedItem();
            mVarTree.setModel((DefaultTreeModel)mScopeModels.get(scope));
        }
        mScopeList.setSelectedItem(scope);
        mStack.push(mCurrentScope);
        mCurrentScope = scope;
    }

    // again, called by ExecutionManager
    public void functionExited(String scope)
    {
        if(mScopeModels.containsKey(scope))
        {
            mScopeModels.remove(scope);
            mScopes.remove(scope);
            mScopeList.removeItem(scope);

            String lastScope = (String)mStack.pop();
            DefaultTreeModel model = (DefaultTreeModel)mScopeModels.get(lastScope);

            if(model==null){
                model = mEmptyModel;
            }

            mVarTree.setModel(model);

            mScopeList.setSelectedItem(lastScope);
            mCurrentScope = lastScope;
            IMTrace.out("*** changing scope model to " + lastScope);
        }
    }

    public void interpreterFinished(){}

    public void breakPointSet(int line){}

    public void breakPointRemoved(int line){}

    public void clearBreakpoints(){}

    // -------- implementation of ValueSwitch interface -----------

    public Object caseAArrayValue(ArrayValue v)
    {
        String parentName = v.getVarName();
        IMTrace.out("makeNode(ArrayValue v), array name = " + parentName);
        RValueTreeNode a = new RValueTreeNode(v);

        for(int ka=0;ka<v.getLength();ka++){
            //arraylevel = ka;
            try{
                Value arrayElement = v.getValueAt(ka);
                arrayElement.setVarName(parentName + "[" + ka + "]");
                ValueTreeNode field = (ValueTreeNode)arrayElement.apply(this);
                a.add(field);
            }catch(ArrayValueException ae){ae.printStackTrace();}
        }

        return a;
    }

    public Object caseAClassValue(ClassValue v)
    {
        IMTrace.out("makeNode(ClassValue v)");
        RValueTreeNode c = new RValueTreeNode(v);
        String[] fldNames = v.fieldNames();

        IMTrace.out("*** caseAClassValue:: class fields:-");
        //for(int k=0;k<fldNames.length;k++) IMTrace.out(fldNames[k]);

        for(int k=0;k<fldNames.length;k++){
            try{
                ValueTreeNode field =
                  (ValueTreeNode)v.getValue(fldNames[k]).apply(this);
                // ok, the line below sets the name of a field in a class.

                //v.getValue(fldNames[k]).setVarName(fldNames[k]);
                field.setVarName(fldNames[k]);

                c.add(field);
            }catch(Exception e){ e.printStackTrace(); }
        }
        return c;
    }

    public Object caseAStringValue(StringValue v)
    {
        IMTrace.out("makeNode(StringValue v)");
        return new PValueTreeNode(v);
    }

    public Object caseAIntValue(IntValue v)
    {
        IMTrace.out("makeNode(IntValue v)");
        return new PValueTreeNode(v);
    }

    public Object caseADoubleValue(DoubleValue v)
    {
        IMTrace.out("makeNode(DoubleValue v)");
        return new PValueTreeNode(v);
    }

    public Object caseACharValue(CharValue v)
    {
        IMTrace.out("makeNode(CharValue v)");
        return new PValueTreeNode(v);
    }

    public Object caseABoolValue(BoolValue v)
    {
        IMTrace.out("makeNode(BoolValue v)");
        return new PValueTreeNode(v);
    }

    public Object caseANullValue(NullValue v)
    {
        IMTrace.out("makeNode(NullValue v)");
        return new PValueTreeNode(v);
    }

    // ---------------- Tree modification helpers -----------------

    private DefaultMutableTreeNode findNode(String varName, DefaultMutableTreeNode parent)
    {
        int dot = varName.indexOf('.');
        String receiver = (dot==-1) ? varName : varName.substring(0,dot);

        if(receiver.indexOf('[')!=-1){
            parent = findArrayNode(receiver,parent,0);
            if(parent==null) return null;
            if(dot==-1) return parent;
            else{
                int oldDot = dot;
                dot = varName.indexOf('.',oldDot+1);
                if(dot!=-1) receiver = varName.substring(oldDot+1,dot);
                else receiver = varName.substring(oldDot+1);
            }
        }

        for(Enumeration e = parent.children();e.hasMoreElements();)
        {
            ValueTreeNode child = (ValueTreeNode)e.nextElement();
            if(child.getVarName().equals(receiver)){
                if(dot==-1) return child;
                else return findNode(varName.substring(dot+1),child);
            }
        }
        return null;
    }

    private ValueTreeNode findArrayNode(String arrName,
      DefaultMutableTreeNode parent, int depth)
    {
        int bracket = arrName.indexOf('[');
        String part = (bracket==-1) ? arrName : arrName.substring(0,bracket);

        for(int ka=1;ka<=depth;ka++){
            IMTrace.out("in loop, depth = " + depth + ", ka = " + ka);
            bracket = arrName.indexOf(']',bracket+1);

            if(bracket!=-1) part = arrName.substring(0,bracket+1);
            else return null;
        }

        IMTrace.out("*** arrName = " + arrName);
        IMTrace.out("*** looking for: " + part);

        ValueTreeNode array = null;

        int matches = 0;

        for(Enumeration e = parent.children();e.hasMoreElements();)
        {
            ValueTreeNode child = (ValueTreeNode)e.nextElement();
            if(child.getVarName().equals(part)){
                array = child;
                //break;
                IMTrace.out((++matches) + " matches for " + part + " found.");
            }
        }

        if(array!=null){
            if(arrName.equals(part)){
                IMTrace.out("Array " + arrName + " found!");
                return array;
            }
            else return findArrayNode(arrName,array,depth+1);
        }else{
            return null;
        }
    }

    // ------------------------------------------------------------

    public static void main(String[] args)
    {
        JFrame ka = new JFrame();
        ka.getContentPane().add(new KVarWatchPanel());
        ka.pack();
        ka.show();
        ka.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    }
}