/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001,2002 Robert Chatley, Alan Chan                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

/**
 * ExecutionListener.java
 *
 * Interface which objects which are interested in the execution of
 * the interpreter need to implement.&nbsp;The object will need to implement
 * this interface, and then register themselves with the ExecutionManager, and
 * the ExecutionManager will notify them of the interpreter execution events
 * as and when they occur by calling them on the methods below.
 *
 * @author askc98 2002/2
 */

package com.chatley.kenya.interpreter;

import com.chatley.kenya.interpreter.value.*;

public interface ExecutionListener
{
    /**
     * The interpreter has reached the start of a new statement, but
     * has not executed it yet.
     * @param line the line on which this statement occurs.
     */
    public void startingStep(int line);

    /**
     * The interpreter has finished executing a statement.
     * @param line the line on which this statement occurs.
     */
    public void endingStep(int line);

    /**
     * A variable has been assigned a new value.
     * @param varName the name of the variable.
     * @param scope the name of the scope the variable was declared in.
     * @param v the new value of the variable.
     */
    public void variableChanged(String varName, String scope, Value v);

    /**
     * The interpreter has just entered a new function call.
     * @param scope the scope name given to the function call.
     */
    public void functionEntered(String scope);

    /**
     * The interpreter has just exited a function call.
     * @param scope the scope name given to the function call.
     */
    public void functionExited(String scope);

    /**
     * The interpreter has finished execution.
     */
    public void interpreterFinished();

    /**
     * A breakpoint has been set.
     * @param line the line on which the breakpoint was set.
     */
    public void breakPointSet(int line);

    /**
     * A breakpoint has been removed.
     * @param line the line on which the breakpoint was removed.
     */
    public void breakPointRemoved(int line);

    /**
     * All breakpoints have been removed.
     */
    public void clearBreakpoints();
}