/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001,2002 Robert Chatley, Alan Chan                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

/**
 * ExecutionManager.java
 *
 * Manages the interpretation of a kenya program by halting execution
 * at breakpoints, or when in step-through mode.<br>Also notifies interested
 * parties of events which have occurred in the interpreter, see
 * <code>ExecutionListener</code> for more details.
 */

package com.chatley.kenya.interpreter;

import com.chatley.kenya.interpreter.value.*;
import com.chatley.kenya.utils.IMTrace;
import java.util.*;


public class ExecutionManager
{
    /** Whether we are operating in step-thru mode.*/
    private boolean mStepModeOn;
    /** The set of lines on which breakpoints have been set.*/
    private Set mBreakPts;
    private Object mLock;
    /** Listeners to notify of interpreter events.*/
    private Set mListeners;

    public ExecutionManager()
    {
        mStepModeOn = false;
        mBreakPts = Collections.synchronizedSet(new HashSet());
        mLock = new Object();
        mListeners = Collections.synchronizedSet(new HashSet());
    }

    //-------------------------------------------------------------
    // This method is called by the interpreter thread just before
    // it executes each statement in a kenya program. It checks to
    // see whether the current line is a breakpoint line, or whether
    // the user is stepping thru code, and suspends execution if
    // either of these is true. Interpretation does not continue until
    // the interpreter thread is woken up by a call from another thread
    // (usually the Swing event dispatch thread) to the run() or
    // step() methods (see below).
    //-------------------------------------------------------------
    /*public*/ synchronized void checkHalt(int line)
    {
        if(mStepModeOn || isBreakPoint(line)){
            try{
                IMTrace.out("Sleeping on line: " + line);
                // notify listeners
                synchronized(mListeners) {
                    for(Iterator i = mListeners.iterator();i.hasNext();){
                        ExecutionListener el = (ExecutionListener)i.next();
                        el.startingStep(line);
                    }
                }
                wait();

                IMTrace.out("Waking up on line: " + line);
                // notify listeners
                synchronized(mListeners) {
                    for(Iterator i = mListeners.iterator();i.hasNext();){
                        ExecutionListener el = (ExecutionListener)i.next();
                        el.endingStep(line);
                    }
                }
            }catch(InterruptedException ie){}
        }
    }

    /**
     * Call this method to set a breakpoint on a line.
     * @param line the line to set breakpoint at.
     */
    public void setBreakpoint(int line)
    {
        synchronized(mBreakPts){
            mBreakPts.add(new Integer(line));
        }
        synchronized(mListeners) {
            for(Iterator i = mListeners.iterator();i.hasNext();){
                ExecutionListener el = (ExecutionListener)i.next();
                el.breakPointSet(line);
            }
        }
    }

    /**
     * Call this method to remove a breakpoint on a line.
     * @param line the line to remove breakpoint from.
     */
    public void removeBreakpoint(int line)
    {
        synchronized(mBreakPts){
            mBreakPts.remove(new Integer(line));
        }
        synchronized(mListeners) {
            for(Iterator i = mListeners.iterator();i.hasNext();){
                ExecutionListener el = (ExecutionListener)i.next();
                el.breakPointRemoved(line);
            }
        }
    }

    /**
     * Call this method to remove all previously set breakpoints
     */
    public void clearBreakpoints()
    {
        synchronized(mBreakPts){
            mBreakPts.clear();
        }
        synchronized(mListeners) {
            for(Iterator i = mListeners.iterator();i.hasNext();){
                ExecutionListener el = (ExecutionListener)i.next();
                el.clearBreakpoints();
            }
        }
    }

    public boolean isBreakPoint(int line)
    {
        synchronized(mBreakPts){
            return mBreakPts.contains(new Integer(line));
        }
    }

    /**
     * Called by the interpreter thread so that we can notify any
     * interested listeners that a variable has had its value changed.
     * @param varName the name of the variable the has been changed.
     * @param scope the scope in which the variable was declared.
     * @param v the new value of the variable.
     */
    /*public*/ void variableChanged(String varName, String scope, Value v)
    {
        IMTrace.out("*** setVariable():: varName = " + varName + " scope = " + scope);
        synchronized(mListeners) {
            for(Iterator i = mListeners.iterator();i.hasNext();){
                ExecutionListener el = (ExecutionListener)i.next();
                el.variableChanged(varName,scope,v);
            }
        }
    }

    /**
     * Called by the interpreter thread so that we can notify any
     * interested listeners that we are just about to execute a funtion
     * call.
     * @param scope the name of the function.
     */
    /*public*/ void functionEntered(String scope)
    {
        synchronized(mListeners) {
            for(Iterator i = mListeners.iterator();i.hasNext();){
                ExecutionListener el = (ExecutionListener)i.next();
                el.functionEntered(scope);
            }
        }
    }

    /**
     * Called by the interpreter thread so that we can notify any
     * interested listeners that we have finished executing a function.
     * @param scope the name of the function.
     */
    /*public*/ void functionExited(String scope)
    {
        synchronized(mListeners) {
            for(Iterator i = mListeners.iterator();i.hasNext();){
                ExecutionListener el = (ExecutionListener)i.next();
                el.functionExited(scope);
            }
        }
    }

    /**
     * Wakes up the interpreter thread, letting it know
     * that it can continue execution.
     */
    public synchronized void step()
    {
        notifyAll();
    }

    /**
     * Turns on step by step execution mode
     */
    public void stepModeOn()
    {
        synchronized(mLock){
            mStepModeOn = true;
        }
    }

    /**
     * Turns off step by step execution mode, and restarts
     * any blocked interpreter threads.
     */
    public synchronized void stepModeOff()
    {
        synchronized(mLock){
            mStepModeOn = false;
            notifyAll();
        }
    }

    /**
     * Wakes up any blocked interpreter threads so that
     * they run till the next breakpoint (if one exists) or to the end.
     */
    public synchronized void run()
    {
        stepModeOff();
    }

    /**
     * Called by the interpreter when it's finished.&nbsp;This
     * makes the interpreter thread wait for all it's output
     * to be printed on the screen before it can go and die.
     */
    /*public synchronized void waitForOutput()
    {
        while(!mOutputDone)
            try{
                wait();
            }catch(InterruptedException ie){}
    }*/

    /**
     * Called by the output reading thread in KOutputWindow
     * to let the interpreter thread know that it has finished
     * reading it's output, and that it can go and die.
     */
    /*public synchronized void outputDone()
    {
        mOutputDone = true;
        notifyAll();
    }*/

    /**
     * Register a listener to receive interpreter execution events,
     * @param el the listener to register
     */
    public void addListener(ExecutionListener el)
    {
        synchronized(mListeners){
            mListeners.add(el);
        }
    }

    /**
     * Remove the provided listener if it is already registered for receiving
     * interpreter execution events.
     * @param el the listener to remove
     */
    public void removeListener(ExecutionListener el)
    {
        synchronized(mListeners){
            mListeners.remove(el);
        }
    }
}