/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001,2002 Robert Chatley, Alan Chan                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/


/**
 * Interface for Kenya Interpreter<p>
 *
 * This class should be used to run the kenya interpreter, simply
 * create an instance of this class, and then call <code>execute()</code>
 * to run the interpreter.<p>
 *
 * In addition, there is a limited amount of runtime type checking,
 * so it's best if you run the Kenya type checker before running
 * the interpreter.
 */

package com.chatley.kenya.interpreter;

import com.chatley.kenya.minijava.node.*;
import java.io.*;


public class KInterpreter
{

    private Reader mStdin              = null;
    private PrintWriter mStdout        = null;
    private ExecutionManager mEm       = null;
    private Start mAst                 = null;
    private KInterpreterEngine mEngine = null;

    /**
     * Creates a new KInterpreter object. Use this constructor
     * if the input/output for the interpreter is to come from
     * a GUI.&nbsp;Using this constructor also enables line by
     * line step through of the interpreted kenya program, as
     * well as breakpoints.
     * @param aStdin a reader which the interpreter will use to get input.
     * @param aStdout a PrintWriter to which the interpreter will write output.
     * @param aEm execution manager for managing breakpoints and step through.
     * @param aAst the parsed abstract syntax tree of the kenya program to
     *             interpret.
     */
    public KInterpreter(Reader aStdin,PrintWriter aStdout,
      ExecutionManager aEm, Start aAst)
    {
        mStdin  = aStdin;
        mStdout = aStdout;
        mEm     = aEm;
        mAst    = aAst;
        mEngine = new KInterpreterEngine(mStdin,mStdout,mEm);
    }

    /**
     * Creates a new KInterpreter object. Use this constructor
     * if the input/output for the interpreter is to come from a
     * non graphical environment, such as the command line.&nbsp;Breakpoints
     * and line by line step through are disabled.
     * @param aStdin a reader which the interpreter will use to get input.
     * @param aStdout a PrintWriter to which the interpreter will write output.
     * @param aAst the parsed abstract syntax tree of the kenya program to
     *             interpret.
     */
    public KInterpreter(Reader aStdin,PrintWriter aStdout, Start aAst)
    {
        mStdin  = aStdin;
        mStdout = aStdout;
        mAst    = aAst;
        mEngine = new KInterpreterEngine(mStdin,mStdout);
    }

    /**
     * executes the interpeter.
     * @throws KInterException if an error occurs during interpretation.
     */
    public void execute() throws KInterException
    {
        mAst.apply(mEngine);
        mEngine.execute();
    }

    /**
     * requests that the interpeter halts.
     */
    public void requestHalt()
    {
        mEngine.requestHalt();
    }
}