/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001,2002 Robert Chatley, Alan Chan                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

/**
 * Interpreter for Kenya<p>
 *
 * Any program using this class should be ready to catch
 * a <code>KInterException</code>, as all methods can potentially throw
 * one!<p>
 *
 * To understand the mess here fully, you need to look at
 * the kenya grammar at the same time, as this was created
 * directly from the kenya grammar, and follows the same layout.<p>
 *
 * This class should not be used directly, instead an instance of
 * the KInterpreter class should be used.&nbsp;In addition,
 * there is a limited amount of runtime type checking, so it's best
 * if you run the Kenya type checker before running this interpreter.
 */

package com.chatley.kenya.interpreter;

import com.chatley.kenya.minijava.analysis.*;
import com.chatley.kenya.minijava.node.*;
import com.chatley.kenya.interpreter.value.*;
import com.chatley.kenya.interpreter.io.*;
import com.chatley.kenya.utils.*;
import java.io.*;
import java.util.*;

class KInterpreterEngine extends DepthFirstAdapter
{
    /** stores values of variables.*/
    private HashMap mVariables;

    /** temporary values of partially evaluated expressions. */
    private HashMap mValues;

    /** Temporary area used in class declarations. */
    private HashMap mClassTemp;

    /** Stores user defined class names plus templates of them.*/
    private HashMap mClasses;

    /** Whether we are in a class declaration.*/
    private boolean mInClass = false;

    /** Whether we are in preprocessing mode.*/
    private boolean mPreProcessMode = true;

    /**
     * A separate instance of this class is used to interpret
     * each function call. This variable keeps a pointer to
     * another instance of this class that represents the enclosing
     * scope of the function this instance is executing.
     */
    private KInterpreterEngine mParent;

    /** holds entry points to functions in program.*/
    private HashMap mVtable;

    /** Temporary function call stack. */
    private Vector mFStack = null;

    /** Return value (only used if this instance is executing a func call)*/
    private Object mReturnVal = null;

    /**
     * Name of function we are executing with this instance (if we are
     * executing a function that is).
     */
    private String mFName = null;

    /**
     * Output which would usually go to stdout will go to this
     * instead, which usually is a pipe connected to output window,
     * but you could potentially connect this to whatever you want.
     */
    private PrintWriter mStdout = null;

    /**
     * Pipe connected to output window.&nsbp;Allows reading of user input
     * into program, acts like stdin.
     */
    private Reader mStdin = null;

    /**
     * Utility class for handling input made by kenya programs.
     */
    private InputHandler mHandler;

    /**
     * Monitor which synchronises interpreter threads and the output
     * window threads.
     */
    private ExecutionManager mExecutionManager = null;

    /**
     * This will be set to true if the user explicitly requests
     * that the interpreter should halt.
     */
    private boolean mHalt = false;

    private String mScope = "<Globals>";

    private HashMap mScopes;

    /**
     * Use this constructor if running interpreter from command line,
     * break points and step through are disabled.
     * @param stdin pipe to source for receiving user input.
     * @param stdout pipe to destination for printing program output.
     */
    KInterpreterEngine(Reader stdin, PrintWriter stdout)
    {
        this(null,stdin,stdout,null,"<Globals>",null);
    }

    /**
     * Use this constructor if running interpreter with GUI, if you're
     * not using a GUI, then you will have to implement some other
     * method of setting breakpoints etc at command line by interfacing with the
     * <code>ExecutionManager class</code>.
     * @param stdin pipe to source for receiving user input.
     * @param stdout pipe to destination for printing program output.
     * @param em monitor for synchronising interpreter and gui threads.
     */
    KInterpreterEngine(Reader stdin, PrintWriter stdout, ExecutionManager em)
    {
        this(null,stdin,stdout,em,"<Globals>",null);
    }

    /**
     * Use this constructor for the main interpreter (can also be
     * thought of as the one running the main program).
     * @param vTable table of entry points of functions in program,
     *        use the KInterPreproc class to produce this.
     * @param stdin pipe to source for receiving user input.
     * @param stdout pipe to destination for printing program output.
     * @param em monitor for synchronising interpreter and gui threads.
     * @param the name of the scope which we shall be interpreting.
     * @param the scopes which are currently executing.
     */
    KInterpreterEngine(HashMap vTable,Reader stdin, PrintWriter stdout,
      ExecutionManager em, String scope, HashMap scopes)
    {
        mParent           = null; // superfluous?
        mVariables        = null;
        mValues           = null;
        mFStack           = null;
        mVtable           = vTable;
        mStdin            = stdin;
        mStdout           = stdout;
        mHandler          = new InputHandler(mStdin);
        mExecutionManager = em;
        mScopes           = scopes;

        // setup function scope tables, needed for listners and
        // variable watch.
        if(mScopes==null){
            IMTrace.out(IMTrace.ERROR,"mScopes hashtable is null!");
            mScopes = new HashMap();
        }

        if(!mScopes.containsKey(scope)){
            mScopes.put(scope,new Integer(1));
            mScope = scope;
        }else{
            int count = ((Integer)mScopes.get(scope)).intValue();
            mScope = scope + "_" + (++count);
            mScopes.put(scope,new Integer(count));
        }

        // notify listeners
        if(mExecutionManager!=null){
            mExecutionManager.functionEntered(mScope);
        }
    }

    /**
     * Should never be used from outside of this class.&nsbp;Used to execute
     * a function so that it will have it's own environment to run
     * in (like it's own local variables and stuff).
     * @param aParent The KInterpreterEngine instance that spawned this one
     * @param aVtable table of entry points of functions in program.
     * @param aFName name of function this instance will execute.
     * @param stdin pipe to output window for receiving user input.
     * @param stdout pipe to output window for printing program output.
     * @param em monitor for synchronising interpreter and gui threads.
     * @param the name of the scope which we shall be interpreting.
     * @param the scopes which are currently executing.
     */
    private KInterpreterEngine(KInterpreterEngine aParent, HashMap aVtable,
      String aFName, Reader stdin, PrintWriter stdout, ExecutionManager em,
      String scope, HashMap scopes, HashMap classTable)
    {
        this(aVtable,stdin,stdout,em,scope,scopes);
        mParent = aParent;
        mFName = aFName;
        mPreProcessMode = false;
        mClasses = classTable;
    }

    /**
     * Call this method to start the interpreter executing the
     * main() method of the kenya program.
     */
    void execute()
    {
        // check function existance
        if(!mVtable.containsKey("main"))
        {
            throwError("Unrecoverable error! main() not found (1).");
        }
        FuncInfo f = (FuncInfo)mVtable.get("main");
        if(f==null) throwError("Unrecoverable error! main() not found (2).");

        try{
            mPreProcessMode = false;
            mScope = "main"; // set the scope to main() function.
            // notify listeners
            if(mExecutionManager!=null){
                mExecutionManager.functionEntered(mScope);
            }
            f.getEntryPt().apply(this);//do the call
        }
        catch(KInterReturnException re){}
        finally{
            // notify listeners of termination of interpreter
            if(mExecutionManager!=null){
                mExecutionManager.functionExited(mScope);
                mExecutionManager.functionExited("<Globals>");
            }
        }
    }

    /**
     * stores the value of evaluating a node in the AST for use later by
     * other nodes in the tree.
     *
     * @param node a reference to the node in the AST.
     * @param value the value gained by evaluating this node.
     */
    private void setValue(Node node, Value value)
    {
        if(mValues==null) mValues = new HashMap();
        mValues.put(node,value);
    }

    /**
     * retrieves the value which was stored earlier as the result of evaluating
     * the given node in the AST.
     *
     * @param node the Node we wish to retrieve the evaluated value for
     */
    private Value getValue(Node node)
    {
        Object o = (mValues==null)? null: mValues.get(node);
        return ((Value) o);
    }

    // used for cleaning up after yourself, remove intermediate
    // values created when evaluating expressions.
    private void cleanValue(Node node)
    {
        mValues.remove(node);
    }

    /**
     * Get the value of the kenya variable specified.
     * @param varName fully qualified name of variable to get.
     */
    public Value getVariable(String varName) throws ArrayValueException,
      NullValueException, InvalidFieldException
    {
        // the thing we will return in the end
        Value o = null;
        int start = 0;
        // possible index of first '.'
        int nextDot    = varName.indexOf('.');
        boolean done = false;
        String bitBeforeDot;

        // this loop walks a string representing a fully qualified
        // variable name, and gets the Values of each part of the
        // qualified name. For example, if we have a.b.c, it will get
        // a, then query a for b, then query b for c.
        do
        {
            if(nextDot==-1){
                bitBeforeDot = varName.substring(start);
                done = true;
            }else{
                bitBeforeDot = varName.substring(start,nextDot);
            }

            // possible index of first bracket in array access
            int openBracket = bitBeforeDot.indexOf('[');

            // o represents the receiver at a particular iteration of the
            // loop. For example, if we want to evaluate a.b.c, o will be
            // null to begin with, then it will be a, then it will be b
            // finally it will be c, and we are done and return.
            if(o==null)
            {
                if(openBracket==-1) // simple variable access
                {
                    // search local variables
                    o = (mVariables==null)? null:
                      (Value)mVariables.get(bitBeforeDot);
                    // search global
                    if(o==null && mParent!=null)
                        o = mParent.getVariable(bitBeforeDot);
                }
                else // array access
                {
                    String name = bitBeforeDot.substring(0,openBracket);

                    // try local vars, if not there, check global.
                    if((o = (Value)mVariables.get(name)) == null)
                        o = mParent.getVariable(name);

                    if(o.isArray()){
                        o = ((ArrayValue)o).getValueAt(bitBeforeDot);
                    }else{
                        throwError("variable " + name + " is not an array.");
                    }
                }
            }
            else
            {
                if(o.isNull()) // is the receiver null?
                    throw new NullValueException(
                      "unable to access field: " + bitBeforeDot +
                      ", receiver is null -> " + varName);

                else if(!o.isArray()) // normally we have class field access
                    o = ((ClassValue)o).getValue(bitBeforeDot);

                // special case is when we have length of an array
                else if(bitBeforeDot.equals("length"))
                    o = ((ArrayValue)o).getLengthAsValue();
            }
            start = nextDot;
            nextDot = varName.indexOf('.',++start);

        }while(!done);

        return o;
    }

    /**
     * Stores the value of a variable in the kenya program which is being
     * interpetered. The stored value can be retrieved later using the
     * getVariable method.
     *
     * @param varName the name of the variable as declared in the kenya code.
     * @param v the value to set the variable to.
     */
    public void setVariable(String varName, Value v) throws
      ArrayValueException, NullValueException, InvalidFieldException
    {
        int openBracket;
        int lastDot = varName.lastIndexOf('.');
        String field = (lastDot==-1) ? varName : varName.substring(lastDot+1);

        // notify listeners of variable value change
        if(mExecutionManager!=null && !mInClass)
            mExecutionManager.variableChanged(varName,mScope,v);

        if(mVariables == null) mVariables = new HashMap();

        // we are setting a field in an object
        if(lastDot!=-1){
            Value receiver = getVariable(varName.substring(0,lastDot));
            if(receiver.isNull()){
                throw new NullValueException(
                  "Unable to set field, receiver is null --> " +
                    varName.substring(0,lastDot));
            }
            else if(!receiver.isClass()){
                throw new InvalidFieldException("Unable to set field " + varName);
            }else{
                ((ClassValue)receiver).setValue(field,v);
            }
        }
        // we are setting a 'normal' stand alone variable
        else
        {
            // we are setting a scalar
            if((openBracket=varName.indexOf('['))==-1)
            {
                IMTrace.out("in setVariable. varName = " + varName + " mInClass = " + mInClass);

                if(mInClass) //we're initialising fields in a class declaration
                    mClassTemp.put(varName,v);
                else // we're handling a normal var dec
                    mVariables.put(varName,v);
            }
            else // we are setting an element of an array
            {
                Value a;
                int closeBracket;

                String name = varName.substring(0,openBracket);
                IMTrace.out("name = " + name);
                closeBracket = varName.indexOf(']');

                a = (Value)mVariables.get(name);

                if(a.isArray()){
                    ((ArrayValue)a).setValueAt(varName,v);
                }else{
                    throw new ArrayValueException("variable " + name +
                      " is not an array.");
                }
            }
        }
    }

    /**
     * This will throw an exception which should cause interpreter execution
     * to stop.&nbsp;Clients of this class should catch this exception
     * and report the error message to the user.
     */
    private void throwError(String msg)
    {
        throw new KInterException(msg);
    }

    /**
     * This will throw an exception which should cause interpreter execution
     * to stop.&nbsp;Clients of this class should catch this exception
     * and report the error message to the user.
     */
    private void throwError(String msg, int line)
    {
        throw new KInterException(msg,line);
    }

    private boolean isPrimitiveType(String type)
    {
        return ( type.equals("int") || type.equals("double") ||
                 type.equals("String") || type.equals("boolean") ||
                 type.equals("char") );
    }

    /**
     * Escape sequences in kenya strings have to be converted ourselves, this
     * is done here.&nbsp;The Kenya grammar only supports a subset of those
     * supported by Java, so we only need to cater for those cases.
     * @param s the string containing escape sequences to escape
     * @param lineNum the line on which this string occurs in the kenya prog.
     */
    private String escapeString(String s, int lineNum)
    {
        char BS = '\b';         /* \u0008: backspace BS */
        char HT = '\t';         /* \u0009: horizontal tab HT */
        char LF = '\n';         /* \u000a: linefeed LF */
        char FF = '\f';         /* \u000c: form feed FF */
        char CR = '\r';         /* \u000d: carriage return CR */
        char DQ = '\"';         /* \u0022: double quote " */
        char SQ = '\'';         /* \u0027: single quote ' */
        char SL = '\\';         /* \u005c: backslash \ */

        StringBuffer result = new StringBuffer();

        for(int i=0;i<s.length();i++){
            char c = s.charAt(i);
            if(c == '\\' && ((i + 1) < s.length())){
                char next = s.charAt(i+1);
                switch(next){
                    case 'b':  result.append(BS); break;
                    case 't':  result.append(HT); break;
                    case 'n':  result.append(LF); break;
                    case 'f':  result.append(FF); break;
                    case 'r':  result.append(CR); break;
                    case '\"': result.append(DQ); break;
                    case '\'': result.append(SQ); break;
                    case '\\': result.append(SL); break;
                    default:   throwError("Invalid escape sequence",lineNum);
                }
                i++;
            }else{
                result.append(c);
            }
        }
        return result.toString();
    }

    private Value initVariable(String type)
    {
        Value v = null;

        if("int".equals(type)){
            v = ValueFactory.makeValue(0);
        }else if("double".equals(type)){
            v = ValueFactory.makeValue(0.0d);
        }else if("char".equals(type)){
            v = ValueFactory.makeValue('\0');
        }else if("String".equals(type)){
            v = ValueFactory.makeValue("");
        }else if("boolean".equals(type)){
            v = ValueFactory.makeValue(false);
        }else{ // class declaration
            if(mClasses!= null && mClasses.containsKey(type)){
                v = (Value)mClasses.get(type);
                try{
                    v = (Value)v.clone();
                }catch(Exception e){e.printStackTrace();}
            }else{
                throwError("Unable to resolve type name: " + type);
            }
        }
        return v;
    }

    private Value setupArray(ArrayDim aDim, String type) throws ArrayValueException
    {
        Value array = null;

         // are we making an array of objs?
        if(mClasses!=null && mClasses.containsKey(type))
        {
            Value v = (Value)mClasses.get(type);
            try{
                v = (Value)v.clone();
            }catch(Exception e){e.printStackTrace();}
            array = ValueFactory.makeValue(aDim.getDim(1),v);

            if(aDim.getNumDims()>1)
            {
                ArrayValue dim2 =
                  (ArrayValue) ValueFactory.makeValue(aDim.getDim(2),v);
                addArrayDimensions((ArrayValue)array,dim2,aDim,3,v);
            }
        }
        else
        { // no, an array of primitive types
            array = ValueFactory.makeValue(aDim.getDim(1),type);

            if(aDim.getNumDims()>1)
            {
                Value v = ValueFactory.getDefaultPrimValue(type);
                IMTrace.out("making more dims for prims");
                ArrayValue dim2 =
                  (ArrayValue) ValueFactory.makeValue(aDim.getDim(2),v);
                addArrayDimensions((ArrayValue)array,dim2,aDim,3,v);
            }
        }

        IMTrace.out("array in setup array is null: " + (array==null));
        return array;
    }

    /**
     * 'Allocates' memory for a regular shaped array, and initialises
     * the cells to the appropriate default values for the specified type, we
     * make heavy use of cloning of templates for the initialisation.
     *
     * @param dim1 ArrayValue object representing a template of xth dimension
     *             of the array.
     * @param dim2 ArrayValue object represeting a template of x+1 th dimension
     *             of the array.
     * @param aDim the dimesions of the array.
     * @param thirdDim the x+2 th dimension as a number
     * @param klass template of the type of elements of this array
     */
    private void addArrayDimensions(ArrayValue dim1,ArrayValue dim2,
      ArrayDim aDim,int thirdDim, Value klass) throws ArrayValueException
    {
        IMTrace.out("dim1.getLength() = " + dim1.getLength());
        for(int i=0;i<dim1.getLength();i++)
        {
            dim1.setValueAt(i,dim2);
            if(thirdDim<=aDim.getNumDims())
            {
                IMTrace.out("A third dimension exists");
                ArrayValue dim3 = (ArrayValue)
                  ValueFactory.makeValue(aDim.getDim(thirdDim),klass);

                addArrayDimensions((ArrayValue)dim1.getValueAt(i),
                  dim3,aDim,thirdDim+1,klass);
            }
        }
    }

    private String stripQuotes(String s)
    {
        if("\"\"".equals(s)) return "";
        else return s.substring(1,s.length()-2);
    }

    /**
     * If this instance of this class is interpreting a function call, and this
     * function call is supposed to return a value, the return value can be
     * retrieved thru this method.&nbsp;This method is usually called by another
     * instance of this class, which represents the caller of the function being
     * executed by this instance.
     */
    public Value getRetVal()
    {
        return (Value)mReturnVal;
    }

    /**
     * Call this method to request the interpreter stop execution.
     */
    public void requestHalt()
    {
        mHalt = true;
    }

    public void inAListStatements(AListStatements node)
    {
        if(mHalt) throw new KInterHaltException();
    }

    // declaration
    public void caseADecStatement(ADecStatement node)
    {
        node.getDeclaration().apply(this);
    }

    public void inAFunctioncallStatement(AFunctioncallStatement node)
    {
        IMTrace.out("FunctioncallStatement: " + node.getSemicolon().getLine());
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getSemicolon().getLine());
    }

    // assignment
    public void caseAAssignmentStatement(AAssignmentStatement node)
    {
        IMTrace.out("AssignmentStatement: " + node.getSemicolon().getLine());

        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getSemicolon().getLine());

        // get name of variable we are assigning to
        node.getFieldAccess().apply(this);
        String var = getValue(node.getFieldAccess()).toString();
        //evaluate expression
        node.getExpression().apply(this);

        try{
            setVariable(var,getValue(node.getExpression()));
        }catch(ValueException ae){
            IMTrace.out("-----> line ----> " + node.getAssign().getLine());
            throwError(ae.getMessage(),node.getAssign().getLine());
        }

        setValue(node,getValue(node.getExpression()));
    }

    // if
    public void caseAIfStatement(AIfStatement node)
    {
        IMTrace.out("IfStatement: " + node.getIf().getLine());
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getIf().getLine());

        // evaluate if condition
        node.getBoolExpression().apply(this);

        Value result = getValue(node.getBoolExpression());

        if(result.isBool())
        {
            if(((BoolValue)result).getBool()) node.getBlock1().apply(this);
            else{
                if(node.getElse()!=null) node.getBlock2().apply(this);
            }
        }else{
            throw new KInterException("Result of If condition not " +
              "a boolean.",node.getIf().getLine());
        }
    }

    // while loop
    public void caseAWhileStatement(AWhileStatement node)
    {
        IMTrace.out("WhileStatement: " + node.getWhile().getLine());
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getWhile().getLine());

        node.getBoolExpression().apply(this);
        try{
            while(((BoolValue)getValue(node.getBoolExpression())).getBool()){
                cleanValue(node.getBoolExpression());
                node.getBlock().apply(this);
                node.getBoolExpression().apply(this);
            }
        }
        catch(KInterBreakException be){}//break from loop, just continue
        catch(ClassCastException e){
            throw new KInterException("While loop condition did " +
              "not evaluate to a boolean.",node.getWhile().getLine());
        }
    }

    // return statement
    public void caseAReturnStatement(AReturnStatement node)
    {
        IMTrace.out("ReturnStatement: " + node.getReturn().getLine());
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getReturn().getLine());

        if(mFName!=null){ // make sure we are in a function call
            if(!((FuncInfo)mVtable.get(mFName)).getType().equals("void"))
            { // only evaluate expression if we have to (and can).
                if(node.getExpression()!=null)
                    node.getExpression().apply(this);
                else
                    throwError("Return statement without expression.",
                      node.getReturn().getLine());

                mReturnVal = getValue(node.getExpression());
            }
        } // signal that we want to return from function call.
        throw new KInterReturnException(node.getReturn().getLine());
    }

    // switch statement
    public void caseASwitchStatement(ASwitchStatement node)
    {
        IMTrace.out("ASwitchStatement: " + node.getSwitch().getLine());
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getSwitch().getLine());

        node.getExpression().apply(this);
        Value v = getValue(node.getExpression());

        // throw exception if not trying to switch on int or char.
        if(!v.isInt() && !v.isChar()){
            throw new KInterException(
              "Switch can only be performed char or int.",
              node.getSwitch().getLine());
        }// store the result of evaluating switch condition...
        setValue(node,getValue(node.getExpression()));
        node.getSwitchBlock().apply(this);
        cleanValue(node);
        cleanValue(node.getExpression());
    }

    public void caseASwitchBlock(ASwitchBlock node)
    {
        Object temp[] = node.getPossibleCase().toArray();
        setValue(node,getValue(node.parent())); // value of switch exp
        for(int i = 0; i < temp.length; i++)
        {
            try{
                ((PPossibleCase) temp[i]).apply(this);
                // if one of the cases has matched switch, start
                // fall through mode...
                if(getValue((Node)temp[i])!=null){
                    setValue(node,getValue((Node)temp[i]));
                    cleanValue((Node)temp[i]);
                }// a break means we stop looking for matching cases
            }catch(KInterBreakException e){break;}
        }
    }

    public void caseAPossibleCase(APossibleCase node)
    {
        try{
            // first we see if fall thru mode is on, if so, don't
            // bother evaluating case, just execute code block
            if(getValue(node.parent()).isBool() &&
               (((BoolValue)getValue(node.parent())).getBool()==true)){
                node.getBlock().apply(this);
                return;
            }
            // otherwise, evaluate case and see if we match switch condition
            node.getSwitchLabel().apply(this);
            Value v = getValue(node.getSwitchLabel());
            cleanValue(node.getSwitchLabel());
            if(v.isBool() && (((BoolValue)v).getBool()==true)) // default?
                node.getBlock().apply(this);
            else if(v.isInt()) // is it a int switch?
            {
                if(((IntValue)v).getInt() ==
                   ((IntValue)getValue(node.parent())).getInt()){
                    node.getBlock().apply(this);
                    setValue(node,ValueFactory.makeValue(true)); // signal fall thru
                }
            }
            else if(v.isChar()) // assume char switch
            {
                if(((CharValue)v).getChar() ==
                   ((CharValue)getValue(node.parent())).getChar()){
                    node.getBlock().apply(this);
                    setValue(node,ValueFactory.makeValue(true)); // signal fall thru
                }
            }else{
                throwError("switch case not int or char.");
            }
        }catch(ClassCastException e){
            throwError("Unrecoverable type error in switch case.");
        }
    }

    public void caseACaseSwitchLabel(ACaseSwitchLabel node)
    {
        node.getExpression().apply(this);
        setValue(node,getValue(node.getExpression()));
        cleanValue(node.getExpression());
    }

    public void caseADefaultSwitchLabel(ADefaultSwitchLabel node)
    {
        setValue(node,ValueFactory.makeValue(true));
    }


    // for loop
    public void caseAForStatement(AForStatement node)
    {
        node.getForControl().apply(this);
        try{
            while(((BoolValue)getValue(node.getForControl())).getBool()==true)
            {
                try{
                    node.getBlock().apply(this);
                    node.getForControl().apply(this);
                }catch(KInterBreakException e){break;}
            }
            cleanValue(node.getForControl()); // garbage collection
        }catch(ClassCastException e)
        {
            throwError("For loop condition did " +
              "not evaluate to a boolean.",node.getFor().getLine());
        }
    }

    public void caseABreakStatement(ABreakStatement node)
    {
        IMTrace.out("BreakStatement: " + node.getBreak().getLine());
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getBreak().getLine());

        throw new KInterBreakException(node.getBreak().getLine());
    }

// {assert} assert bool_expression colon_string? semicolon

    public void caseAAssertStatement(AAssertStatement node)
    {
        IMTrace.out("AssertStatement: " + node.getAssert().getLine());
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getAssert().getLine());

        if(node.getBoolExpression()!=null)
        {
            node.getBoolExpression().apply(this);
            Value v = getValue(node.getBoolExpression());
            cleanValue(node.getBoolExpression());

            try
            {
                if(!Value.getBool(v))
                {
                    String errorMsg = "Assertion failed. ";

                    if(node.getColonString()!=null){
                        node.getColonString().apply(this);
                        errorMsg += getValue(node.getColonString());
                        cleanValue(node.getColonString());
                    }
                    throwError(errorMsg,node.getAssert().getLine());
                }
            }catch(NotABoolException nabe){
                throwError("Assertion expression did not return a boolean.",
                  node.getAssert().getLine());
            }
        }
    }

    public void caseAColonString(AColonString node)
    {
        if(node.getExpression()!=null)
        {
            node.getExpression().apply(this);
            setValue(node,
              ValueFactory.makeValue(getValue(node.getExpression()).toString()));
        }
    }

    // array copy
    public void caseAArraycopyStatement(AArraycopyStatement node)
    {
        IMTrace.out("ArraycopyStatement: " + node.getArraycopy().getLine());
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getArraycopy().getLine());

        node.getSrc().apply(this);
        node.getDest().apply(this);

        try{
            Value src = getVariable(getValue(node.getSrc()).toString());
            Value dest = getVariable(getValue(node.getDest()).toString());

            if(!src.isArray() || !dest.isArray())
                throwError("Array copy attempted on non array types",
                  node.getArraycopy().getLine());

            setVariable(getValue(node.getDest()).toString(),
               (Value)((ArrayValue)src).clone());
        }catch(CloneNotSupportedException ce){
            throwError("Error performing array copy",
              node.getArraycopy().getLine());
        }catch(ValueException ae){
            throwError(ae.getMessage(),node.getArraycopy().getLine());
        }

        cleanValue(node.getSrc());
        cleanValue(node.getDest());
    }

    // print statement
    public void caseAPrintStrStatement(APrintStrStatement node)
    {
        IMTrace.out("PrintStatement: " + node.getPrint().getLine());
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getPrint().getLine());

        node.getExpression().apply(this);
        if(mStdout!=null){
            mStdout.print(getValue(node.getExpression()).toString());
            mStdout.flush();
        }
        else
            throwError("Cannot print output, stream null.",
              node.getPrint().getLine());
    }

    // println statement
    public void caseAPrintlnStrStatement(APrintlnStrStatement node)
    {
        IMTrace.out("PrintlnStatement: " + node.getPrintln().getLine());
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getPrintln().getLine());

        node.getExpression().apply(this);
        if(mStdout!=null)
            mStdout.println(getValue(node.getExpression()).toString());
        else
            throwError("Cannot print output, stream null.",
              node.getPrintln().getLine());
    }

    // java version of for loop control statment
    public void caseAJavaForControl(AJavaForControl node)
    {
        IMTrace.out("JavaForControl: " + node.getS1().getLine());
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getS1().getLine());

        if(getValue(node)==null)
        {
            node.getExpression().apply(this);
            node.getName().apply(this);
            try{
                setVariable(getValue(node.getName()).toString(),
                  getValue(node.getExpression()));
            }catch(ValueException ae){
                throwError(ae.getMessage(),node.getS1().getLine());
            }
        }else{
            node.getUnaryExp().apply(this);
        }

        node.getBoolExpression().apply(this);
        setValue(node,getValue(node.getBoolExpression()));
    }

    // 'simple' version of loop control statement
    public void caseASimpleForControl(ASimpleForControl node)
    {
        IMTrace.out("SimpleForControl: " + node.getTo().getLine());
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getTo().getLine());

        double newResult=0; // new value that loop var will take
        if(getValue(node)==null)
        {
            node.getExp1().apply(this);
            node.getName().apply(this);
            try{
                setVariable(getValue(node.getName()).toString(),
                  getValue(node.getExp1()));
            }catch(ValueException ae){
                throwError(ae.getMessage(),node.getAssign().getLine());
            }

            if(getValue(node.getExp1()).isNumber()){
                newResult =
                  ((NumberValue)getValue(node.getExp1())).getNumber();
            }
            else{
                throwError("Expecting number, got " +
                      "something else!",node.getAssign().getLine());
            }

        }else{
            double step = 1;
            if(node.getStepBit()!=null){ // test for step size
                node.getStepBit().apply(this);

                if(getValue(node.getStepBit()).isNumber()){
                    step =
                      ((NumberValue)getValue(node.getStepBit())).getNumber();
                }else{
                    throwError("Step size must be a number.",
                      node.getAssign().getLine());
                }

            } // test for decreasing
            if(node.getDecreasing()!=null) step *= -1;
            node.getName().apply(this);

            //increase loop variable
            try{
                newResult =
                  ((NumberValue)getVariable(getValue(node.getName()).toString())).getNumber();

                newResult+= step;

                if(getVariable(getValue(node.getName()).toString()).isInt())
                    ((IntValue)getVariable(getValue(node.getName()).toString())).setInt((int)newResult);
                else
                    ((DoubleValue)getVariable(getValue(node.getName()).toString())).setDouble(newResult);

             }catch(ClassCastException e2){
                    throwError("Expecting number, got something else!",
                      node.getAssign().getLine());
             }catch(ValueException ae){
                throwError(ae.getMessage(),node.getAssign().getLine());
             }
         }
         node.getExp2().apply(this);
         try{
            if(node.getDecreasing()!=null){
                setValue(node,
                  ValueFactory.makeValue(newResult>=
                    ((NumberValue)getValue(node.getExp2())).getNumber()));
            }else{
                setValue(node,
                  ValueFactory.makeValue(newResult<=
                    ((NumberValue)getValue(node.getExp2())).getNumber()));
            }
         }catch(ClassCastException e3){
            throwError("Expecting number, got something else!",
              node.getAssign().getLine());
         }
    }

    public void caseAStepBit(AStepBit node)
    {
        node.getExpression().apply(this);
        setValue(node,getValue(node.getExpression()));
        cleanValue(node.getExpression());
    }

    // block as in DepthFirstAdapter

    // ------------------- graphics primitives --------------------
    // -------------- not implemented in this version -------------

    public void caseAClearStatement(AClearStatement node)
    {
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getClear().getLine());

        mStdout.println("\nClear() not implemented in current version" +
          " of Kenya interpreter.");
    }

    public void caseADrawlineStatement(ADrawlineStatement node)
    {
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getDrawline().getLine());

        mStdout.println("\nDrawline() not implemented in current version" +
          " of Kenya interpreter.");
    }

    public void caseADrawboxStatement(ADrawboxStatement node)
    {
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getDrawbox().getLine());

        mStdout.println("\nDrawbox() not implemented in current version" +
          " of Kenya interpreter.");
    }

    public void caseADrawtriangleStatement(ADrawtriangleStatement node)
    {
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getDrawtriangle().getLine());

        mStdout.println("\nDrawtriangle() not implemented in current version" +
          " of Kenya interpreter.");
    }

    public void caseADrawellipseStatement(ADrawellipseStatement node)
    {
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getDrawellipse().getLine());

        mStdout.println("\nDrawellipse() not implemented in current version" +
          " of Kenya interpreter.");
    }

    public void caseADrawstringStatement(ADrawstringStatement node)
    {
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getDrawstring().getLine());

        mStdout.println("\nDrawstring() not implemented in current version" +
          " of Kenya interpreter.");
    }

    public void caseASetcolourStatement(ASetcolourStatement node)
    {
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getSetcolour().getLine());

        mStdout.println("\nSetcolour() not implemented in current version" +
          " of Kenya interpreter.");
    }

    public void caseASetbgcolourStatement(ASetbgcolourStatement node)
    {
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getSetbgcolour().getLine());

        mStdout.println("\nSetbgcolour() not implemented in current version" +
          " of Kenya interpreter.");
    }

    public void caseAFillStatement(AFillStatement node)
    {
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getFill().getLine());

        mStdout.println("\nFill() not implemented in current version" +
          " of Kenya interpreter.");
    }

    /* ************************************************************
     function application = name l_parenthese actual_param_list r_parenthese;
    **************************************************************/
    public void caseAFunctionApplication(AFunctionApplication node)
    {
        String fName = node.getName().toString().trim();
        if(!mVtable.containsKey(fName)){ // check function existance
            throwError("unknown function " + fName,
                        node.getLParenthese().getLine());
        }

        // setup parameters for function call
        // we evaluate the expressions for the parameters, and store
        // them in mFStack. If a parameter involves nested function calls,
        // we have to save previous mFStack contents, and restore later
        Vector oldStack = null;
        if(mFStack!=null) oldStack = mFStack; // save mFStack if nested calls
        mFStack = new Vector();

        node.getActualParamList().apply(this);

        FuncInfo f = (FuncInfo)mVtable.get(fName);

        IMTrace.out("f.numParams() = " + f.numParams());
        IMTrace.out("mFStack.size() = " + mFStack.size());

        if(f.numParams()!=mFStack.size()){ // check the number of params
            throwError("Wrong number of parameters in function call.",
                       node.getLParenthese().getLine());
        }

        // param type check
        // The function call parameter type check has been moved to the
        // typechecker, this will allow programs that aren't interpretered,
        // to have this type check done as well. askc98 2002/6
        /*int i = 0;
        for (Enumeration e = mFStack.elements(); e.hasMoreElements();i++)
        {
            String expected = f.getParamType(i);
            String found = ((Value)e.nextElement()).getType();
            if( !found.equals(expected) ){
                throwError("param " + (i+1) + " of function call is wrong type." +
                  " Expected: " + expected + ", found: " + found,
                  node.getLParenthese().getLine());
            }

        }*/

        //prepare parameter stack, and do the function call
        KInterpreterEngine functionCall = new
          KInterpreterEngine(this,mVtable,fName,mStdin,mStdout,
            mExecutionManager,fName,mScopes,mClasses);
        int i = 0;
        for(Enumeration e = mFStack.elements(); e.hasMoreElements();i++)
        {
            try{
                functionCall.setVariable(f.getParam(i),(Value)e.nextElement());
            }catch(ValueException ae){
                throwError(ae.getMessage(),node.getLParenthese().getLine());
            }
        }

        try{
            f.getEntryPt().apply(functionCall);//do the call
        }catch(KInterReturnException re){}
        if (!f.getType().equals("void"))
            setValue(node,functionCall.getRetVal());

        mFStack.clear(); // remove stuff from stack;

        // restore saved stacks if necessary
        if(oldStack!=null) mFStack = oldStack;
        else mFStack = null;

        // clean up scope table
        int count = ((Integer)mScopes.get(fName)).intValue();

        // notify listeners of function call return
        //IMTrace.out("*** removing scope:: " + fName + "_" + count);
        if(count>1){
            if(mExecutionManager!=null){
                mExecutionManager.functionExited(fName + "_" + count);
            }
            mScopes.put(fName, new Integer(--count));
        }
        else{
            if(mExecutionManager!=null){
                mExecutionManager.functionExited(fName);
            }
            mScopes.remove(fName);
        }
        IMTrace.out("done function call");
    }

    /* ************************************************************
    actual_param_list = {list} actual_param_list comma expression |
                        {exp} expression |
                        {empty};
    **************************************************************/
    public void caseAListActualParamList(AListActualParamList node)
    {
        node.getActualParamList().apply(this);
        node.getExpression().apply(this);

        mFStack.add(getValue(node.getExpression()));
        cleanValue(node.getExpression());
    }

    public void caseAExpActualParamList(AExpActualParamList node)
    {
        node.getExpression().apply(this);
        mFStack.add(getValue(node.getExpression()));
        cleanValue(node.getExpression());
    }

    public void caseAEmptyActualParamList(AEmptyActualParamList node){}

    /* ************************************************************
    declaration =
    {class_dec} klass identifier l_brace declaration* r_brace |
    {func_dec}  type bracket_pair* identifier l_parenthese formal_param_list? r_parenthese block |
    {const_dec} const type identifier initialiser? semicolon |
    {var_dec}   type bracket_pair* type_param? identifier array_access* initialiser? semicolon;

    **************************************************************/

    public void caseAClassDecDeclaration(AClassDecDeclaration node)
    {
        if(mInClass)
            throwError("Nested classes are not supported in this version " +
              "of the interpreter.",node.getKlass().getLine());

        IMTrace.out("ClassDeclaration: " + node.getKlass().getLine());
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getKlass().getLine());

        mInClass = true;

        String className = node.getIdentifier().toString().trim();

        if(mClasses!=null && mClasses.containsKey(className)){
             throw new
               KInterException("Class " + className + "already declared.",
                               node.getIdentifier().getLine());
        }

        LinkedList decs = node.getDeclaration();
        ListIterator decsIterator = decs.listIterator(0);

        mClassTemp = new HashMap();

        while( decsIterator.hasNext() )
        {
            Node t = (Node)decsIterator.next();
            t.apply(this);
        }

        if(mClasses==null) mClasses = new HashMap();
        mClasses.put(className,new ClassValue(className,mClassTemp));
        mInClass = false;
    }

    public void caseAFuncDecDeclaration(AFuncDecDeclaration node)
    {
        if(mInClass)
            throwError("class methods are not supported in this version " +
              "of the interpreter.",node.getLParenthese().getLine());

        String fName  = node.getIdentifier().toString().trim();
        if(mVtable == null) mVtable = new HashMap();

        // check for function redefinition.
        if(mVtable.containsKey(fName)){
             throwError("Function " + fName + "already defined.",
                        node.getIdentifier().getLine());
        }

        FuncInfo f = new FuncInfo(fName,
                                  node.getType().toString().trim(),
                                  node.getBlock());

        if(node.getFormalParamList() != null)
        {

            // function call parameter type check has been moved to the
            // typechecker
            // askc98 2002/6
            /*StringTokenizer st = new
              StringTokenizer(node.getFormalParamList().toString().trim(),
                              "[] ,");
            while (st.hasMoreTokens()) // get param names and types
            {
                 String type, name;
                 type = st.nextToken();
                 name = st.nextToken();
                 f.addParam(name,type);
            }*/

            AFormalParamList params = (AFormalParamList)node
              .getFormalParamList();

            ATypeName t = (ATypeName)params.getTypeName();
            String type = ASTUtils.parseTypeName(t);
            String name = t.getIdentifier().toString().trim();
            f.addParam(name,type);

            Object temp[] = params.getCommaTypeName().toArray();
            for(int i = 0; i < temp.length; i++)
            {
                t = (ATypeName) ((ACommaTypeName) temp[i]).getTypeName();
                type = ASTUtils.parseTypeName(t);
                name = t.getIdentifier().toString().trim();
                f.addParam(name,type);
            }
        }
        mVtable.put(fName,f);
        IMTrace.out(f.toString());
    }

    public void caseAConstDecDeclaration(AConstDecDeclaration node)
    {
        IMTrace.out("AVariableDeclaration: " + node.getSemicolon().getLine());

        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getSemicolon().getLine());

        String varName,type;
        varName = node.getIdentifier().toString().trim();
        type    = node.getType().toString().trim();

        IMTrace.out("found constant dec: " + varName);
        try{
            if(node.getInitialiser()!=null)
            {
                node.getInitialiser().apply(this);
                setVariable(varName,getValue(node.getInitialiser()));
                cleanValue(node.getInitialiser());
            }else
                setVariable(varName,initVariable(type));
        }catch(ValueException ae){
            throwError(ae.getMessage(),node.getSemicolon().getLine());
        }
    }



    public void caseAVarDecDeclaration(AVarDecDeclaration node)
    {
        IMTrace.out("AVariableDeclaration: " + node.getSemicolon().getLine());

        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getSemicolon().getLine());


        String varName, type;

        varName = node.getIdentifier().toString().trim();
        type    = node.getType().toString().trim();

        try
        {
            IMTrace.out("found scalar dec: " + varName);
            if(node.getInitialiser()!=null)
            {
                node.getInitialiser().apply(this);
                setVariable(varName,getValue(node.getInitialiser()));
                cleanValue(node.getInitialiser());
            }
            // initialise 'references' to classes within classes to null
            else if(mInClass && !isPrimitiveType(type)){
                setVariable(varName,ValueFactory.makeNull(type,varName));
            }else {
                setVariable(varName,initVariable(type));
            }
        }catch(ValueException ae){
            throwError(ae.getMessage(),node.getSemicolon().getLine());
        }
    }

    public void caseAArrayDecDeclaration(AArrayDecDeclaration node)
    {
        String varName = node.getIdentifier().toString().trim(),
               type    = node.getType().toString().trim();

        IMTrace.out("Array declaration found: " + varName);
        if(!mPreProcessMode && mExecutionManager!=null)
            mExecutionManager.checkHalt(node.getSemicolon().getLine());

        // get the number of dimensions the array is declared as having.
        int numDeclaredDims = node.getBracketPair().size();

        if(numDeclaredDims > ArrayDim.MAX_SIZE){
            throwError("The current interpreter implementation does" +
              " not support arrays with more than " + ArrayDim.MAX_SIZE
              + " dimensions.",node.getSemicolon().getLine());
        }

        Value v = null;

        if(node.getArrayInitialiser()!=null){
            node.getArrayInitialiser().apply(this);
            v = getValue(node.getArrayInitialiser());
            cleanValue(node.getArrayInitialiser());
        }else{
            v = new ArrayValue(type);
            v.setVarName(varName);
            IMTrace.out("Making empty array of type " + type);
        }

        IMTrace.out("v in array declaration is null: " + (v==null));

        try{
            if(mInClass) mClassTemp.put(varName,v); // class dec
            else setVariable(varName,v);         // normal dec
        }catch(ValueException ae){
            throwError(ae.getMessage(),node.getSemicolon().getLine());
        }
    }

    /* ************************************************************

    initialiser = assign expression;

    array_initialiser =
      {static} assign array_init |
      {dynamic} assign array_allocate;

    array_init = l_brace init_list r_brace;

    init_list =
      {scalar} expression comma_exp* |
      {array}  array_init comma_array_init*;

    comma_exp = comma expression;

    comma_array_init = comma array_init;

    **************************************************************/

    // scalar initialiser
    public void caseAInitialiser(AInitialiser node)
    {
        node.getExpression().apply(this);
        setValue(node,getValue(node.getExpression()));
        cleanValue(node.getExpression());
    }

    public void caseAStaticArrayInitialiser(AStaticArrayInitialiser node)
    {
        node.getArrayInit().apply(this);
        setValue(node,getValue(node.getArrayInit()));
        cleanValue(node.getArrayInit());
    }

    public void caseADynamicArrayInitialiser(ADynamicArrayInitialiser node)
    {
        node.getArrayAllocate().apply(this);
        setValue(node,getValue(node.getArrayAllocate()));
        cleanValue(node.getArrayAllocate());
    }

    public void caseAArrayInit(AArrayInit node)
    {
        node.getInitList().apply(this);
        setValue(node,getValue(node.getInitList()));
        cleanValue(node.getInitList());
    }

    public void caseAScalarInitList(AScalarInitList node)
    {
        Vector data = new Vector();

        if ( node.getExpression() != null )
        {
            node.getExpression().apply( this );
            data.add(getValue(node.getExpression()));
            cleanValue(node.getExpression());
        }


        if ( node.getCommaExp() != null )
        {
            LinkedList commaexps = node.getCommaExp();
            ListIterator ceIterator = commaexps.listIterator(0);

            while( ceIterator.hasNext() ){
                Node next = (Node)ceIterator.next();
                next.apply( this );
                data.add(getValue(next));
                cleanValue(next);
            }
        }

        Value[] _data = new Value[data.size()];
        for(int i=0;i<data.size();i++) _data[i] = (Value)data.get(i);
        setValue(node,new ArrayValue(_data,_data[0].getType()));
    }

    public void caseAArrayInitList(AArrayInitList node)
    {
        Vector data = new Vector();

        if ( node.getArrayInit() != null )
        {
            node.getArrayInit().apply( this );
            data.add(getValue(node.getArrayInit()));
            cleanValue(node.getArrayInit());
        }


        if ( node.getCommaArrayInit() != null )
        {
            LinkedList commaarrayinits = node.getCommaArrayInit();
            ListIterator ceIterator = commaarrayinits.listIterator(0);

            while( ceIterator.hasNext() ){
                Node next = (Node)ceIterator.next();
                next.apply( this );
                data.add(getValue(next));
                cleanValue(next);
            }
        }

        Value[] _data = new Value[data.size()];
        for(int i=0;i<data.size();i++) _data[i] = (Value)data.get(i);
        setValue(node,new ArrayValue(_data,_data[0].getType()));
        IMTrace.out("In caseAArrayInitList, size = " + _data.length);
    }

    public void caseACommaExp(ACommaExp node)
    {
        node.getExpression().apply(this);
        setValue(node,getValue(node.getExpression()));
        cleanValue(node.getExpression());
    }

    public void caseACommaArrayInit(ACommaArrayInit node)
    {
        node.getArrayInit().apply(this);
        setValue(node,getValue(node.getArrayInit()));
        cleanValue(node.getArrayInit());
    }

    /* ************************************************************

    booleanliteral =
        {true} true |
        {false} false;

    **************************************************************/

    public void caseATrueBooleanliteral(ATrueBooleanliteral node)
    {
        setValue(node,ValueFactory.makeValue(true));
    }

    public void caseAFalseBooleanliteral(AFalseBooleanliteral node)
    {
        setValue(node,ValueFactory.makeValue(false));
    }

    /* ************************************************************

    type =
    {basic_type}     basic_type |
    {reference_type} reference_type;

    **************************************************************/

    public void caseABasicTypeType(ABasicTypeType node)
    {
        node.getBasicType().apply(this);
        setValue(node,getValue(node.getBasicType()));
    }


    /* ************************************************************

    basic_type =
    {char}    char |
    {int}     int |
    {real}    real |
    {string}  string |
    {boolean} boolean |
    {void}    void;

    **************************************************************/

    public void caseACharBasicType(ACharBasicType node)
    {
        setValue(node,ValueFactory.makeValue(node.getChar().getText()));
    }

    public void caseAIntBasicType(AIntBasicType node)
    {
        setValue(node,ValueFactory.makeValue(node.getInt().getText()));
    }

    public void caseARealBasicType(ARealBasicType node)
    {
        setValue(node,ValueFactory.makeValue(node.getReal().getText()));
    }

    public void caseAStringBasicType(AStringBasicType node)
    {
        setValue(node,ValueFactory.makeValue(node.getString().getText()));
    }

    public void caseABooleanBasicType(ABooleanBasicType node)
    {
        setValue(node,ValueFactory.makeValue(node.getBoolean().getText()));
    }

    public void caseAVoidBasicType(AVoidBasicType node)
    {
        setValue(node,ValueFactory.makeValue(node.getVoid().getText()));
    }

    /* ************************************************************
    expression =
      {boolexp} bool_expression |
      {array_int} array_allocate;

    array_allocate = new type array_access+
    **************************************************************/

    public void caseABoolexpExpression( ABoolexpExpression node)
    {
        node.getBoolExpression().apply(this);
        setValue(node, getValue(node.getBoolExpression()));
        //cleanValue(node.getExpression());
    }

    public void caseAArrayIntExpression(AArrayIntExpression node)
    {
        node.getArrayAllocate().apply(this);
        setValue(node,getValue(node.getArrayAllocate()));
        cleanValue(node.getArrayAllocate());
    }

    public void caseAArrayAllocate(AArrayAllocate node)
    {
        if(node.getArrayAccess().size()>ArrayDim.MAX_SIZE){
            throwError("The current interpreter implementation does" +
                      " not support arrays with more than " + ArrayDim.MAX_SIZE
                      + " dimensions.",node.getNew().getLine());
        }

        String type = node.getType().toString().trim();
        ArrayDim a = new ArrayDim();
        String tmp;
        Object temp[] = node.getArrayAccess().toArray();
        for(int i = 0; i < temp.length; i++)
        {
            ((PArrayAccess) temp[i]).apply(this);
            tmp = getValue((Node)temp[i]).toString();
            tmp = tmp.substring(tmp.indexOf('[')+1,tmp.indexOf(']'));
            a.addDim(Integer.parseInt(tmp));
        }
        try{
            Value array = setupArray(a,type); // 'allocate' mem for array
            setValue(node,array);
        }catch(ValueException ae){
            throwError(ae.getMessage(),node.getNew().getLine());
        }
    }


    /* ************************************************************

    math_expression =
    {term} term |
    {plus} math_expression plus term |
    {minus} math_expression minus term ;

    **************************************************************/


    // math expression term
    public void caseATermMathExpression(ATermMathExpression node)
    {
        node.getTerm().apply(this);
        setValue(node, getValue(node.getTerm()));
    }

    // plus expression
    public void caseAPlusMathExpression( APlusMathExpression node )
    {
        // evaluate LHS of plus
        node.getMathExpression().apply(this);
        // evaluate RHS of plus
        node.getTerm().apply(this);

        try
        {
            Value LHS = getValue(node.getMathExpression()),
                  RHS = getValue(node.getTerm());

            if(LHS.isInt() && RHS.isInt())
            {
                int result = Value.getInt(LHS) + Value.getInt(RHS);
                setValue(node,ValueFactory.makeValue(result));
            }
            else if(LHS.isNumber() && RHS.isNumber())
            {
                double result = Value.getNumber(LHS) + Value.getNumber(RHS);
                setValue(node,ValueFactory.makeValue(result));
            }
            else // not both operands are numbers, assume string concat
            {
                setValue(node,
                  ValueFactory.makeValue(LHS.toString() + RHS.toString()));
            }
        }
        catch(Exception e)
        {
            throwError("One of the operands in addition is not a number.",
              node.getPlus().getLine());
        }
    }

    // minus expression
    public void caseAMinusMathExpression( AMinusMathExpression node )
    {
        // evaluate LHS of subtract
        node.getMathExpression().apply(this);
        // evaluate RHS of subtract
        node.getTerm().apply(this);

        try
        {
            if(getValue(node.getMathExpression()).isInt() &&
               getValue(node.getTerm()).isInt())
            {
                int result =
                  ((IntValue)getValue(node.getMathExpression())).getInt() -
                  ((IntValue)getValue(node.getTerm())).getInt();
                setValue(node,ValueFactory.makeValue(result));
            }else{
                double result =
                  ((DoubleValue)getValue(node.getMathExpression())).getNumber()
                   - ((DoubleValue)getValue(node.getTerm())).getNumber();
                setValue(node,ValueFactory.makeValue(result));
            }
        }
        catch(Exception e)
        {
            throwError("One of the operands in subtraction is not a number.",
              node.getMinus().getLine());
        }
    }

    /* ************************************************************

    term =
    {unary} unary_exp |
    {mult} term times unary_exp |
    {div} term divide unary_exp |
    {mod} term mod unary_exp |
    {power} term power unary_exp;

    **************************************************************/

    public void caseAUnaryTerm(AUnaryTerm node)
    {
        node.getUnaryExp().apply(this);
        setValue(node,getValue(node.getUnaryExp()));
    }

    public void caseAMultTerm(AMultTerm node)
    {
        node.getTerm().apply(this);
        node.getUnaryExp().apply(this);
        try
        {
            double result =
              ((NumberValue)getValue(node.getTerm())).getNumber() *
              ((NumberValue)getValue(node.getUnaryExp())).getNumber();

            if(getValue(node.getUnaryExp()).isInt() &&
               getValue(node.getTerm()).isInt())
            {
                setValue(node,ValueFactory.makeValue((int)result));
            }else{
                setValue(node,ValueFactory.makeValue(result));
            }
        }
        catch(ClassCastException e)
        {
            throwError("One of the operands in multiplication is not a number.",
              node.getTimes().getLine());
        }
    }

    public void caseADivTerm(ADivTerm node)
    {
        node.getTerm().apply(this);
        node.getUnaryExp().apply(this);
        try
        {
            double result =
                ((NumberValue)getValue(node.getTerm())).getNumber() /
                ((NumberValue)getValue(node.getUnaryExp())).getNumber();

            if(getValue(node.getUnaryExp()).isInt() &&
               getValue(node.getTerm()).isInt())
            {
                setValue(node,ValueFactory.makeValue((int)result));
            }else{
                setValue(node,ValueFactory.makeValue(result));
            }
        }
        catch(ClassCastException e)
        {
            throwError("One of the operands in division is not a number.",
              node.getDivide().getLine());
        }
    }

    public void caseAModTerm(AModTerm node)
    {
        node.getTerm().apply(this);
        node.getUnaryExp().apply(this);
        try
        {
            double result =
              ((NumberValue)getValue(node.getTerm())).getNumber() %
              ((NumberValue)getValue(node.getUnaryExp())).getNumber();

            if(getValue(node.getUnaryExp()).isInt() &&
               getValue(node.getTerm()).isInt())
            {
                setValue(node,ValueFactory.makeValue((int)result));
            }else{
                setValue(node,ValueFactory.makeValue(result));
            }
        }
        catch(ClassCastException e)
        {
            throwError("One of the operands in mod is not a number.",
              node.getMod().getLine());
        }
    }

    public void caseAPowerTerm(APowerTerm node)
    {
        node.getTerm().apply(this);
        node.getUnaryExp().apply(this);
        try
        {
            double result =
              Math.pow(((NumberValue)getValue(node.getTerm())).getNumber(),
                  ((NumberValue)getValue(node.getUnaryExp())).getNumber());

            if(getValue(node.getUnaryExp()).isInt() &&
               getValue(node.getTerm()).isInt())
            {
                setValue(node,ValueFactory.makeValue((int)result));
            }else{
                setValue(node,ValueFactory.makeValue(result));
            }
        }
        catch(ClassCastException e)
        {
            throwError("One of the operands in power expression isn't numeric.",
              node.getPower().getLine());
        }
    }


    /* ************************************************************

    unary_exp =
    {preincr} preincr |
    {predecr} predecr |
    {minus} minus unary_exp |
    {plus} plus unary_exp |
    {negate} not unary_exp |
    {postfix_exp} postfix_exp;

    **************************************************************/

    public void caseAPreincrUnaryExp(APreincrUnaryExp node)
    {
        node.getPreincr().apply(this);
        setValue(node, getValue(node.getPreincr()));
    }

    public void caseAPredecrUnaryExp(APredecrUnaryExp node)
    {
        node.getPredecr().apply(this);
        setValue(node, getValue(node.getPredecr()));
    }

    public void caseAMinusUnaryExp(AMinusUnaryExp node) //throws KInterException
    {
        node.getUnaryExp().apply(this);
        Value result = getValue(node.getUnaryExp());

        try{
            if(result.isInt()){
                setValue(node,
                  ValueFactory.makeValue(((IntValue)result).getInt() * -1));
            }
            else{
                setValue(node,
                  ValueFactory.makeValue(((NumberValue)result).getNumber() * -1));
            }
        }
        catch(Exception e)
        {
            throwError("Operand in negation expression is not a number.",
              node.getMinus().getLine());
        }
    }

    public void caseAPlusUnaryExp(APlusUnaryExp node) //throws KInterException
    {
        node.getUnaryExp().apply(this);
        Value result = getValue(node.getUnaryExp());

        try{
            if(result.isInt()){
                setValue(node,
                  ValueFactory.makeValue(Math.abs(((IntValue)result).getInt())));
            }
            else{
                setValue(node,
                  ValueFactory.makeValue(Math.abs(((NumberValue)result).getNumber())));
            }
        }
        catch(Exception e)
        {
           throwError("Operand in abs (or is concat?) is not a number.",
              node.getPlus().getLine());
        }
    }

    public void caseANegateUnaryExp(ANegateUnaryExp node)
    {
        node.getUnaryExp().apply(this);
        try{
            boolean r = ((BoolValue)getValue(node.getUnaryExp())).getBool();
            setValue(node,ValueFactory.makeValue(!r));
        }
        catch(ClassCastException e){
            throwError("Tried to negate a non boolean value.",
              node.getNot().getLine());
        }
    }

    public void caseAPostfixExpUnaryExp(APostfixExpUnaryExp node)
    {
        node.getPostfixExp().apply(this);
        setValue(node, getValue(node.getPostfixExp()));
    }

    /* ************************************************************

    postfix_exp =
    {factor} factor |
    {postincr} postincr |
    {postdecr} postdecr;

    **************************************************************/

    public void caseAFactorPostfixExp(AFactorPostfixExp node)
    {
        node.getFactor().apply(this);
        setValue(node,getValue(node.getFactor()));
    }

    public void caseAPostincrPostfixExp(APostincrPostfixExp node)
    {
        node.getPostincr().apply(this);
        setValue(node,getValue(node.getPostincr()));
    }

    public void caseAPostdecrPostfixExp(APostdecrPostfixExp node)
    {
        node.getPostdecr().apply(this);
        setValue(node,getValue(node.getPostdecr()));
    }

    /* postdecr = postfix_exp minusminus *************************/
    public void caseAPostdecr(APostdecr node)
    {
        Value v;

        // evaluate the postfix expression
        node.getPostfixExp().apply(this);
        v = getValue(node.getPostfixExp());
        try{
            if(v.isInt())
            {
                int i = ((IntValue)v).getInt();
                setValue(node,ValueFactory.makeValue(i));
                // if we are incrementing a var, update it.
                if(v.getVarName()!=null){
                    //((IntValue)getVariable(v.getVarName())).setInt(--i);
                    setVariable(v.getVarName(), ValueFactory.makeValue(--i));
                }
            }
            else // assume double
            {
                double d = ((DoubleValue)v).getDouble();
                setValue(node,ValueFactory.makeValue(d));
                // if we are incrementing a var, update it.
                if(v.getVarName()!=null){
                    //((DoubleValue)getVariable(v.getVarName())).setDouble(--d);
                    setVariable(v.getVarName(),ValueFactory.makeValue(--d));
                }
            }
        }
        catch(ValueException ae){
            throwError(ae.getMessage(),node.getMinusminus().getLine());
        }
        catch(ClassCastException e3){
            throwError("Operand in post decrement is not a number.",
              node.getMinusminus().getLine());
        }
    }

    /* postincr = postfix_exp plusplus ***************************/
    public void caseAPostincr(APostincr node)
    {
        Value v;

        // evaluate the postfix expression
        node.getPostfixExp().apply(this);
        v = getValue(node.getPostfixExp());
        try{
            if(v.isInt())
            {
                int i = ((IntValue)v).getInt();
                setValue(node,ValueFactory.makeValue(i));
                // if we are incrementing a var, update it.
                if(v.getVarName()!=null){
                    //((IntValue)getVariable(v.getVarName())).setInt(++i);
                    setVariable(v.getVarName(),ValueFactory.makeValue(++i));
                }
            }
            else // assume double
            {
                double d = ((DoubleValue)v).getDouble();
                setValue(node,ValueFactory.makeValue(d));
                // if we are incrementing a var, update it.
                if(v.getVarName()!=null){
                    //((DoubleValue)getVariable(v.getVarName())).setDouble(++d);
                    setVariable(v.getVarName(),ValueFactory.makeValue(++d));
                }
            }
        }
        catch(ValueException ae){
            throwError(ae.getMessage(),node.getPlusplus().getLine());
        }
        catch(ClassCastException e3){
            throwError("Operand in post increment is not a number.",
              node.getPlusplus().getLine());
        }
    }

    /* predecr = minusminus postfix_exp **************************/
    public void caseAPredecr(APredecr node)
    {
        Value v;

        // evaluate the postfix expression
        node.getPostfixExp().apply(this);
        v = getValue(node.getPostfixExp());
        try{
            if(v.isInt())
            {
                int i = ((IntValue)v).getInt();
                setValue(node,ValueFactory.makeValue(--i));

                // if we are incrementing a var, update it.
                if(v.getVarName()!=null){
                    setVariable(v.getVarName(),
                      ValueFactory.makeValue(v.getVarName(),i));
                }
            }
            else // assume double
            {
                double d = ((DoubleValue)v).getDouble();
                setValue(node,ValueFactory.makeValue(--d));

                // if we are incrementing a var, update it.
                if(v.getVarName()!=null){
                    setVariable(v.getVarName(),
                      ValueFactory.makeValue(v.getVarName(),d));
                }
            }
        }
        catch(ValueException ae){
            throwError(ae.getMessage(),node.getMinusminus().getLine());
        }
        catch(ClassCastException e3){
            throwError("Operand in pre decrement is not a number",
            node.getMinusminus().getLine());
        }
    }

    /* postincr = plusplus postfix_exp ***************************/
    public void caseAPreincr(APreincr node)
    {
        Value v;

        // evaluate the postfix expression
        node.getPostfixExp().apply(this);
        v = getValue(node.getPostfixExp());
        try{
            if(v.isInt()) // it's an int!
            {
                int i = ((IntValue)v).getInt();
                setValue(node,ValueFactory.makeValue(++i));

                // if we are incrementing a var, update it.
                if(v.getVarName()!=null){
                    setVariable(v.getVarName(),
                      ValueFactory.makeValue(v.getVarName(),i));
                }
            }
            else // no it's not! assume double...
            {
                double d = ((DoubleValue)v).getDouble();
                setValue(node,ValueFactory.makeValue(++d));

                // if we are incrementing a var, update it.
                if(v.getVarName()!=null){
                    setVariable(v.getVarName(),
                      ValueFactory.makeValue(v.getVarName(),d));
                }
            }
        }
        catch(ValueException ae){
            throwError(ae.getMessage(),node.getPlusplus().getLine());
        }
        catch(ClassCastException e3){
            throwError("Operand in pre increment is not a number.",
              node.getPlusplus().getLine());
        }
    }

    /* ************************************************************

    factor =
    {null} null |
    {fieldaccess} field_access |
    {function} function_application |
    {number} number |
    {stringliteral} stringliteral |
    {charliteral} charliteral |
    {boolliteral} booleanliteral |
    {readint} readint |
    {readreal} readreal |
    {readstring} readstring |
    {readchar} readchar |
    {read} read |
    {endoffile} endoffile |
    {sin} sin l_parenthese math_expression r_parenthese |
    {cos} cos l_parenthese math_expression r_parenthese |
    {tan} tan l_parenthese math_expression r_parenthese |
    {asin} asin l_parenthese math_expression r_parenthese |
    {acos} acos l_parenthese math_expression r_parenthese |
    {atan} atan l_parenthese math_expression r_parenthese |
    {sqrt} sqrt l_parenthese math_expression r_parenthese |
    {log} log l_parenthese math_expression r_parenthese |
    {exp} exp l_parenthese math_expression r_parenthese |
    {abs} abs l_parenthese math_expression r_parenthese |
    {floor} floor l_parenthese math_expression r_parenthese |
    {ceil} ceil l_parenthese math_expression r_parenthese |
    {round} round l_parenthese math_expression r_parenthese |
    {random} random l_parenthese r_parenthese |
    {expression} l_parenthese bool_expression r_parenthese;

    **************************************************************/

    // null literal
    public void caseANullFactor(ANullFactor node)
    {
        setValue(node,new NullValue());
    }


    // a variable access
    public void caseAFieldaccessFactor(AFieldaccessFactor node)
    {
        try{
            node.getFieldAccess().apply(this);
            // get value of the variable
            Value v = getVariable(getValue(node.getFieldAccess()).toString());
            // keep a note of the variable name... may be useful later...
            v.setVarName(getValue(node.getFieldAccess()).toString());
            setValue(node,v);
            cleanValue(node.getFieldAccess());
        }catch(ValueException ae){
            throwError(ae.getMessage(),getLineNum(node.getFieldAccess()));
        }
    }

    /**
     * utility method for returning the line number of a field access. USed
     * for producing better error messages usually.
     */
    private int getLineNum(PFieldAccess node)
    {
        if(node instanceof AArrayFieldAccess)
        {
            return getLineNumFromName(((AArrayFieldAccess)node).getName());
        }else{
            return getLineNumFromName(((AScalarFieldAccess)node).getName());
        }
    }

    private int getLineNumFromName(PName name)
    {
        if(name instanceof ASimpleNameName)
        {
            PSimpleName p = ((ASimpleNameName)name).getSimpleName();
            return ((ASimpleName) p).getIdentifier().getLine();
        }else{
            PQualifiedName ka = ((AQualifiedNameName)name).getQualifiedName();
            return ((AQualifiedName)ka).getIdentifier().getLine();
        }
    }


    // function call
    public void caseAFunctionFactor(AFunctionFactor node)
    {
        node.getFunctionApplication().apply(this);
        setValue(node,getValue(node.getFunctionApplication()));
        cleanValue(node.getFunctionApplication());
    }

    // a number literal
    public void caseANumberFactor(ANumberFactor node)
    {
        node.getNumber().apply(this);
        setValue(node, getValue(node.getNumber()));
    }

    // a string literal
    public void caseAStringliteralFactor(AStringliteralFactor node)
    {
        int ln = node.getStringliteral().getLine();
        setValue(node,ValueFactory.makeValue(
            escapeString(
             stripQuotes(node.getStringliteral().toString()),ln)));
    }

    // a character literal
    public void caseACharliteralFactor(ACharliteralFactor node)
    {
        String s = node.toString().trim();
        int ln = node.getCharliteral().getLine();

        s = escapeString(s.substring(1,s.length()-1),ln);
        if(s.length()> 1)
            throwError("Invalid escape sequence",ln);
        else
            setValue(node,ValueFactory.makeValue(s.charAt(0)));
    }

    // a boolean literal
    public void caseABoolliteralFactor(ABoolliteralFactor node)
    {
        node.getBooleanliteral().apply(this);
        setValue(node,getValue(node.getBooleanliteral()));
    }

    public void caseAReadintFactor(AReadintFactor node)
    {
        setValue(node,ValueFactory.makeValue(mHandler.readInt()));
    }

    public void caseAReadrealFactor(AReadrealFactor node)
    {
        setValue(node,ValueFactory.makeValue(mHandler.readDouble()));
    }

    public void caseAReadstringFactor(AReadstringFactor node)
    {
        setValue(node,ValueFactory.makeValue(mHandler.readString()));
    }

    public void caseAReadcharFactor(AReadcharFactor node)
    {
        setValue(node,ValueFactory.makeValue(mHandler.readChar()));
    }

    public void caseAReadFactor(AReadFactor node)
    {
        setValue(node,ValueFactory.makeValue(mHandler.read()));
    }

    public void caseAEndoffileFactor(AEndoffileFactor node)
    {
        setValue(node,ValueFactory.makeValue(mHandler.isEOF()));
    }

    public void caseASinFactor(ASinFactor node)
    {
        node.getMathExpression().apply(this);
        try{
            double r =
              ((NumberValue)getValue(node.getMathExpression())).getNumber();
            setValue(node,ValueFactory.makeValue(Math.sin(r)));
        }
        catch(ClassCastException e){
            throwError("Tried to evaluate sin of non numeric expression.",
              node.getSin().getLine());
        }
    }

    public void caseACosFactor(ACosFactor node)
    {
        node.getMathExpression().apply(this);
        try{
            double r =
              ((NumberValue)getValue(node.getMathExpression())).getNumber();
            setValue(node,ValueFactory.makeValue(Math.cos(r)));
        }
        catch(ClassCastException e){
            throwError("Tried to evaluate cos of non numeric" +
              " expression.",node.getCos().getLine());
        }
    }

    public void caseATanFactor(ATanFactor node)
    {
        node.getMathExpression().apply(this);
        try{
            double r =
              ((NumberValue)getValue(node.getMathExpression())).getNumber();
            setValue(node,ValueFactory.makeValue(Math.tan(r)));
        }
        catch(ClassCastException e){
            throwError("Tried to evaluate tan of non numeric" +
              " expression.",node.getTan().getLine());
        }
    }

    public void caseAAsinFactor(AAsinFactor node)
    {
        node.getMathExpression().apply(this);
        try{
            double r =
              ((NumberValue)getValue(node.getMathExpression())).getNumber();
            setValue(node,ValueFactory.makeValue(Math.asin(r)));
        }
        catch(ClassCastException e){
            throwError("Tried to evaluate asin of non numeric" +
              " expression.",node.getAsin().getLine());
        }
    }

    public void caseAAcosFactor(AAcosFactor node)
    {
        node.getMathExpression().apply(this);
        try{
            double r =
              ((NumberValue)getValue(node.getMathExpression())).getNumber();
            setValue(node,ValueFactory.makeValue(Math.acos(r)));
        }
        catch(ClassCastException e){
            throwError("Tried to evaluate acos of non numeric" +
              " expression.",node.getAcos().getLine());
        }
    }

    public void caseAAtanFactor(AAtanFactor node)
    {
        node.getMathExpression().apply(this);
        try{
            double r =
              ((NumberValue)getValue(node.getMathExpression())).getNumber();
            setValue(node,ValueFactory.makeValue(Math.atan(r)));
        }
        catch(ClassCastException e){
            throwError("Tried to evaluate atan of non numeric" +
              " expression.",node.getAtan().getLine());
        }
    }

    public void caseASqrtFactor(ASqrtFactor node)
    {
        node.getMathExpression().apply(this);
        try{
            double r =
              ((NumberValue)getValue(node.getMathExpression())).getNumber();
            setValue(node,ValueFactory.makeValue(Math.sqrt(r)));
        }
        catch(ClassCastException e){
            throwError("Tried to evaluate sqrt of non numeric" +
              " expression.",node.getSqrt().getLine());
        }
    }

    public void caseALogFactor(ALogFactor node)
    {
        node.getMathExpression().apply(this);
        try{
            double r =
              ((NumberValue)getValue(node.getMathExpression())).getNumber();
            setValue(node,ValueFactory.makeValue(Math.log(r)));
        }
        catch(ClassCastException e){
            throwError("Tried to evaluate log of non numeric" +
              " expression.",node.getLog().getLine());
        }
    }

    public void caseAExpFactor(AExpFactor node)
    {
        node.getMathExpression().apply(this);
        try{
            double r =
              ((NumberValue)getValue(node.getMathExpression())).getNumber();
            setValue(node,ValueFactory.makeValue(Math.exp(r)));
        }
        catch(ClassCastException e){
            throwError("Tried to evaluate exp of non numeric" +
              " expression.",node.getExp().getLine());
        }
    }

    public void caseAAbsFactor(AAbsFactor node)
    {
        node.getMathExpression().apply(this);
        try{
            double r =
              ((NumberValue)getValue(node.getMathExpression())).getNumber();
            setValue(node,ValueFactory.makeValue(Math.abs(r)));
        }
        catch(ClassCastException e){
            throwError("Tried to evaluate log of non numeric" +
              " expression.",node.getAbs().getLine());
        }
    }

    public void caseAFloorFactor(AFloorFactor node)
    {
        node.getMathExpression().apply(this);
        try{
            double r =
              ((NumberValue)getValue(node.getMathExpression())).getNumber();
            setValue(node,ValueFactory.makeValue(Math.floor(r)));
        }
        catch(ClassCastException e){
            throwError("Tried to evaluate floor of non numeric" +
              " expression.",node.getFloor().getLine());
        }
    }

    public void caseACeilFactor(ACeilFactor node)
    {
        node.getMathExpression().apply(this);
        try{
            double r =
              ((NumberValue)getValue(node.getMathExpression())).getNumber();
            setValue(node,ValueFactory.makeValue(Math.ceil(r)));
        }
        catch(ClassCastException e){
            throwError("Tried to evaluate ceiling of non numeric" +
              " expression.",node.getCeil().getLine());
        }
    }

    public void caseARoundFactor(ARoundFactor node)
    {
        node.getMathExpression().apply(this);
        try{
            double r =
              ((NumberValue)getValue(node.getMathExpression())).getNumber();
            setValue(node,ValueFactory.makeValue(Math.round(r)));
        }
        catch(ClassCastException e){
            throwError("Tried to round a non numeric" +
              " expression.",node.getRound().getLine());
        }
    }

    public void caseARandomFactor(ARandomFactor node)
    {
        setValue(node,ValueFactory.makeValue(Math.random()));
    }

    public void caseAExpressionFactor(AExpressionFactor node)
    {
        node.getBoolExpression().apply(this);
        setValue(node,getValue(node.getBoolExpression()));
    }

    /* ************************************************************

    number = {i} intnumber |
             {f} fpnumber;

    **************************************************************/

    // an integer constant
    public void caseAINumber(AINumber node)
    {
        try{
            //Integer i = new Integer(node.getIntnumber().getText());
            setValue(node,
              ValueFactory.makeValue(
                Integer.parseInt(node.getIntnumber().getText()))); // set value
        }
        catch(NumberFormatException e){
            throwError("Number error in line ",node.getIntnumber().getLine());
        }
    }

    // a floating point constant
    public void caseAFNumber(AFNumber node)
    {
        try{
            setValue(node,ValueFactory.makeValue(
                Double.parseDouble(node.getFpnumber().getText()))); // set value
        }
        catch(NumberFormatException e){
            throwError("Number error in line ",node.getFpnumber().getLine());
        }
    }

    /* ************************************************************

    bool_expression =
        {term} bool_term |
        {or} bool_expression or bool_term |
        {xor} bool_expression xor bool_term;

    **************************************************************/

    public void caseATermBoolExpression( ATermBoolExpression node)
    {
        node.getBoolTerm().apply(this);
        setValue(node, getValue(node.getBoolTerm()));
    }

    public void caseAOrBoolExpression(AOrBoolExpression node)
    {
        node.getBoolExpression().apply(this);
        //node.getBoolTerm().apply(this);
        // short circuit implementation
        try
        {
            boolean LHS =
              ((BoolValue)getValue(node.getBoolExpression())).getBool();
            if(LHS) setValue(node,ValueFactory.makeValue(true));
            else{
                try{
                    node.getBoolTerm().apply(this);
                    boolean RHS =
                      ((BoolValue)getValue(node.getBoolTerm())).getBool();
                    setValue(node,ValueFactory.makeValue(LHS || RHS));
                }catch(ClassCastException e2){
                    throwError("RHS of OR Op is not a boolean.",
                      node.getOr().getLine());
                }
            }
        }catch(ClassCastException e1){
            throwError("Error! LHS of OR Op is not a boolean.",
              node.getOr().getLine());
        }
    }

    public void caseAXorBoolExpression(AXorBoolExpression node)
    {
        node.getBoolExpression().apply(this);
        node.getBoolTerm().apply(this);

        try
        {
            boolean LHS =
              ((BoolValue)getValue(node.getBoolExpression())).getBool();
            try{
                node.getBoolTerm().apply(this);
                boolean RHS =
                  ((BoolValue)getValue(node.getBoolTerm())).getBool();
                setValue(node,ValueFactory.makeValue(LHS ^ RHS));
            }catch(ClassCastException e2){
                throwError("RHS of XOR Op is not a boolean.",
                  node.getXor().getLine());
            }

        }catch(ClassCastException e1){
            throwError("LHS of XOR Op is not a boolean.",
              node.getXor().getLine());
        }
    }

    /* ************************************************************

     bool_term =
       {equality} equality |
       {and} bool_term and equality;

    **************************************************************/

    public void caseAEqualityBoolTerm( AEqualityBoolTerm node)
    {
        node.getEquality().apply(this);
        setValue(node, getValue(node.getEquality()));
    }

    public void caseAAndBoolTerm(AAndBoolTerm node)
    {
        node.getBoolTerm().apply(this);
        //node.getEquality().apply(this);
        // short circuit implementation
        try
        {
            boolean LHS =
              ((BoolValue)getValue(node.getBoolTerm())).getBool();
            if(!LHS)
                setValue(node,ValueFactory.makeValue(false));
            else
            {
                try{
                    node.getEquality().apply(this);
                    boolean RHS =
                      ((BoolValue)getValue(node.getEquality())).getBool();
                    setValue(node,ValueFactory.makeValue(LHS && RHS));
                }catch(ClassCastException e2){
                    throwError("RHS of And Op is not a boolean.",
                      node.getAnd().getLine());
                }
            }
        }catch(ClassCastException e1){
            throwError("LHS of And Op is not a boolean.",
              node.getAnd().getLine());
        }
    }

    /* ************************************************************

    equality =
        {relational} relational |
        {eq} equality equal relational |
        {neq} equality noteqal relational;

    **************************************************************/

    public void caseARelationalEquality( ARelationalEquality node)
    {
        node.getRelational().apply(this);
        setValue(node, getValue(node.getRelational()));
    }

    public void caseAEqEquality(AEqEquality node)
    {
        node.getEquality().apply(this);
        node.getRelational().apply(this);

        Value LHS = getValue(node.getEquality()),
              RHS = getValue(node.getRelational());

        try{ // equality means different things in different contexts
            if(LHS.isNumber() && RHS.isNumber())
                setValue(node, ValueFactory.makeValue(
                    Value.getNumber(LHS) == Value.getNumber(RHS)));
            else if(LHS.isBool() && RHS.isBool())
                setValue(node, ValueFactory.makeValue(
                    Value.getBool(LHS) == Value.getBool(RHS)));
            else if(LHS.isString() && RHS.isString())
                setValue(node, ValueFactory.makeValue(
                    Value.getString(LHS).equals(Value.getString(RHS))));
            else if(LHS.isChar() && RHS.isChar())
                setValue(node, ValueFactory.makeValue(
                    Value.getChar(LHS) == Value.getChar(RHS)));
            else if(LHS.getType().equals(RHS.getType()))
                setValue(node,ValueFactory.makeValue(LHS==RHS));
            else{
                throwError("Operator == cannot be applied" +
                   " to " + LHS.getType() + "," + RHS.getType(),
                  node.getEqual().getLine());
            }
        }catch(ValueException e){
            throwError("Failed to retrieve one of the" +
              " operands for == operation.", node.getEqual().getLine());
        }
    }

    public void caseANeqEquality(ANeqEquality node)
    {
        node.getEquality().apply(this);
        node.getRelational().apply(this);

        Value LHS = getValue(node.getEquality()),
              RHS = getValue(node.getRelational());

        try{
            if(LHS.isNumber() && RHS.isNumber())
                setValue(node, ValueFactory.makeValue(
                    Value.getNumber(LHS) != Value.getNumber(RHS)));
            else if(LHS.isBool() && RHS.isBool())
                setValue(node, ValueFactory.makeValue(
                    Value.getBool(LHS) != Value.getBool(RHS)));
            else if(LHS.isString() && RHS.isString())
                setValue(node, ValueFactory.makeValue(
                    !Value.getString(LHS).equals(Value.getString(RHS))));
            else if(LHS.isChar() && RHS.isChar())
                setValue(node, ValueFactory.makeValue(
                    Value.getChar(LHS) != Value.getChar(RHS)));
            else{
                throwError("Operator != cannot be applied" +
                   " to " + LHS.getType() + "," + RHS.getType(),
                  node.getNotequal().getLine());
            }
        }catch(ValueException e){
            throwError("Failed to retrieve one of the" +
              " operands for != operation.", node.getNotequal().getLine());
        }
    }
    /* ************************************************************

    relational =
    {math} math_expression |
    {lt} relational less math_expression |
    {gt} relational greater math_expression |
    {lteq} relational lessequal math_expression |
    {gteq} relational greaterequal math_expression;

    **************************************************************/

    public void caseAMathRelational(AMathRelational node)
    {
        node.getMathExpression().apply(this);
        setValue(node, getValue(node.getMathExpression()));
    }

    public void caseALtRelational(ALtRelational node)
    {
        node.getRelational().apply(this);
        node.getMathExpression().apply(this);

        Value LHS = getValue(node.getRelational()),
              RHS = getValue(node.getMathExpression());

        try{
            if(LHS.isNumber() && RHS.isNumber())
                setValue(node, ValueFactory.makeValue(
                    Value.getNumber(LHS) < Value.getNumber(RHS)));
            else if(LHS.isString() && RHS.isString())
                setValue(node, ValueFactory.makeValue(
                    LHS.toString().compareTo(RHS.toString())<0));
            else if(LHS.isChar() && RHS.isChar())
                setValue(node, ValueFactory.makeValue(
                    Value.getChar(LHS) < Value.getChar(RHS)));
            else{
                throwError("Operator < cannot be applied" +
                   " to " + LHS.getType() + "," + RHS.getType(),
                  node.getLess().getLine());
            }
        }catch(ValueException e){
            throwError("Failed to retrieve one of the" +
              " operands for < operation.", node.getLess().getLine());
        }
    }

    public void caseAGtRelational(AGtRelational node)
    {
        node.getRelational().apply(this);
        node.getMathExpression().apply(this);

        Value LHS = getValue(node.getRelational()),
              RHS = getValue(node.getMathExpression());

        try{
            if(LHS.isNumber() && RHS.isNumber())
                setValue(node, ValueFactory.makeValue(
                    Value.getNumber(LHS) > Value.getNumber(RHS)));
            else if(LHS.isString() && RHS.isString())
                setValue(node, ValueFactory.makeValue(
                    LHS.toString().compareTo(RHS.toString())>0));
            else if(LHS.isChar() && RHS.isChar())
                setValue(node, ValueFactory.makeValue(
                    Value.getChar(LHS) > Value.getChar(RHS)));
            else{
                throwError("Operator > cannot be applied" +
                   " to " + LHS.getType() + "," + RHS.getType(),
                  node.getGreater().getLine());
            }
        }catch(ValueException e){
            throwError("Failed to retrieve one of the" +
              " operands for > operation.", node.getGreater().getLine());
        }
    }

    public void caseALteqRelational(ALteqRelational node)
    {
        node.getRelational().apply(this);
        node.getMathExpression().apply(this);

        Value LHS = getValue(node.getRelational()),
              RHS = getValue(node.getMathExpression());

        try{
            if(LHS.isNumber() && RHS.isNumber())
                setValue(node, ValueFactory.makeValue(
                    Value.getNumber(LHS) <= Value.getNumber(RHS)));
            else if(LHS.isString() && RHS.isString())
                setValue(node, ValueFactory.makeValue(
                    LHS.toString().compareTo(RHS.toString())<=0));
            else if(LHS.isChar() && RHS.isChar())
                setValue(node, ValueFactory.makeValue(
                    Value.getChar(LHS) <= Value.getChar(RHS)));
            else{
                throwError("Operator <= cannot be applied" +
                   " to " + LHS.getType() + "," + RHS.getType(),
                  node.getLessequal().getLine());
            }
        }catch(ValueException e){
            throwError("Failed to retrieve one of the" +
              " operands for <= operation.", node.getLessequal().getLine());
        }
    }

    public void caseAGteqRelational(AGteqRelational node)
    {
        node.getRelational().apply(this);
        node.getMathExpression().apply(this);

        Value LHS = getValue(node.getRelational()),
              RHS = getValue(node.getMathExpression());

        try{
            if(LHS.isNumber() && RHS.isNumber())
                setValue(node, ValueFactory.makeValue(
                    Value.getNumber(LHS) >= Value.getNumber(RHS)));
            else if(LHS.isString() && RHS.isString())
                setValue(node, ValueFactory.makeValue(
                    LHS.toString().compareTo(RHS.toString())>=0));
            else if(LHS.isChar() && RHS.isChar())
                setValue(node, ValueFactory.makeValue(
                    Value.getChar(LHS) >= Value.getChar(RHS)));
            else{
                throwError("Operator >= cannot be applied" +
                   " to " + LHS.getType() + "," + RHS.getType(),
                  node.getGreaterequal().getLine());
            }
        }catch(ValueException e){
            throwError("Failed to retrieve one of the" +
              " operands for >= operation.", node.getGreaterequal().getLine());
        }
    }

    /* ************************************************************

    field_access =
    {array} name array_access+ |
    {scalar} name;

    **************************************************************/

    public void caseAScalarFieldAccess(AScalarFieldAccess node)
    {
        node.getName().apply(this);
        setValue(node, getValue(node.getName()));
    }

    /* ************************************************************

    name =
    {simple_name}     simple_name |
    {qualified_name}  qualified_name;

    **************************************************************/

    public void caseASimpleNameName(ASimpleNameName node)
    {
        node.getSimpleName().apply(this);
        setValue(node, getValue(node.getSimpleName()));
        cleanValue(node.getSimpleName());
    }

    public void caseAQualifiedNameName(AQualifiedNameName node)
    {
        node.getQualifiedName().apply(this);
        setValue(node, getValue(node.getQualifiedName()));
        cleanValue(node.getQualifiedName());
    }


    /* ************************************************************
      simple_name = identifier
      simple variable
     *************************************************************/

    public void caseAArrayFieldAccess(AArrayFieldAccess node)
    {
        String fullName;
        node.getName().apply(this);
        fullName = getValue(node.getName()).toString();
        cleanValue(node.getName());

        Object temp[] = node.getArrayAccess().toArray();
        for(int i = 0; i < temp.length; i++)
        {
            ((PArrayAccess) temp[i]).apply(this);
            fullName += getValue((Node)temp[i]).toString();
            cleanValue((Node)temp[i]);
        }
        setValue(node,ValueFactory.makeValue(fullName));
    }

    public void caseASimpleName(ASimpleName node)
    {
        //setValue(node,getVariable(node.getIdentifier().getText()));
        setValue(node,ValueFactory.makeValue(node.getIdentifier().getText()));
    }

    public void caseAQualifiedName(AQualifiedName node)
    {
        String fullQualifiedName = node.getIdentifier().getText().trim();

        node.getFieldAccess().apply(this);
        fullQualifiedName = getValue(node.getFieldAccess()).toString() +
         "." + fullQualifiedName;

        setValue(node,ValueFactory.makeValue(fullQualifiedName));
        IMTrace.out(fullQualifiedName);
        cleanValue(node.getFieldAccess());
    }

    public void caseAArrayAccess(AArrayAccess node)
    {
        node.getMathExpression().apply(this);
        setValue(node,
          ValueFactory.makeValue("["+
            getValue(node.getMathExpression()).toString() + "]"));
        cleanValue(node.getMathExpression());
    }


    /**
     * KInterBreakException
     * Exception thrown when a break statement is encountered.<br>
     * Should only be used inside this class. Used to simulate the
     * behaviour of java break statement, not very elegant i know, but
     * only way i could think of doing it without backtracking thru the
     * AST.
     */
    private class KInterBreakException extends RuntimeException
    {
        public KInterBreakException(int lineNum)
        {
            super("Error! break outside of switch or loop. " +
              "Line: " + lineNum);
        }
    }

    /**
     * KInterReturnException
     * Exception thrown when a return statement is encountered.<br>
     * Should only be used inside this class. Used to simulate the
     * behaviour of java return statement, not very elegant i know, but
     * only way i could think of doing it without backtracking thru the
     * AST.
     */
    private class KInterReturnException extends RuntimeException
    {
        public KInterReturnException(int lineNum)
        {
            super("Program returned. (Line: " + lineNum + ")");
        }
    }
}

