/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001,2002 Robert Chatley, Alan Chan                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

/*
 * A member of a set of classes which has usage similar to that of a
 * union type in C. See the class Value for more details.
 *
 * This class is used to store the value of an array variable in a Kenya
 * program when running the Kenya interpreter, a Vector is used to model
 * the array.
 */

package com.chatley.kenya.interpreter.value;

import java.util.*;
import com.chatley.kenya.interpreter.*;
import com.chatley.kenya.utils.IMTrace;

public class ArrayValue extends ReferenceValue
{

    private Value[] mArray = null;

    /** Type of components stored by this array.*/
    private String mType;


    /**
     * Constructs a new ArrayValue obj representing an array of kenya class objs
     * of the specified type, with the given variable name and specified length.
     * @param size the length of the array which this object will represent.
     * @param varName the variable name of this array.
     * @param klass the name of the class of objects this array will hold.
     *
     * @throws ArrayValueException if error occurred during array creation.
     */
    public ArrayValue(int size, String varName, Value klass)
      throws ArrayValueException
    {
        mArray = new Value[size];
        try{
            for(int i=0;i<size;i++)
                mArray[i] = (Value)klass.clone();
        }catch(Exception e){
            throw new ArrayValueException("Error creating array of " +
              klass + " ---> " + varName);
        }

        setVarName(varName);
        mType = klass.getType();
    }

    /**
     * Constructs a new ArrayValue object representing an array of primitives
     * of the specified type, with the given variable name and length.
     * @param size the length of the array which this object will represent.
     * @param varName the variable name of this array.
     * @param aType  the type of primitives this array will hold as a string,
     *        for example, "int".
     */
    public ArrayValue(int size, String varName, String aType)
    {
        mArray = new Value[size];

        for(int i=0;i<size;i++)
            mArray[i] = ValueFactory.getDefaultPrimValue(aType);

        setVarName(varName);

        mType = aType;
    }

    /**
     * Constructs a new ArrayValue object representing an array of primitives
     * of the specified type, with the specified length.
     * @param size the length of the array which this object will represent.
     * @param aType  the type of primitives this array will hold.
     */
    public ArrayValue(int size, String aType)
    {
        this(size,"",aType);
    }

    /**
     * Constructs a new ArrayValue object representing an array of objects
     * of the specified class, with the specified length.
     * @param size the length of the array which this object will represent.
     * @param aClass  the type of primitives this array will hold.
     *
     * @throws ArrayValueException if error occurred during array creation.
     */
    public ArrayValue(int size, Value aClass) throws ArrayValueException
    {
        this(size,"",aClass);
    }


    /**
     * Constructs a new ArrayValue object representing an array of objects
     * initialised with the array passed as parameter.
     * @param aData the array of Value of objects to intialise with.
     * @param aType the type of objects held by the array.
     */
    public ArrayValue(Value[] aData, String aType)
    {
        mArray = aData;
        mType = aType;
    }

    /**
     * creates an empty ArrayValue object of primitives where the array is null.
     * @param aType type that this array will store.
     * @param aDims dimensions of array
     */
    public ArrayValue(String aType)
    {
        mType = aType;
    }

    /**
     * parses an array access of the form <code>arrayName[x][y][z]...</code>
     * and returns the value at the specified index.
     * @param indexAsString the string representation of the array access.
     * @returns the value in the array at the specified index
     * @throws ArrayValueException if the index is out of bounds.
     */
    public Value getValueAt(String indexAsString) throws ArrayValueException
    {

        if(mArray == null) throw new
          ArrayValueException("Tried to access uninitialised Array -> " +
            indexAsString);

        int openBracket=indexAsString.indexOf('[');
        String name = indexAsString.substring(0,openBracket);
        int closeBracket = indexAsString.indexOf(']');
        Value returnV = null;

        int dims = 1;
        int index = 0;

        try{
            index =
              Integer.parseInt(indexAsString.substring(openBracket+1,closeBracket));

            returnV = mArray[index];

            while((openBracket=indexAsString.indexOf('[',closeBracket))!=-1)
            {
                IMTrace.out("dims = " + dims);
                closeBracket = indexAsString.indexOf(']',openBracket);
                index =
                  Integer.parseInt(indexAsString.substring(openBracket+1,closeBracket));

                returnV = ((ArrayValue)returnV).getValueAt(index);

                dims++;

            }
        }catch(ArrayIndexOutOfBoundsException e){
            e.printStackTrace();
            throw new ArrayValueException("Array index out of bounds: " +
              indexAsString + " --> " + index);
        }
        return returnV;
    }

    /**
     * returns the element at the given index.
     * @param index the index to return
     * @returns the value in the array at the specified index.
     * @throws ArrayValueException if the index is out of bounds.
     */
    public Value getValueAt(int index) throws ArrayValueException
    {
        //IMTrace.out(mArray.elementAt(index).toString());
        //return (Value)mArray.elementAt(index);

        if(mArray == null) throw new
          ArrayValueException("Tried to access uninitialised Array -> " + index);

        try{
            return mArray[index];
        }catch(ArrayIndexOutOfBoundsException e){
            e.printStackTrace();
            throw new ArrayValueException("Array index out of bounds: " +
              index);
        }
    }

    /**
     * parses an array access of the form <code>arrayName[x][y][z]...</code>
     * and set the value at the specified index to the value passed.
     * @param indexAsString the string representation of the arrary access.
     * @param val the value to set
     * @throws ArrayValueException if the index is out of bounds.
     */
    public void setValueAt(String indexAsString, Value val) throws ArrayValueException
    {
        IMTrace.out("indexAsString = " + indexAsString);

        if(mArray == null) throw new
          ArrayValueException("Tried to access uninitialised Array -> " +
            indexAsString);

        int openBracket=indexAsString.indexOf('[');
        String name = indexAsString.substring(0,openBracket);
        int closeBracket = indexAsString.indexOf(']');
        Value v = null;

        int dims = 0;
        int[] indexes = new int[ArrayDim.MAX_SIZE];

        int index =
          Integer.parseInt(indexAsString.substring(openBracket+1,closeBracket));

        indexes[dims++] = index;

        while((openBracket=indexAsString.indexOf('[',closeBracket))!=-1 &&
          dims<ArrayDim.MAX_SIZE)
        {
            closeBracket = indexAsString.indexOf(']',openBracket);
            index =
              Integer.parseInt(indexAsString.substring(openBracket+1,closeBracket));
            indexes[dims++] = index;
        }

        IMTrace.out("dims = " + dims);

        try
        {
           Value ka; // just a temp var
           // we only clone if we have a primitive type, coz we want
           // aliasing with classes in order to be 'faithful to java'
           if(val.isClass()) ka = val;
           else ka = (Value)val.clone();


           if(dims==1){
               mArray[indexes[0]] = ka;//(Value)val.clone();
           }
           else{
               v = mArray[indexes[0]];
               for(int i=1;i<dims-1;i++){
                   v = ((ArrayValue)v).getValueAt(indexes[i]);
               }
               ((ArrayValue)v).setValueAt(indexes[dims-1],ka/*(Value)val.clone()*/);
           }
        }catch(CloneNotSupportedException e){
            e.printStackTrace();
        }catch(ArrayIndexOutOfBoundsException e){
            e.printStackTrace();
            throw new ArrayValueException("Array index out of bounds: " +
              index);
        }
    }

    /**
     * set the element at the given index.
     * @param index the index to return
     * @param v the value to set.
     * @throws ArrayValueException if the index is out of bounds.
     */
    public void setValueAt(int index, Value v) throws ArrayValueException
    {
        if(mArray == null) throw new
          ArrayValueException("Tried to access uninitialised Array -> " +
            index);

        try{
            mArray[index] = (Value)v.clone();
            IMTrace.out("setting value at: " + index + " " + v);
        }catch(CloneNotSupportedException e){
            e.printStackTrace();
        }catch(ArrayIndexOutOfBoundsException e){
            e.printStackTrace();
            throw new ArrayValueException("Array index out of bounds: " +
              index);
        }
    }

    /**
     * Reports the type of this object as the type of Values it stores, for
     * example, if this is an array of <code>int</code>s, the string
     * <code>int</code> will be returned.
     */
    public String getType()
    {
        return mType;
    }

    public boolean isArray()
    {
        return true;
    }

    /**
     * returns the length of this array encapsulated in a Value object.
     */
    public Value getLengthAsValue()
    {
        return ValueFactory.makeValue(getLength());
    }

    /**
     * returns the length of this array, or 0 if the array is null.
     */
    public int getLength()
    {
    	return ((mArray==null)? 0 : mArray.length);
    }

    public Object clone() throws CloneNotSupportedException
    {
        if(getLength()>0){
           Value[] deepCopy = new Value[mArray.length];
           for(int i=0;i<mArray.length;i++)
               deepCopy[i] = (Value)mArray[i].clone();
           ArrayValue clone = new ArrayValue(deepCopy,mType);
           clone.setVarName(new String(getVarName()));
           return (Object)clone;
        }
        else{
            ArrayValue ka = new ArrayValue(mType);
            ka.setVarName(new String(getVarName()));
            return (Object) ka;
        }
    }

    /**
     * implementation of the <code>ValueSwitch</code> interface.
     */
    public Object apply(ValueSwitch vs)
    {
        return vs.caseAArrayValue(this);
    }

    public String toString()
    {
        return ("Array of length " + getLength());
    }
}