/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001,2002 Robert Chatley, Alan Chan                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

/*
 * A member of a set of classes which has usage similar to that of a
 * union type in C. See the class Value for more details.
 *
 * This class is used to store the value of a class variable in a Kenya
 * program when running the Kenya interpreter. A HashMap of Value objects
 * maps field names to their values.
 *
 * As an example, consider the following kenya class declaration:-
 *
 * class kaka{
 *   int ka;
 *   String name;
 * }
 *
 * If we now have a variable of this type:-
 *
 * kaka karen;
 *
 * then, the variable karen will be represented using a instance of this class,
 * and the mFields HashMap will have two keys called 'ka' and 'name',
 * respectively, and then mClassName field will hold 'kaka'.
 *
 *
 * Alan Chan 2002/2
 */

package com.chatley.kenya.interpreter.value;

import java.util.*;

public class ClassValue extends ReferenceValue
{

   /**
    * The fields of the class are represented using a HashMap of
    * Value objects, with the field name as the key.
    */
    private HashMap mFields;

    /** Name of class, which is the name used in the class declaration in
     *  the kenya program.*/
    private String mClassName;

    public ClassValue(String aClassName, HashMap aFields)
    {
        mClassName = aClassName;
        mFields = aFields;
    }

    public ClassValue(String aClassName, String varName)
    {
       mClassName = aClassName;
       setVarName(varName);
    }

    public Value getValue(String varName) throws
      InvalidFieldException, ArrayValueException
    {
        int firstDot = varName.indexOf('.');
        String var = (firstDot==-1)? varName : varName.substring(0,firstDot);

        int openBracket = var.indexOf('[');
        String field = (openBracket==-1)? var : var.substring(0,openBracket);

        if(!mFields.containsKey(field)){
            throw new InvalidFieldException("field " + field +
              " not found in class " + getType());
        }

        Value result = (Value)mFields.get(field);

        if(openBracket!=-1) // we have an array access
        {
            if(result.isArray())
                result = ((ArrayValue)result).getValueAt(var);
            else
                throw new InvalidFieldException("array access " + var
                  + " attempted on field that's not an array");
        }

        return result;
    }

    public void setValue(String varName, Value v) throws
      InvalidFieldException, ArrayValueException
    {
        //IMTrace.out("Inside ClassValue.setValue(), varName = " + varName);
        int firstDot = varName.indexOf('.');
        String var = (firstDot==-1)? varName : varName.substring(0,firstDot);

        int openBracket = var.indexOf('[');
        String field = (openBracket==-1)? var : var.substring(0,openBracket);

        if(!mFields.containsKey(field)){
            throw new InvalidFieldException("field " + field +
              " not found in class " + getType() + ", cannot write.");
        }

        Value result = (Value)mFields.get(field);

        if(openBracket!=-1) // we have an array access
        {
            if(result.isArray()){
                ((ArrayValue)result).setValueAt(var,v);
                //IMTrace.out("Setting array field " + var + " value = " + v.toString());
            }
            else
                throw new InvalidFieldException("array access " + var
                  + " attempted on field that's not an array");
        }else{
            try{ // in order to 'be faithful to java' we need to have aliasing
                if(v.isClass()) mFields.put(field,v);
                else mFields.put(field,v.clone());
                //IMTrace.out("Setting field " + field + " value = " + mFields.get(field).toString());
            }catch(CloneNotSupportedException e){e.printStackTrace();}
        }
    }

    /**
     * Reports the type of this class.
     */
    public String getType()
    {
        return mClassName;
    }

    public boolean isClass()
    {
        return true;
    }

    /**
     * peforms a deep copy
     */
    public Object clone() throws CloneNotSupportedException
    {
        HashMap deepCopy = new HashMap();

        for(Iterator i = mFields.keySet().iterator();i.hasNext();)
        {
            String key = (String)i.next();
            Object v = mFields.get(key);
            v = ((Value)v).clone();
            deepCopy.put(new String(key),v);

        }

        ClassValue clone =  new ClassValue(new String(mClassName),deepCopy);
        clone.setVarName(new String(getVarName()));
        return (Object)clone;
    }

    public Object apply(ValueSwitch vs)
    {
        return vs.caseAClassValue(this);
    }

    /**
     * returns the names of the fields of by this class as
     * an array of strings.
     */
    public String[] fieldNames()
    {
        Set fieldNames = mFields.keySet();
        String[] names = new String[fieldNames.size()];
        int k = 0;
        for(Iterator i = fieldNames.iterator();i.hasNext();k++)
            names[k] = (String)i.next();

        return names;
    }

    public String toString()
    {
        StringBuffer des = new StringBuffer();

        des.append("Class: ").append(getType()).append(" [\n\r");

        for(Iterator i = mFields.keySet().iterator();i.hasNext();)
        {
            String fldName = (String)i.next();
            Value ka = (Value)mFields.get(fldName);
            des.append(fldName).append("\t= ");

            if(!ka.isClass())
            {
                 des.append(ka.toString());
            }
            // need to have this to stop infinite loop
            // when we have back references
            else
            {
                des.append("Class value");
            }
            des.append("\n\r");
        }
        des.append("]\n\r");

        return des.toString();
    }
}