/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001,2002 Robert Chatley, Alan Chan                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

/*
 * Class that encapsulates Kenya data types in an object.
 * This class is used sort of like a union data structure in C,
 * however the memory savings of a union are not reflected in this
 * class =(
 *
 * This class is used to store the value of a variable in a Kenya
 * program when running the Kenya interpreter.
 *
 * askc98 2002/2
 */

package com.chatley.kenya.interpreter.value;

public abstract class Value implements Cloneable
{
    /** If a variable has this value, we can keep a note here.*/
    protected String mVarName  = "";


    /**
     * Keep a note that the variable with name varName has
     * the value represented by this object.
     *
     * @param varName The name of the variable that has the value
     *                represented by this object.
     */
    public void setVarName(String varName)
    {
        mVarName = varName;
    }

    /**
     * Returns the name of a variable that *may* have the value
     * represented by this object. Note that it is up to the
     * programmer to keep this field updated for it to be useful.
     * Also note that this method can also return null if no note
     * of a variable name was registered with this object.
     */
    public String getVarName()
    {
        return mVarName;
    }

    /**
     * Returns the type of this object as a string.
     * If the object does not hold one of the standard
     * Kenya data types, the string "null" is returned.
     */
    public abstract String getType();

    /**
     * Returns true if this Value object holds an int,
     * false otherwise.
     */
    public boolean isInt()
    {
        return false;
    }

    public boolean isDouble()
    {
        return false;
    }

    public boolean isNumber()
    {
        return false;
    }

    public boolean isBool()
    {
        return false;
    }

    public boolean isString()
    {
        return false;
    }

    public boolean isChar()
    {
        return false;
    }

    public boolean isArray()
    {
        return false;
    }

    public boolean isClass()
    {
        return false;
    }

    public boolean isNull()
    {
        return false;
    }

    /**
     * returns true iff this value object represents a primitive value,
     * i.e., an int, double, char, boolean or string (strings are considered
     * as primitive values in kenya).
     */
    public boolean isPrimitiveValue()
    {
        return false;
    }

    /**
     * returns true iff this value object represents a reference value,
     * this includes arrays, the null value and user defined classes.
     */
    public boolean isReferenceValue()
    {
        return false;
    }


    public abstract Object apply(ValueSwitch v);

    public static int getInt(Value v) throws NotAnIntException
    {
        if(v.isInt()) return ((IntValue)v).getInt();
        else throw new NotAnIntException();
    }

    public static boolean getBool(Value v) throws NotABoolException
    {
        if(v.isBool()) return ((BoolValue)v).getBool();
        else throw new NotABoolException();
    }

    public static double getDouble(Value v) throws NotADoubleException
    {
        if(v.isDouble()) return ((DoubleValue)v).getDouble();
        else throw new NotADoubleException();
    }

    public static double getNumber(Value v) throws NANException
    {
        if(v.isNumber()) return ((NumberValue)v).getNumber();
        else throw new NANException();
    }

    public static char getChar(Value v) throws NotACharException
    {
        if(v.isChar()) return ((CharValue)v).getChar();
        else throw new NotACharException();
    }

    public static String getString(Value v) throws NotAStringException
    {
        if(v.isString()) return ((StringValue)v).getString();
        else throw new NotAStringException();
    }

    public Object clone() throws CloneNotSupportedException
    {
        return super.clone();
    }
}