/* *******************************************************************************
 *   Kenya - an introductory programming language and development environment    *
 *   Copyright (C) 2001,2002 Robert Chatley, Alan Chan                           *
 *                                                                               *
 *   This program is free software; you can redistribute it and/or               *
 *   modify it under the terms of the GNU General Public License                 *
 *   as published by the Free Software Foundation; either version 2              *
 *   of the License, or (at your option) any later version.                      *
 *                                                                               *
 *   This program is distributed in the hope that it will be useful,             *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 *   GNU General Public License for more details.                                *
 *                                                                               *
 *   You should have received a copy of the GNU General Public License           *
 *   along with this program; if not, write to the Free Software                 *
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
 *                                                                               *
 *   The author can be contacted by email at robert@chatley.com                  *
 *                                                                               *
 *********************************************************************************/

/**
 * ValueFactory.java
 *
 * Contains a set of overloaded static convenience methods for creating
 * Value objects.
 *
 * askc98 2002/2
 */

package com.chatley.kenya.interpreter.value;

import com.chatley.kenya.interpreter.*;

public class ValueFactory
{
    private ValueFactory(){} // private constructor

    /////////////////////// Basic types ///////////////////////////////////

    /** Initialise to hold an integer. */
    public static Value makeValue(int m)
    {
        return new IntValue(m);
    }

    /** Initialise to hold string. */
    public static Value makeValue(String s)
    {
        return new StringValue(s);
    }

    /** Initialise to hold a character. */
    public static Value makeValue(char c)
    {
        return new CharValue(c);
    }

    /** Initialise to hold a boolean. */
    public static Value makeValue(boolean b)
    {
        return new BoolValue(b);
    }

    /** Initialise to hold a double. */
    public static Value makeValue(double d)
    {
        return new DoubleValue(d);
    }

    /**
     * Initialise to hold an integer, keeping a note
     * that the variable with name varName has this value.
     */
    public static Value makeValue(String varName, int m)
    {
        return new IntValue(m,varName);
    }

    /**
     * Initialise to hold a String, keeping a note
     * that the variable with name varName has this value.
     */
    public static Value makeValue(String varName, String s)
    {
        return new StringValue(s,varName);
    }

    /**
     * Initialise to hold a character, keeping a note
     * that the variable with name varName has this value.
     */
    public static Value makeValue(String varName, char c)
    {
        return new CharValue(c,varName);
    }

    /**
     * Initialise to hold a boolean, keeping a note
     * that the variable with name varName has this value.
     */
    public static Value makeValue(String varName, boolean b)
    {
        return new BoolValue(b,varName);
    }

    /**
     * Initialise to hold a double, keeping a note
     * that the variable with name varName has this value.
     */
    public static Value makeValue(String varName, double d)
    {
        return new DoubleValue(d,varName);
    }

    /**
     * Initialise to hold null, keeping a note of its type, and
     * that the variable with name varName has this value.
     */
    public static Value makeNull(String type, String varName)
    {
        return new NullValue(type,varName);
    }

    /**
     * Initialise to hold null, keeping a note
     * of its type.
     */
    public static Value makeNull(String type)
    {
        return new NullValue(type);
    }

    public static Value getDefaultPrimValue(String primType)
    {
        Value value;

        if(primType.equals("int"))
            value = ValueFactory.makeValue(0);
        else if(primType.equals("double"))
            value = ValueFactory.makeValue(0.0d);
        else if(primType.equals("boolean"))
            value = ValueFactory.makeValue(false);
        else if(primType.equals("String"))
            value = ValueFactory.makeValue("");
        else if(primType.equals("char"))
            value = ValueFactory.makeValue('\0');
        else value = null;

        return value;

    }

    //////////////////////////// Arrays ///////////////////////////////////

    /**
     * Constructs a new ArrayValue object representing an array of primitives
     * of the specified type, with the given variable name and length.
     * @param size the length of the array required.
     * @param varName the variable name of this array.
     * @param aType  the type of primitives this array will hold as a string,
     *         for example, "int".
     */
    public static Value makeValue(String varName,int size, String type)
    {
        return new ArrayValue(size,varName,type);
    }

    /**
     * Constructs a new ArrayValue object representing an array of primitives
     * of the specified type, with the specified length.
     * @param size the length of the array which this object will represent.
     * @param aType  the type of primitives this array will hold.
     */
    public static Value makeValue(int size, String type)
    {
        return new ArrayValue(size,type);
    }

    /**
     * Constructs a new ArrayValue obj representing an array of kenya class objs
     * of the specified type, with the given variable name and specified length.
     * @param size the length of the array required.
     * @param varName the variable name of this array.
     * @param klass the name of the class of objects the array will hold.
     *
     * @throws ArrayValueException if error occurred during array creation.
     */
    public static Value makeValue(String varName,int size, Value klass)
      throws ArrayValueException
    {
        return new ArrayValue(size,varName,klass);
    }

    /**
     * Constructs a new ArrayValue object representing an array of objects
     * of the specified class, with the specified length.
     * @param size the length of the array which this object will represent.
     * @param aClass  the type of primitives this array will hold.
     *
     * @throws ArrayValueException if error occurred during array creation.
     */
    public static Value makeValue(int size, Value aClass)
      throws ArrayValueException
    {
        return new ArrayValue(size,aClass);
    }

}